/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : alarm.c                                                  */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : Provides timer functions for use by postit node.         */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Jun 2001                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include <postit.h> /* Specific header for this dll                            */

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <threads.h>
#include <time.h>
#include <cmqc.h>


/*
 * Reset the alarm element for a new postit 
 */
void setPostitTimeoutAlarm(IN postitCreateInstance_t *postitInstanceP) {

	/*
	 * Assumption:
	 *
	 * On entry to this function, the caller has locked the postit chain
	 */
	time_t currentTime;
	int timeoutInterval;
	postitCreateInstance_t *currentPostitP;

    TRACE_ENTRY("setPostitTimeoutAlarm", ("postitInstanceP=%.8x", postitInstanceP));

	/* calculate the new timeout interval */
	timeoutInterval = postitInstanceP->deployedPostitCreateP->timeOutInterval;
    TRACE_MID(TRACE_INFO, ("Timeout interval is %d", timeoutInterval));
	time(&currentTime);
	postitInstanceP->timeoutTime = currentTime + timeoutInterval;
    TRACE_MID(TRACE_INFO, ("Timeout value is %d", postitInstanceP->timeoutTime));

	/* find position in alarm chain for this postit based on its expiry time */
    currentPostitP = &instanceAnchor;
    TRACE_MID(TRACE_INFO, ("Starting search of alarm chain at anchor (%.8x)",
		                   currentPostitP));

	while(TRUE) {
        TRACE_MID(TRACE_INFO, ("Checking postit instance at %.8x "
			                   "(timeout value = %d)", currentPostitP,
							   currentPostitP->timeoutTime));

		if (!currentPostitP->nextAlarmP ||
			currentPostitP->nextAlarmP->timeoutTime > currentPostitP->timeoutTime) {
            TRACE_MID(TRACE_INFO, ("Inserting alarm after postit at %.8x",
				                   currentPostitP));
			break;
		}

		currentPostitP = currentPostitP->nextAlarmP;
	}

    /* add after current alarm member */
	INSERT_AFTER_IN_ALARM_CHAIN(currentPostitP, postitInstanceP);
    TRACE_RETURN_VOID(RSN_001);
}

/******************************************************************************/
/* processAlarms                                                              */
/******************************************************************************/
threadFunction_t processAlarms(void *dataP) {
	time_t currentTime;
	int timeToWait;
	postitCreateInstance_t *thisAlarmP,
		                   *nextAlarmP;
	bool_t globalScope = FALSE,
		   brokerScope = FALSE;

    TRACE_ENTRY("processAlarms", ("dataP=%.8x", dataP));

	/* let the parent thread know that we're up and running */
	alarmThreadInitialised = TRUE;

	/* lock the alarm thread sync mutex and post the parent thread */
    TRACE_MID(TRACE_INFO, ("About to signal parent for initialisation complete"));
	threadMutexLock(alarmThreadSyncMutex);
	threadCondSignal(alarmThreadSyncCondVar);
	threadMutexUnlock(alarmThreadSyncMutex);

	/* 
	 * The postit chain and the alarm chain are closely related, so we lock
	 * both via a single mutex.
	 */
	TRACE_MID(TRACE_INFO, ("About to lock postitChainMutex"));
	threadMutexLock(postitChainMutex);

	/* keep processing alarm chain until requested to shutdown */
	while(!shutdownAlarmRequested) {
        TRACE_MID(TRACE_INFO, ("Processing alarm element chain"));

		thisAlarmP = instanceAnchor.nextAlarmP;

		/* process any expired alarm elements */
	    while(thisAlarmP) {
 		    /* run down the list of alarm elements */
            TRACE_MID(TRACE_INFO, ("Processing alarm element at %.8x",
				                    thisAlarmP));

			/* address next alarm element, just in case we need to free the current */
            TRACE_MID(TRACE_INFO, ("thisAlarmP->nextAlarmP = %.8x",
				                    thisAlarmP->nextAlarmP));
			nextAlarmP = thisAlarmP->nextAlarmP;

			/* 
			 * if the Timeout interval has expired, then process any
			 * messages we've received up till this point.
			 */
			time(&currentTime);
            TRACE_MID(TRACE_INFO, ("thisAlarmP->timeoutTime = %d, currentTime=%d",
				                    thisAlarmP->timeoutTime, currentTime));
			if (thisAlarmP->timeoutTime <= currentTime) {
				TRACE_MID(TRACE_INFO, ("Postit instance %.8x has expired",
										thisAlarmP));

				// if this is a global postit, need to take appropriate lock
				globalScope = thisAlarmP->global;
				if (globalScope) {
				    TRACE_MID(TRACE_INFO, ("Taking global lock"));
					if (thisAlarmP->scope == SCOPE_BROKER) {
						brokerScope = TRUE;
					}
					else {
						brokerScope = FALSE;
					}
				    shmLock(brokerScope);
				}
				freePostit(thisAlarmP);
				if (globalScope) {
				    TRACE_MID(TRACE_INFO, ("Freeing global lock"));
				    shmUnlock(brokerScope);
				}

				/* process next in list... */
				thisAlarmP = nextAlarmP;
				continue;
			}

			/* alarms are in order of expiry time...so nothing more to do here */
			break;
		}

		/* reset the alarm thread to pop at next expiry interval */
		if (instanceAnchor.nextAlarmP) {
			/* reset the alarm for the next element (milliseconds) */
			timeToWait = 
				(instanceAnchor.nextAlarmP->timeoutTime - currentTime) * 1000;
            TRACE_MID(TRACE_INFO, ("Next expiry interval time is %d", timeToWait));
		}
		else {
			timeToWait = INFINITE;
            TRACE_MID(TRACE_INFO, ("Next expiry interval time is INFINITE"));
		}

	    /* wait for next alarm to pop or a new element to appear on the chain */
        TRACE_MID(TRACE_INFO, ("Alarm thread about to sleep for expiry interval"));
		threadCondTimedWait(alarmThreadWorkCondVar, timeToWait);
        TRACE_MID(TRACE_INFO, ("Alarm thread woken"));
	}

    TRACE_MID(TRACE_INFO, ("Alarm thread terminating"));

	threadMutexUnlock(postitChainMutex);

    TRACE_RETURN_TYPE(RSN_001, FUNCTION_RETURN_TYPE, 0);
}

