/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : apply.c                                                  */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : Contains functions for adding postit data to messages    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Jun 2001                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include <postit.h>      /* Specific header for this dll                       */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <BipCni.h>		/* Function prototypes for Plug-in Node functions     */

/* local function prototypes */
globalPostitCreateInstance_t *copyFromGlobalToLocal(
				                   IN globalPostitCreateInstance_t *globalInstanceP);
/******************************************************************************/
/* evaluatePostitApply                                                        */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  destinationList (input) - address of the input destination list for the   */
/*                            message                                         */
/*  exceptionList   (input) - address of the exception list for the message   */
/*  message         (input) - address of the input message object             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Applies previously stored postit data to the specified location in the     */
/* current message.  Postits are matched by either MsgId or CorrelId or both. */
/*                                                                            */
/******************************************************************************/

typedef struct fullKey {
	char msgId[MQ_MSG_ID_LENGTH],
		 correlId[MQ_CORREL_ID_LENGTH];
} fullKey_t;

int evaluatePostitApply(IN context_t* contextP,
 			   		    IN CciMessage *destinationList,
		                IN CciMessage *exceptionList,
		                IN CciMessage *message)
{
	deployedPostitApply_t *deployedPostitApplyP;
	void *correlIdElementP,
		 *messageIdElementP;
    struct CciByteArray byteArrayDetails;
    CciSize numBytes;
	fullKey_t matchKey;
	int rc,
		retCode,
		numMatches,
		i;
	bool_t nullCorrelId,
		   nullMsgId;
	char nullId[MQ_MSG_ID_LENGTH];
	postitCreateInstance_t **postitList;
	CciMessage *newMsg;
	CciElement *copyToElement;
	CciMessageContext *msgContextP;
	postitCreateInstance_t *currentInstanceP;
	char mbString[MAX_STRING_SIZE + 1],
		 msgBuffer[1000];
    elementType_t elementType;
    static char *functionName="evaluatePostitApply";

    TRACE_ENTRY("evaluatePostitApply", ("contextP=%.8x, destinationList=%.8x, "
		                     "exceptionList=%.8x, message=%.8x",
							 contextP, destinationList,
							 exceptionList, message));

	/* Using the match conditions specified on the PostitApply node, look for any
	 * matching postits.  There may be multiple matches, in which case, apply each
	 * postit in order of its creation.
	 */

	deployedPostitApplyP = (deployedPostitApply_t *)(contextP->nodeDataP);

	/* extract the message id and correl id values */
	messageIdElementP = getMessageElement(message, MESSAGE_ID);

	if (!messageIdElementP)
	{
		/* generate an error, exit */
		/* log an error message and send invalid message to failure terminal */
		TRACE_MID(TRACE_ERROR, ("Message Id not found in message"));

		/* Log a failure message */
		cciLog(&rc, 
			   CCI_LOG_ERROR, 
			   __FILE__, 
			   __LINE__, 
			   functionName, 
			   C.PLUGIN_MESSAGE_CATALOGUE, 
			   MJB_FIELD_NOT_FOUND, 
			   "MessageId field missing in input message",
#if !defined(WIN32)
			   brokerName,
#endif
			   "MessageId",
			   (char *)0);

		/* send message to failure terminal */
		rc = propagateMessage(contextP,
						      C.TERM_FAILURE,
						      message,
						      destinationList,
						      exceptionList,
						      FALSE); /* not a new message */

		if (rc != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message "
					                "to 'failure' terminal"));
			RETHROW_IF_EXCEPTION(rc);
			TRACE_RETURN(RSN_046, TT_PROPAGATION_ERROR);
		}
		TRACE_RETURN(RSN_001, TT_MESSAGE_ID_NOT_FOUND);
	}

	/* copy message id, assume it is of type CCI_VALUE_TYPE_BYTEARRAY */
	byteArrayDetails.pointer = (void *)matchKey.msgId;
	byteArrayDetails.size = sizeof(matchKey.msgId);
	numBytes = cniElementByteArrayValue(&rc,
										messageIdElementP,
										&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract Message Id from message"));
		TRACE_RETURN(RSN_015, TT_MESSAGE_ID_NOT_FOUND);
	}

	TRACE_MID(TRACE_INFO, ("Incoming Message Id value is..."));
	TRACE_MIDHEX(TRACE_INFO, matchKey.msgId, MQ_MSG_ID_LENGTH);

	/* extract the correlId */
	correlIdElementP = getMessageElement(message, CORREL_ID);

	if (!correlIdElementP)
	{
		/* log an error message and send invalid message to failure terminal */
		TRACE_MID(TRACE_ERROR, ("Correlation Id not found in message"));

		/* Log a failure message */
		cciLog(&rc, 
			   CCI_LOG_ERROR, 
			   __FILE__, 
			   __LINE__, 
			   functionName, 
			   C.PLUGIN_MESSAGE_CATALOGUE, 
			   MJB_FIELD_NOT_FOUND, 
			   "CorrelId field missing in input message",
#if !defined(WIN32)
			   brokerName,
#endif
			   "CorrelId",
			   (char *)0);

		/* send message to failure terminal */
		rc = propagateMessage(contextP,
						      C.TERM_FAILURE,
						      message,
						      destinationList,
						      exceptionList,
						      FALSE); /* not a new message */

		if (rc != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message "
					                "to 'failure' terminal"));
			RETHROW_IF_EXCEPTION(rc);
			TRACE_RETURN(RSN_046, TT_PROPAGATION_ERROR);
		}

		TRACE_RETURN(RSN_003, TT_CORRELID_NOT_FOUND);
	}

	/* copy correl id, assume it is of type CCI_VALUE_TYPE_BYTEARRAY */
	byteArrayDetails.pointer = (void *)matchKey.correlId;
	byteArrayDetails.size = sizeof(matchKey.correlId);
	numBytes = cniElementByteArrayValue(&rc,
										correlIdElementP,
										&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract Correl Id from message"));
		TRACE_RETURN(RSN_015, TT_CORRELID_NOT_FOUND);
	}

	TRACE_MID(TRACE_INFO, ("Incoming Correl Id value is..."));
	TRACE_MIDHEX(TRACE_INFO, matchKey.correlId, MQ_CORREL_ID_LENGTH);

	/* Check for null keys, i.e. either MsgId or CorrelId set to binary zeroes */
	nullCorrelId = FALSE;
	nullMsgId = FALSE;
	memset(nullId, 0, sizeof(nullId));
	if (!memcmp(matchKey.correlId, nullId, sizeof(nullId)))
	{
		nullCorrelId = TRUE;
        TRACE_MID(TRACE_INFO, ("CorrelId is null"));
	}
	if (!memcmp(matchKey.msgId, nullId, sizeof(nullId)))
	{
		nullMsgId = TRUE;
        TRACE_MID(TRACE_INFO, ("MsgId is null"));
	}

	TRACE_MID(TRACE_INFO, ("About to lock postitChainMutex"));
	threadMutexLock(postitChainMutex);

	/* search required index for matching postits */ 
	numMatches = 0;
	switch (deployedPostitApplyP->matchValue)
	{
		case MATCH_TYPE_CORRELID_MSGID:
			if (nullCorrelId)
			{
		        TRACE_MID(TRACE_INFO, ("No match on null CorrelId"));
				break;
			}
	        TRACE_MID(TRACE_INFO, ("Using CorrelId value to search MsgId index"));
			numMatches = searchIndex(deployedPostitApplyP->postitName,
				                     matchKey.correlId, 
									 INDEX_MSGID, 
									 deployedPostitApplyP->scope,
									 &postitList);
			break;

		case MATCH_TYPE_MSGID_MSGID:
			if (nullMsgId)
			{
		        TRACE_MID(TRACE_INFO, ("No match on null MsgId"));
				break;
			}
	        TRACE_MID(TRACE_INFO, ("Using MsgId value to search MsgId index"));
			numMatches = searchIndex(deployedPostitApplyP->postitName,
				                     matchKey.msgId, 
									 INDEX_MSGID, 
									 deployedPostitApplyP->scope,
									 &postitList);
			break;

		case MATCH_TYPE_CORRELID_CORRELID:
			if (nullCorrelId)
			{
		        TRACE_MID(TRACE_INFO, ("No match on null CorrelId"));
				break;
			}
	        TRACE_MID(TRACE_INFO, ("Using CorrelId value to search CorrelId index"));
			numMatches = searchIndex(deployedPostitApplyP->postitName,
				                     matchKey.correlId, 
									 INDEX_CORRELID, 
									 deployedPostitApplyP->scope,
									 &postitList);
			break;

		case MATCH_TYPE_MSGID_CORRELID:
			if (nullMsgId)
			{
		        TRACE_MID(TRACE_INFO, ("No match on null MsgId"));
				break;
			}
	        TRACE_MID(TRACE_INFO, ("Using MsgId value to search CorrelId index"));
			numMatches = searchIndex(deployedPostitApplyP->postitName,
				                     matchKey.msgId, 
									 INDEX_CORRELID, 
									 deployedPostitApplyP->scope,
									 &postitList);
			break;

		case MATCH_TYPE_BOTH:
			if (nullMsgId && nullCorrelId)
			{
		        TRACE_MID(TRACE_INFO, ("No match on null MsgId AND null CorrelId"));
				break;
			}
	        TRACE_MID(TRACE_INFO, ("Using MsgId+CorrelId value to search Both index"));
			numMatches = searchIndex(deployedPostitApplyP->postitName,
				                     (char *)&matchKey, 
									 INDEX_BOTH, 
									 deployedPostitApplyP->scope,
									 &postitList);
			break;

		case MATCH_TYPE_ALL:
	        TRACE_MID(TRACE_INFO, ("Matching on all messages"));
			numMatches = searchIndex(deployedPostitApplyP->postitName,
				                     NULL, 
									 INDEX_ALL, 
									 deployedPostitApplyP->scope,
									 &postitList);
			break;

		default:
			// should never get here...
	        TRACE_MID(TRACE_ERROR, ("Deployed postit match value is invalid (%d)"));
			break;
	}

	if (!numMatches)
	{
		TRACE_MID(TRACE_INFO, ("No matching postit found"));

		TRACE_MID(TRACE_INFO, ("About to unlock postitChainMutex"));
		threadMutexUnlock(postitChainMutex);

		/* propagate the message to the noMatch terminal */
		TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'noMatch' "
							   "terminal", message));
		retCode = propagateMessage(contextP,
								   C.TERM_NO_MATCH,
								   message,
								   destinationList,
								   exceptionList,
								   FALSE); /* this is not a new message */

		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'noMatch' terminal"));
			RETHROW_IF_EXCEPTION(retCode);
			TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
		}

		TRACE_RETURN(RSN_019, TT_SUCCESS);
	}

	/* create a copy of message so we can add postit data */
	msgContextP = cniGetMessageContext(&rc, message);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to get message context "
			                    "(cniGetMessageContext rc=%d)", rc));
		TRACE_RETURN(RSN_014, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}
	TRACE_MID(TRACE_INFO, ("Message context got at %.8x", msgContextP));

    newMsg = cniCreateMessage(&rc, msgContextP);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to create new message "
			                    "(cniCreateMessage rc=%d)", rc));
		TRACE_RETURN(RSN_015, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}
	TRACE_MID(TRACE_INFO, ("New message created at %.8x", newMsg));

	rc = copyMessage(newMsg, message);
	if (rc != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to copy old message to new message"));
		TRACE_MID(TRACE_INFO, ("About to unlock postitChainMutex"));
		threadMutexUnlock(postitChainMutex);
		TRACE_RETURN(RSN_034, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}

	/* locate (and create if necessary) specified output element */
    elementType	= ELEMENT_TYPE_NAME;  // assume that data will be stored under a NAME element
    copyToElement = locateElementInMessage(newMsg,
		                            deployedPostitApplyP->copyToLocation,
									TRUE,  /* create new element(s) if required */  
									&elementType,
									NULL);

	/* 
	 * Propagate message to the 'failure' terminal if we can't find/create the element
	 */
    if (!copyToElement)
	{
		TRACE_MID(TRACE_INFO, ("About to unlock postitChainMutex"));
		threadMutexUnlock(postitChainMutex);

		UCsprintf(msgBuffer, "Failed to locate specified syntax element '%S' in message",
							deployedPostitApplyP->copyToLocation);

        cciLog(&rc, 
		       CCI_LOG_ERROR, 
			   __FILE__, 
			   __LINE__, 
			   functionName, 
			   C.PLUGIN_MESSAGE_CATALOGUE, 
			   MJB_FAILED_TO_LOCATE_SYNTAX_ELEMENT, 
			   msgBuffer,
#if !defined(WIN32)
		       brokerName,
#endif
 			   /* convert copyToLocation to multi byte string */
			   UCToStr(deployedPostitApplyP->copyToLocation,
					   mbString,
					   sizeof(mbString),
					   NULL),
			   (char *)0);

		/* delete the new message */
		cniDeleteMessage(&rc,
						 newMsg);
		
		TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'failure' "
			                   "terminal", message));
		retCode = propagateMessage(contextP,
								   C.TERM_FAILURE,
								   message,
								   destinationList,
								   exceptionList,
								   FALSE); /* this is not a new message */

		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'failure' terminal"));
			RETHROW_IF_EXCEPTION(retCode);
			TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
		}

		TRACE_RETURN(RSN_013, TT_SUCCESS);
	}

	/* run through the list of postits, applying each in turn to the new message */
	TRACE_MID(TRACE_INFO, ("About to process %d postits", numMatches));
	for (i=0; i<numMatches; i++)
	{
		currentInstanceP = *(postitList+i);
		TRACE_MID(TRACE_INFO, ("Processing postit at %.8x", currentInstanceP));

		/* apply postit data to current element */
		retCode = applyDataToElement(copyToElement, currentInstanceP);
		if (retCode != TT_SUCCESS)
		{
			/* handle error here... */
			UCsprintf(msgBuffer, "Failed to inflate serialised data at syntax element '%S' in message",
								deployedPostitApplyP->copyToLocation);

			cciLog(&rc, 
				   CCI_LOG_ERROR, 
				   __FILE__, 
				   __LINE__, 
				   functionName, 
				   C.PLUGIN_MESSAGE_CATALOGUE, 
				   MJB_FAILED_TO_INFLATE_SERIALISED_SYNTAX_ELEMENT_DATA, 
				   msgBuffer,
#if !defined(WIN32)
				   brokerName,
#endif
 				   /* convert copyToLocation to multi byte string */
				   UCToStr(deployedPostitApplyP->copyToLocation,
						   mbString,
						   sizeof(mbString),
						   NULL),
				   (char *)0);

			/* delete the new message */
			cniDeleteMessage(&rc,
							 newMsg);
			
			TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'failure' "
								   "terminal", message));
			retCode = propagateMessage(contextP,
									   C.TERM_FAILURE,
									   message,
									   destinationList,
									   exceptionList,
									   FALSE); /* this is not a new message */

			if (retCode != TT_SUCCESS)
			{
				TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'failure' terminal"));
				RETHROW_IF_EXCEPTION(retCode);
				TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
			}

			TRACE_RETURN(RSN_013, TT_SUCCESS);
		}

		if (currentInstanceP->global)
		{
			/* 
			 * This is a local copy of a global postit.  All storage is contiguous so
			 * we can simply free it up at this stage
			 */
			TRACE_MID(TRACE_INFO, ("About to free local copy of global postit at %.8x",
				                    currentInstanceP));
			free(currentInstanceP);
			continue;
		}

		if (currentInstanceP->deployedPostitCreateP->numApplications)
		{
			currentInstanceP->numTimesApplied++;
			TRACE_MID(TRACE_INFO, ("numTimesApplied incremented to %d", 
				                    currentInstanceP->numTimesApplied));

			/* check if this was the last application */
			if (currentInstanceP->numTimesApplied ==
				currentInstanceP->deployedPostitCreateP->numApplications)
			{
				TRACE_MID(TRACE_INFO, ("This is the last application, removing postit"));
                freePostit(currentInstanceP);
			}
		}
	}

	TRACE_MID(TRACE_INFO, ("Freeing postit list at %.8x", postitList));
	free(postitList);

	TRACE_MID(TRACE_INFO, ("About to unlock postitChainMutex"));
	threadMutexUnlock(postitChainMutex);

	/* propagate message */
	TRACE_MID(TRACE_INFO, ("Propagating message to 'out' terminal"));
	retCode = propagateMessage(contextP,
							   C.TERM_OUT,
							   newMsg,
							   destinationList,
							   exceptionList,
						       TRUE); /* this is a new message */

	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'out' terminal"));
		RETHROW_IF_EXCEPTION(retCode);
		TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
	}

	/* delete the new message */
    cniDeleteMessage(&rc,
		             newMsg);


	TRACE_RETURN(RSN_019, TT_SUCCESS);
}

/******************************************************************************/
/* removeMsgIdIndex                                                           */
/******************************************************************************/
void removeMsgIdIndex(IN msgIdIndexElement_t *msgIndexP)
{

    TRACE_ENTRY("removeMsgIdIndex", ("msgIndexP=%.8x", msgIndexP));

	/* remove entry from message id index */
	REMOVE_FROM_LINKED_LIST(msgIndexP);

	free(msgIndexP);
    TRACE_MID(TRACE_INFO, ("STORAGE(27): free  %.8x", msgIndexP));

	TRACE_RETURN_VOID(RSN_013);
}

/******************************************************************************/
/* searchIndex                                                                */
/******************************************************************************/
int searchIndex(IN CciChar *postitName,
				IN char *key,
				IN indexType_t indexType,
				IN scope_t scope,
				OUT postitCreateInstance_t ***matchListPP)
{
	int numMatches = 0,
	    keyOffset,
		keyLen,
		numSlots = 0,
		nextSlot = 0,
		i;
	bool_t firstTimeThrough,
		   freeInstance,
		   makeCopy,
		   removeFromGlobalChains,
		   nextInstanceSet;
	postitCreateInstance_t **localMatchListP = NULL,
	                       *currentInstanceP,
						   *nextInstanceP = NULL;
	globalPostitCreateInstance_t *globalInstanceP,
		                         *currentGlobalInstanceP,
								 *nextGlobalInstanceP,
		                         *localInstanceCopyP,
								 *deadAnchorP;
	scope_t currentScope;

    TRACE_ENTRY("searchIndex", ("postitName=%S, key=%.8x, indexType=%s, scope=%s", 
	                         postitName ? postitName : C.NULL_LITERAL,
							 key, 
							 indexType==INDEX_MSGID ? "INDEX_MSGID" :
							 indexType==INDEX_CORRELID ? "INDEX_CORRELID" :
	                         indexType==INDEX_ALL ? "INDEX_ALL" : "INDEX_BOTH",
								 scope == SCOPE_EXECUTION_GROUP ? "SCOPE_EXECUTION_GROUP" :
								 scope == SCOPE_BROKER ? "SCOPE_BROKER" : "SCOPE_PLATFORM"));

	/* on entry, assume caller has locked the postit chain mutex */

	/* we always search the local index first for matches */

    if (indexType == INDEX_MSGID)
	{
		currentInstanceP = msgIdIndexAnchorP->nextMsgIdP;
		keyOffset = 0;
		keyLen = MQ_MSG_ID_LENGTH;
	}
	else
    if (indexType == INDEX_CORRELID)
	{
		currentInstanceP = correlIdIndexAnchorP->nextCorrelIdP;
		keyOffset = MQ_MSG_ID_LENGTH;
		keyLen = MQ_CORREL_ID_LENGTH;
	}
	else
    if (indexType == INDEX_BOTH)
	{
		currentInstanceP = bothIndexAnchorP->nextBothP;
		keyOffset = 0;
		keyLen = MQ_MSG_ID_LENGTH + MQ_CORREL_ID_LENGTH;
	}
	else  // match on all messages - we arbitrarily use the MessageId index
	{
		currentInstanceP = msgIdIndexAnchorP->nextMsgIdP; 
		keyOffset = 0;
		keyLen = 0;  // no need to match key, so length is set to zero
	}

	/* search postit instance chain for matching key */
	firstTimeThrough = TRUE;
	nextInstanceSet = FALSE;
	while(currentInstanceP)
	{
		if (!firstTimeThrough)
		{
			if (nextInstanceSet)
			{
				currentInstanceP = nextInstanceP;
				nextInstanceSet = FALSE;
			}
			else
			/* check next in list */
			if (indexType == INDEX_MSGID || indexType == INDEX_ALL)
			{
				currentInstanceP = currentInstanceP->nextMsgIdP;
			}
			else
			if (indexType == INDEX_CORRELID)
			{
				currentInstanceP = currentInstanceP->nextCorrelIdP;
			}
			else
			{
				currentInstanceP = currentInstanceP->nextBothP;
			}
		}
		else
		{
			firstTimeThrough = FALSE;
		}

		if (!currentInstanceP)
		{
			continue;
		}

		TRACE_MID(TRACE_INFO, ("Processing index element at %.8x", currentInstanceP));
		TRACE_MID(TRACE_INFO, ("Processing %s element as local",
 			                    currentInstanceP->global ? "global" : "local"));
		if (!memcmp(currentInstanceP->msgId+keyOffset,
			        key,
			        keyLen))
		{
			/* if postit name was specified, check for match */
			if (postitName)
			{
				// If we find a global postit in the local index, then its contents can
				// be accessed in the same manner as a local postit.  In other words, simple
				// pointers may be used instead of memory handles.
				if (!currentInstanceP->deployedPostitCreateP->postitName ||
					UCcmp(postitName, currentInstanceP->deployedPostitCreateP->postitName))
				{
					continue;
				}
				/* check for null name in postit */
			}
			TRACE_MID(TRACE_INFO, ("Found a %smatching %spostit at %.8x",
									currentInstanceP->global ? "possible " : "",
									currentInstanceP->global ? "global " : "",
									currentInstanceP));

			/* 
			 * If it's a global entry, then we need to recheck it under lock.  This is
			 * because another execution group may have caused the entry's numApplications 
			 * count to reach its threshold.
			 */
			freeInstance = FALSE;
			makeCopy = TRUE;
			if (currentInstanceP->global)
			{
				/* lock the shared memory */
				shmLock(scope==SCOPE_BROKER);

				globalInstanceP = (globalPostitCreateInstance_t *)currentInstanceP;

				if (globalInstanceP->numApplications)
				{

					if (globalInstanceP->numApplications == globalInstanceP->numTimesApplied)
					{
						TRACE_MID(TRACE_INFO, ("Number of applications exceeded for global "
											   "postit at %.8x - removing it from global "
											   "storage", globalInstanceP));
						freeInstance = TRUE;
						makeCopy = FALSE;
					}
					else
					{
						/* increment number of applications and recheck the threshold */
						globalInstanceP->numTimesApplied++;

						if (globalInstanceP->numApplications == globalInstanceP->numTimesApplied)
						{
							TRACE_MID(TRACE_INFO, ("Number of applications has reached the threshold for global "
												   "postit at %.8x - removing it from global "
												   "storage", globalInstanceP));
							freeInstance = TRUE;
						}
					}

				}

				/* make a copy of the global postit for local use, if reqd */
				if (makeCopy)
				{
					localInstanceCopyP = copyFromGlobalToLocal(globalInstanceP);
					TRACE_MID(TRACE_INFO, ("Local copy of postit at %.8x", localInstanceCopyP));
				}

				/* free the global postit if reqd */
				if (freeInstance)
				{
					/* address next in list */
					if (indexType == INDEX_MSGID || indexType == INDEX_ALL)
					{
						nextInstanceP = currentInstanceP->nextMsgIdP;
					}
					else
					if (indexType == INDEX_CORRELID)
					{
						nextInstanceP = currentInstanceP->nextCorrelIdP;
					}
					else
					{
						nextInstanceP = currentInstanceP->nextBothP;
					}
					nextInstanceSet = TRUE;

					/* remove from local and global chains and free postit */
					TRACE_MID(TRACE_INFO, ("Free'ing global postit at %.8x", globalInstanceP));

					// if postit is on the dead chain, its global chains will already have been
					// removed
					freePostit(currentInstanceP);
				}

				/* unlock the shared memory */
				shmUnlock(scope==SCOPE_BROKER);

				if (!makeCopy)
				{
					continue;
				}

				/* continue processing with the local copy */
				currentInstanceP = (postitCreateInstance_t *)localInstanceCopyP;
			}

			/* add address of this instance to the match list */
			if (!localMatchListP || nextSlot == numSlots)
			{
				TRACE_MID(TRACE_INFO, ("Increasing number of slots by 10"));
				numSlots += 10;
				localMatchListP = (postitCreateInstance_t **)realloc(localMatchListP,
					              numSlots * sizeof(postitCreateInstance_t *));
				TRACE_MID(TRACE_INFO, ("Match table allocated at %.8x", localMatchListP));
			}

			*(localMatchListP + nextSlot) = currentInstanceP;
			nextSlot++;
			numMatches++;
		}
	}

	/* If only local indexes were being searched, then return list to caller */
	if (scope == SCOPE_EXECUTION_GROUP)
	{
		/* return the match list */
		*matchListPP = localMatchListP;

		TRACE_RETURN_TYPE(RSN_013, int, numMatches);
	}

	/**********************************************************************/
	/**********************************************************************/
	/* scope is global - we also need to search the global chain(s)       */
	/**********************************************************************/
	/**********************************************************************/
	for (i=0; i<2; i++)
	{
		// first time through, search the Broker scope chain, if attached
		if (!i)
		{
			// if this is Platform scope, then Broker shared memory may not be attached
			if (scope == SCOPE_PLATFORM && !sharedMemoryCBP_Broker) {
			    TRACE_MID(TRACE_INFO, ("Broker scope shared memory not attached"));
				continue;
			}

			TRACE_MID(TRACE_INFO, ("Searching Broker scope"));
			currentScope = SCOPE_BROKER;
		}
		else
		{
			// check if Platform scope chain needs to be searched also
			if (scope != SCOPE_PLATFORM)
			{
				break;
			}
			TRACE_MID(TRACE_INFO, ("Searching Platform scope"));
			currentScope = SCOPE_PLATFORM;
		}

		// take the appropriate lock
		shmLock(currentScope==SCOPE_BROKER); 

		if (indexType == INDEX_MSGID || indexType == INDEX_ALL)
		{
			currentGlobalInstanceP = 
				SHM_HANDLE2ADDRESS((currentScope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor.nextGlobalMsgIdP,
								   globalPostitCreateInstance_t *,
								   currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			keyOffset = 0;
			keyLen = MQ_MSG_ID_LENGTH;
		}
		else
		if (indexType == INDEX_CORRELID)
		{
			currentGlobalInstanceP = 
				SHM_HANDLE2ADDRESS((currentScope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor.nextGlobalCorrelIdP,
								   globalPostitCreateInstance_t *,
								   currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			keyOffset = MQ_MSG_ID_LENGTH;
			keyLen = MQ_CORREL_ID_LENGTH;
		}
		else
		if (indexType == INDEX_BOTH)
		{
			currentGlobalInstanceP = 
				SHM_HANDLE2ADDRESS((currentScope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor.nextGlobalBothP,
								   globalPostitCreateInstance_t *,
								   currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			keyOffset = 0;
			keyLen = MQ_MSG_ID_LENGTH + MQ_CORREL_ID_LENGTH;
		}
		else  // match on all messages - we arbitrarily use the MessageId index
		{
			currentGlobalInstanceP = 
				SHM_HANDLE2ADDRESS((currentScope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor.nextGlobalMsgIdP,
								   globalPostitCreateInstance_t *,
								   currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			keyOffset = 0;
			keyLen = 0;
		}

		/* search postit instance chain for matching key */
		firstTimeThrough = TRUE;
		nextInstanceSet = FALSE;
		while(currentGlobalInstanceP)
		{
			if (!firstTimeThrough)
			{
				if (nextInstanceSet)
				{
					currentGlobalInstanceP = nextGlobalInstanceP;
					nextInstanceSet = FALSE;
				}
				else
				/* check next in list */
				if (indexType == INDEX_MSGID || indexType == INDEX_ALL)
				{
					currentGlobalInstanceP = 
						SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalMsgIdP,
										   globalPostitCreateInstance_t *,
										   currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
				}
				else
				if (indexType == INDEX_CORRELID)
				{
					currentGlobalInstanceP = 
						SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalCorrelIdP,
										   globalPostitCreateInstance_t *,
										   currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
				}
				else
				{
					currentGlobalInstanceP = 
						SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalBothP,
										   globalPostitCreateInstance_t *,
										   currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
				}
			}
			else
			{
				firstTimeThrough = FALSE;
			}

			if (!currentGlobalInstanceP)
			{
				continue;
			}

			TRACE_MID(TRACE_INFO, ("Processing index element at %.8x", currentGlobalInstanceP));

			/* if the owning process is us, then we'll already have seen 
			 * this postit during the local index search - just ignore it
			 */
			if (currentGlobalInstanceP->executionGroupPID == currentPID)
			{
				TRACE_MID(TRACE_INFO, ("Ignoring postit already processed during local search"));
				continue;
			}

			removeFromGlobalChains = FALSE;

			if (!memcmp(currentGlobalInstanceP->msgId+keyOffset,
				        key,
				        keyLen))
			{
				/* if postit name was specified, check for match */
				if (postitName)
				{
					if (!currentGlobalInstanceP->postitName ||
						UCcmp(postitName, 
							  SHM_HANDLE2ADDRESS(currentGlobalInstanceP->postitName, CciChar *,
												 currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)))
					{
						continue;
					}
				}
				TRACE_MID(TRACE_INFO, ("Found a possible matching global postit at %.8x",
										currentGlobalInstanceP));

				/* 
				 * Check the numApplications threshold has not been exceeded
				 */
				if (currentGlobalInstanceP->numApplications)
				{
					if (currentGlobalInstanceP->numApplications == currentGlobalInstanceP->numTimesApplied)
					{
						TRACE_MID(TRACE_INFO, ("Number of applications exceeded for global "
											   "postit at %.8x - ignoring it",
											   currentGlobalInstanceP));
						continue;
					}

					/* increment number of applications and recheck the threshold */
					currentGlobalInstanceP->numTimesApplied++;

					if (currentGlobalInstanceP->numApplications == currentGlobalInstanceP->numTimesApplied)
					{
						/* 
						 * We don't own this entry so all we can do is remove it from the 
						 * global chain so other processes do not keep bumping into it and add it
						 * to the deadInstanceAnchor.
						 */
						TRACE_MID(TRACE_INFO, ("Number of applications has reached the threshold for global "
											   "postit at %.8x - moving it to dead chain", 
											   currentGlobalInstanceP));

						/* remove from global chains */
						removeFromGlobalChains = TRUE;
					}
				}

				/* make a copy of the global postit for local use */
				localInstanceCopyP = copyFromGlobalToLocal(currentGlobalInstanceP);
				TRACE_MID(TRACE_INFO, ("Local copy of postit at %.8x", localInstanceCopyP));

				if (removeFromGlobalChains)
				{
					/* address next global instance before we unchain current one */
					if (indexType == INDEX_MSGID || indexType == INDEX_ALL)
					{
						nextGlobalInstanceP = SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalMsgIdP, globalPostitCreateInstance_t *,
												 currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
					}
					else
					if (indexType == INDEX_CORRELID)
					{
						nextGlobalInstanceP = SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalCorrelIdP, globalPostitCreateInstance_t *,
												 currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
					}
					else
					{
						nextGlobalInstanceP = SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalBothP, globalPostitCreateInstance_t *,
												 currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
					}
					nextInstanceSet = TRUE;
					
					unchainPostit((postitCreateInstance_t *)currentGlobalInstanceP); /* global chains */

					/* add to deadInstanceAnchor */
					deadAnchorP = &(currentScope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalDeadInstanceAnchor;

					INSERT_AFTER_IN_GLOBAL_POSTIT_CHAIN(deadAnchorP, currentGlobalInstanceP, currentScope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
					currentGlobalInstanceP->onDeadChain = TRUE;
				    TRACE_MID(TRACE_INFO, ("Postit moved to dead chain")); 
				}

				/* add address of this local copy instance to the match list */
				if (!localMatchListP || nextSlot == numSlots)
				{
					TRACE_MID(TRACE_INFO, ("Increasing number of slots by 10"));
					numSlots += 10;
					localMatchListP = (postitCreateInstance_t **)realloc(localMatchListP,
									  numSlots * sizeof(postitCreateInstance_t *));
					TRACE_MID(TRACE_INFO, ("Match table allocated at %.8x", localMatchListP));
				}

				*(localMatchListP + nextSlot) = (postitCreateInstance_t *)localInstanceCopyP;
				nextSlot++;
				numMatches++;
			}
		}

		/* unlock shared memory */
		shmUnlock(currentScope==SCOPE_BROKER);
	}

	/* return the match list */
	*matchListPP = localMatchListP;

	TRACE_RETURN_TYPE(RSN_013, int, numMatches);
}

/******************************************************************************/
/* freePostit                                                                 */
/******************************************************************************/
void freePostit(IN postitCreateInstance_t *postitInstanceP)
{
	int rc;
    static char *functionName="freePostit";
    TRACE_ENTRY("freePostit", ("postitInstanceP=%.8x", postitInstanceP));

	/* remove postit from all chains */
	unchainPostit(postitInstanceP);  /* local chains */

	if (postitInstanceP->global)
	{
		rc = shmFree(postitInstanceP, TRUE, postitInstanceP->scope==SCOPE_BROKER);
        CHECK_SHMFREE_ERROR(rc, 
	                        functionName, 
				            platformIsUnix);  
		TRACE_MID(TRACE_INFO, ("global postitInstanceP free'd"));
	}
	else
	{
		/* free up the instance control block */
		if (postitInstanceP->savedDataP)
		{
			TRACE_MID(TRACE_INFO, ("About to free savedDataP"));
			free(postitInstanceP->savedDataP);
		}

		free(postitInstanceP);
		TRACE_MID(TRACE_INFO, ("postitInstanceP free'd"));
	}

    TRACE_RETURN_VOID(RSN_001);
}

/******************************************************************************/
/* unchainPostit                                                              */
/******************************************************************************/
void unchainPostit(IN postitCreateInstance_t *postitInstanceP)
{
	globalPostitCreateInstance_t *globalPostitInstanceP;
	bool_t postitOwnedByProcess = TRUE; 
    TRACE_ENTRY("unchainPostit", ("postitInstanceP=%.8x", 
		                          postitInstanceP));

	/* assume caller has appropriate locks held on entry... */

    // This function removes a postit from the local chains, if called by the owning
	// process.  If the postit is global, it then removes the postit from the global
	// chains. (MsgId, CorrelId and Both)

	// if this is a GLOBAL postit and we don't own it, then don't touch local indices
	if (postitInstanceP->scope != SCOPE_EXECUTION_GROUP) {
 	    globalPostitInstanceP = (globalPostitCreateInstance_t *)postitInstanceP;
		if (globalPostitInstanceP->executionGroupPID != currentPID) {
			postitOwnedByProcess = FALSE;
	        TRACE_MID(TRACE_INFO, ("Postit owned by another process (%d) - "
				                   "local chains not updated", 
								   globalPostitInstanceP->executionGroupPID));
		}
	}

    if (postitOwnedByProcess) {
		// always remove from local chains
		/* unchain postit instance */
		REMOVE_FROM_LINKED_LIST(postitInstanceP);
		TRACE_MID(TRACE_INFO, ("Removed from postit instance chain"));

		/* only remove from alarm chain if timeout value was specified */
		if (postitInstanceP->deployedPostitCreateP->timeOutInterval != INFINITE)
		{
			TRACE_MID(TRACE_INFO, ("About to remove from alarm chain"));
			REMOVE_FROM_ALARM_CHAIN(postitInstanceP);
			TRACE_MID(TRACE_INFO, ("Removed from alarm chain"));
		}

		REMOVE_FROM_MSG_ID_CHAIN(postitInstanceP);
		TRACE_MID(TRACE_INFO, ("Removed from MsgId chain"));

		REMOVE_FROM_CORREL_ID_CHAIN(postitInstanceP);
		TRACE_MID(TRACE_INFO, ("Removed from CorrelId chain"));

		REMOVE_FROM_BOTH_CHAIN(postitInstanceP);
		TRACE_MID(TRACE_INFO, ("Removed from Both chain"));
	}

	if (postitInstanceP->scope == SCOPE_EXECUTION_GROUP)
	{
		// nothing more to do, so return to caller
	    TRACE_RETURN_VOID(RSN_001);
	}

	// remove from appropriate global chain
	globalPostitInstanceP = (globalPostitCreateInstance_t *)postitInstanceP;

	// If postit is on the dead chain, then it will already have been removed from the global
	// chains.  We only need to remove it from the dead chain.
 	if (globalPostitInstanceP->onDeadChain) {
		TRACE_MID(TRACE_INFO, ("Postit is on dead chain."));
        REMOVE_FROM_GLOBAL_POSTIT_CHAIN(globalPostitInstanceP, 
		    globalPostitInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
        TRACE_RETURN_VOID(RSN_003);
	}

	/* assume caller has got shared memory lock */
	REMOVE_FROM_GLOBAL_POSTIT_MSG_ID_CHAIN(globalPostitInstanceP,
		globalPostitInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("Removed from global MsgId chain"));

	REMOVE_FROM_GLOBAL_POSTIT_CORREL_ID_CHAIN(globalPostitInstanceP,
					globalPostitInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("Removed from global CorrelId chain"));

	REMOVE_FROM_GLOBAL_POSTIT_BOTH_CHAIN(globalPostitInstanceP,
		globalPostitInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("Removed from global Both chain"));

    TRACE_RETURN_VOID(RSN_001);
}

/******************************************************************************/
/* applyDataToElement                                                         */
/******************************************************************************/
int applyDataToElement(IN CciElement *elementP, 
					   IN postitCreateInstance_t *postitP)
{
	char *dataBuffer,
		 *currentPos,
		 *endPos;
	int retCode,
		dataBufferLength,
		depth = 0;
	CciElement *localElementP = elementP;
	bool_t moreDataInBuffer,
		   firstTime = TRUE;
    TRACE_ENTRY("applyDataToElement", ("elementP=%.8x, postitP=%.8x", 
		                                elementP, postitP));

	if (postitP->global)
	{
		TRACE_MID(TRACE_INFO, ("Global postit being processed"));
	}
	else
	{
		TRACE_MID(TRACE_INFO, ("Local postit being processed"));
	}

	dataBuffer = postitP->savedDataP;
	currentPos = dataBuffer;
	dataBufferLength = postitP->savedDataLength;

	if (dataBufferLength)
	{
	    moreDataInBuffer = TRUE;
		endPos = dataBuffer + dataBufferLength;
	}
	else
	{
	    moreDataInBuffer = FALSE;
	}

	while(moreDataInBuffer)
	{

		retCode = inflateSyntaxElement(&localElementP,
									   dataBuffer,
									   dataBufferLength,
									   &currentPos,
									   &depth,
									   firstTime,
									   FALSE); /* don't create element for first saved element */
		if (currentPos == endPos)
		{
			moreDataInBuffer = FALSE;
		}
		firstTime = FALSE;
	}

    TRACE_RETURN(RSN_001, TT_SUCCESS);
}


/* local functions */
/******************************************************************************/
/* copyFromGlobalToLocal                                                      */
/******************************************************************************/
globalPostitCreateInstance_t *copyFromGlobalToLocal(
				                   IN globalPostitCreateInstance_t *globalInstanceP)
{
    globalPostitCreateInstance_t *localInstanceP;
	unsigned long offset;

    TRACE_ENTRY("copyFromGlobalToLocal", ("globalInstanceP=%.8x", globalInstanceP));

	localInstanceP = (globalPostitCreateInstance_t *)malloc(globalInstanceP->instanceLength);
	memcpy(localInstanceP, globalInstanceP, globalInstanceP->instanceLength);

	TRACE_MID(TRACE_INFO, ("Copied global postit at %.8x into local "
						   "storage at %.8x - %d bytes copied",
						   globalInstanceP,
						   localInstanceP,
						   globalInstanceP->instanceLength));

	/* 
	 * convert handles in global data to addresses in the local postit copy 
	 */
	/* executionGroupNameP */
	offset = (unsigned long)(SHM_HANDLE2ADDRESS(
				             globalInstanceP->executionGroupNameP, char *,
							 globalInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)) -
                               (unsigned long)globalInstanceP;
	localInstanceP->executionGroupNameP = (shmemHandle_t)(unsigned long)localInstanceP + offset;
	TRACE_MID(TRACE_INFO, ("executionGroupNameP converted to local address %.8x",
				            localInstanceP->executionGroupNameP));

	/* messageFlowNameP */
	offset = (unsigned long)(SHM_HANDLE2ADDRESS(
				             globalInstanceP->messageFlowNameP, char *,
							 globalInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)) -
             (unsigned long)globalInstanceP;
	localInstanceP->messageFlowNameP = (shmemHandle_t)(unsigned long)localInstanceP + offset;
	TRACE_MID(TRACE_INFO, ("messageFlowNameP converted to local address %.8x",
				            localInstanceP->messageFlowNameP));

	/* savedDataP */
	offset = (unsigned long)(SHM_HANDLE2ADDRESS(
				             globalInstanceP->savedDataP, char *,
							 globalInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)) -
             (unsigned long)globalInstanceP;
	localInstanceP->savedDataP = (shmemHandle_t)(unsigned long)localInstanceP + offset;
	TRACE_MID(TRACE_INFO, ("savedDataP converted to local address %.8x",
				            localInstanceP->savedDataP));

	/* postitName */
	offset = (unsigned long)(SHM_HANDLE2ADDRESS(
				             globalInstanceP->postitName, char *,
							 globalInstanceP->scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)) -
             (unsigned long)globalInstanceP;
	localInstanceP->postitName = (shmemHandle_t)(unsigned long)localInstanceP + offset;
	TRACE_MID(TRACE_INFO, ("postitName converted to local address %.8x",
				            localInstanceP->postitName));

	TRACE_RETURN_TYPE(RSN_001, globalPostitCreateInstance_t *, localInstanceP);
}


