/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : create.c                                                 */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : Creates a postit by saving data from the specified       */
/*                   location within the current message.                     */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: June 2001                      */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include <postit.h>      /* Specific header for this dll                      */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <BipCni.h>		/* Function prototypes for Plug-in Node functions     */


/******************************************************************************/
/* Node Utility Functions                                                     */
/*                                                                            */
/* insertAttributeTableEntry                                                  */
/* insertInputTerminalListEntry                                               */
/* insertOutputTerminalListEntry                                              */
/* getInputTerminalHandle                                                     */
/* getOutputTerminalHandle                                                    */
/* traverseMessage                                                                           */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/* evaluatePostitCreate                                                       */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  destinationList (input) - address of the input destination list for the   */
/*                            message                                         */
/*  exceptionList   (input) - address of the exception list for the message   */
/*  message         (input) - address of the input message object             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Extracts contents of specified node in syntax tree and creates a postit    */
/* instance.  Also resets the alarm if a timeout value was specified.         */
/*                                                                            */
/******************************************************************************/
int evaluatePostitCreate(context_t* contextP,
 			   		     CciMessage *destinationList,
		                 CciMessage *exceptionList,
		                 CciMessage *message)
{
	void *correlIdElementP,
		 *messageIdElementP;
	postitCreateInstance_t *postitInstanceP,
		                   *instanceAnchorP;
	globalPostitCreateInstance_t *globalPostitInstanceP;
    struct CciByteArray byteArrayDetails;
    CciSize numBytes;
	int rc,
		retCode,
		length,
		allocSize;
	deployedPostitCreate_t *deployedPostitCreateP;
	CciElement *copyFromElement;
	char *serialisedBuffer,
		 mbString[MAX_STRING_SIZE + 1],
		 msgBuffer[1000];
    char *executionGroupNameP;
	CciChar *messageFlowNameP,
		    *postitNameP;
	void *dataP;
	scope_t scope;
    static char *functionName="evaluatePostitCreate";

    TRACE_ENTRY("evaluatePostitCreate", ("contextP=%.8x, destinationList=%.8x, "
		                     "exceptionList=%.8x, message=%.8x",
							 contextP, destinationList,
							 exceptionList, message));

	deployedPostitCreateP = (deployedPostitCreate_t *)(contextP->nodeDataP);

	scope = deployedPostitCreateP->scope;

	/* Create a new local postit instance.  Even if its a global postit, we build a local
	 * copy first since this makes pointer manipulation somewhat easier.  For global postits
	 * we allocate global storage later and then transfer data from the local copy.
	 */
	postitInstanceP = (postitCreateInstance_t *)malloc(sizeof(postitCreateInstance_t));

	memset(postitInstanceP, 0, sizeof(postitCreateInstance_t));
	TRACE_MID(TRACE_INFO, ("New %s postit instance created at %.8x", 
 		                   scope==SCOPE_EXECUTION_GROUP ? "LOCAL" : "GLOBAL",
		                   postitInstanceP));

	/* extract the message id and correl id values */
	messageIdElementP = getMessageElement(message, MESSAGE_ID);

	if (!messageIdElementP)
	{
		free(postitInstanceP);
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Message Id not found in message"));
		TRACE_RETURN(RSN_014, TT_MESSAGE_ID_NOT_FOUND);
	}

	/* copy message id, assume it is of type CCI_VALUE_TYPE_BYTEARRAY */
	byteArrayDetails.pointer = (void *)(postitInstanceP->msgId);
	byteArrayDetails.size = sizeof(postitInstanceP->msgId);
	numBytes = cniElementByteArrayValue(&rc,
										messageIdElementP,
										&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		free(postitInstanceP);
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract Message Id from message"));
		TRACE_RETURN(RSN_015, TT_MESSAGE_ID_NOT_FOUND);
	}

	TRACE_MID(TRACE_INFO, ("Incoming Message Id value is..."));
	TRACE_MIDHEX(TRACE_INFO, postitInstanceP->msgId, MQ_MSG_ID_LENGTH);

	/* extract the correlId */
	correlIdElementP = getMessageElement(message, CORREL_ID);

	if (!correlIdElementP)
	{
		free(postitInstanceP);
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Correlation Id not found in message"));
		TRACE_RETURN(RSN_014, TT_CORRELID_NOT_FOUND);
	}

	/* copy correl id, assume it is of type CCI_VALUE_TYPE_BYTEARRAY */
	byteArrayDetails.pointer = (void *)(postitInstanceP->correlId);
	byteArrayDetails.size = sizeof(postitInstanceP->correlId);
	numBytes = cniElementByteArrayValue(&rc,
										correlIdElementP,
										&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		free(postitInstanceP);
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract Correl Id from message"));
		TRACE_RETURN(RSN_015, TT_CORRELID_NOT_FOUND);
	}

	TRACE_MID(TRACE_INFO, ("Incoming Correl Id value is..."));
	TRACE_MIDHEX(TRACE_INFO, postitInstanceP->correlId, MQ_CORREL_ID_LENGTH);

	/* initialise other postit create attributes */
	postitInstanceP->deployedPostitCreateP = contextP->nodeDataP;
	postitInstanceP->scope = postitInstanceP->deployedPostitCreateP->scope;

	/* locate the specified element in the syntax tree and copy the data */
	copyFromElement = locateElementInMessage(message,
		                            postitInstanceP->deployedPostitCreateP->copyFromLocation,
									FALSE,   /* don't create element */
									NULL,
									NULL);

	/* 
	 * Propagate message to the 'noDataCopied' terminal if element doesn't exist or 
	 * has no data.
	 */
    if (!copyFromElement)
	{
		UCsprintf(msgBuffer, "Failed to locate specified syntax element '%S' in message",
							postitInstanceP->deployedPostitCreateP->copyFromLocation);

        cciLog(&rc, 
		       CCI_LOG_ERROR, 
			   __FILE__, 
			   __LINE__, 
			   functionName, 
			   C.PLUGIN_MESSAGE_CATALOGUE, 
			   MJB_FAILED_TO_LOCATE_SYNTAX_ELEMENT, 
			   msgBuffer,
#if !defined(WIN32)
		       brokerName,
#endif
 			   /* convert copyFromLocation to multi byte string */
			   UCToStr(postitInstanceP->deployedPostitCreateP->copyFromLocation,
					   mbString,
					   sizeof(mbString),
					   NULL),
			   (char *)0);

		free(postitInstanceP);

		TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'noDataCopied' "
			                   "terminal", message));
		retCode = propagateMessage(contextP,
								   C.TERM_NO_DATA_COPIED,
								   message,
								   destinationList,
								   exceptionList,
								   FALSE); /* this is not a new message */

		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'noDataCopied' terminal"));
			RETHROW_IF_EXCEPTION(retCode);
			TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
		}

		TRACE_RETURN(RSN_013, TT_SUCCESS);
	}

	/* make a copy of the data (and elements) in/under this element */
	retCode = serialiseSyntaxElements(copyFromElement, 
		                              FALSE, 
									  FALSE,
									  &serialisedBuffer, 
									  &length);

	/* propagate to failure terminal if any errors */
	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_INFO, ("About to issue error message"));
		UCsprintf(msgBuffer, "Failed to serialise data from syntax element '%S' in message",
							postitInstanceP->deployedPostitCreateP->copyFromLocation);

        cciLog(&rc, 
		       CCI_LOG_ERROR, 
			   __FILE__, 
			   __LINE__, 
			   functionName, 
			   C.PLUGIN_MESSAGE_CATALOGUE, 
			   MJB_FAILED_TO_SERIALISE_SYNTAX_ELEMENT_DATA, 
			   msgBuffer,
#if !defined(WIN32)
		       brokerName,
#endif
 			   /* convert copyFromLocation to multi byte string */
			   UCToStr(postitInstanceP->deployedPostitCreateP->copyFromLocation,
					   mbString,
					   sizeof(mbString),
					   NULL),
			   (char *)0);

		free(postitInstanceP);
		
		TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'failure' "
			                   "terminal", message));
		retCode = propagateMessage(contextP,
								   C.TERM_FAILURE,
								   message,
								   destinationList,
								   exceptionList,
								   FALSE); /* this is not a new message */

		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'failure' terminal"));
			RETHROW_IF_EXCEPTION(retCode);
			TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
		}

		TRACE_RETURN(RSN_013, TT_SUCCESS);
	}

	/* update the postit with the serialised data */
	postitInstanceP->savedDataLength = length;
	if (scope == SCOPE_EXECUTION_GROUP)
	{
		postitInstanceP->savedDataP = serialisedBuffer;
		postitInstanceP->global = FALSE;
	}
	else
	{
		/* 
		 * Allocate a contiguous block of global storage to hold (in order) 
		 * - the global postit instance
		 * - the name of the Execution Group
		 * - the name of the Message Flow
		 * - the serialised data
		 * - the postit name
		 */
		allocSize = sizeof(globalPostitCreateInstance_t) +
			          strlen(executionGroupName) + 1 +
					  (UClen(contextP->messageFlowName) + 1) * sizeof(CciChar) +
                      postitInstanceP->savedDataLength +
			          (postitInstanceP->deployedPostitCreateP->postitName ?
				      (UClen(postitInstanceP->deployedPostitCreateP->postitName) + 1) 
				       * sizeof(CciChar) : 0);

		TRACE_MID(TRACE_INFO, ("About to allocate %d bytes of global storage",
			                   allocSize));

		/* allocate global storage and copy existing buffer into this */
		globalPostitInstanceP = (globalPostitCreateInstance_t *)shmMalloc(allocSize,
			 										                      FALSE,
																		  scope==SCOPE_BROKER,
													                      &rc);
		CHECK_SHMEM_ERROR(rc, 
			              functionName, 
						  scope==SCOPE_BROKER, 
			              globalPostitInstanceP, 
						  globalPostitCreateInstance_t *, 
						  allocSize,
						  FALSE,  /* lock is not held */
						  platformIsUnix,
						  TRUE);  

		/* copy in the contents of the local postit */
		memcpy(globalPostitInstanceP, postitInstanceP, sizeof(postitCreateInstance_t));

		/* append the additional data to global postit instance */
	    executionGroupNameP = (char *)globalPostitInstanceP + sizeof(globalPostitCreateInstance_t);
		TRACE_MID(TRACE_INFO, ("Storage for executionGroupName located at %.8x",
			                   executionGroupNameP));

        messageFlowNameP = (CciChar *)(executionGroupNameP + strlen(executionGroupName) + 1);
		TRACE_MID(TRACE_INFO, ("Storage for messageFlowName located at %.8x",
			                   messageFlowNameP));

		dataP = (char *)messageFlowNameP + 
			               (UClen(contextP->messageFlowName) + 1) * sizeof(CciChar);
		TRACE_MID(TRACE_INFO, ("Storage for serialised data located at %.8x",
			                   dataP));

        if (postitInstanceP->deployedPostitCreateP->postitName)
		{
			postitNameP = (CciChar *)((char *)dataP + postitInstanceP->savedDataLength);
			TRACE_MID(TRACE_INFO, ("Storage for postit name located at %.8x",
								   postitNameP));
		}
		else
		{
			postitNameP = NULL;
			TRACE_MID(TRACE_INFO, ("Postit does not have a name, no global storage required"));
		} 

		strcpy(executionGroupNameP, executionGroupName);
		UCcpy(messageFlowNameP, contextP->messageFlowName);
		memcpy(dataP, serialisedBuffer, postitInstanceP->savedDataLength);
        if (postitInstanceP->deployedPostitCreateP->postitName)
		{
			UCcpy(postitNameP, postitInstanceP->deployedPostitCreateP->postitName);
		}

		/* since this is a global postit, we must use handles instead of pointers */
		globalPostitInstanceP->executionGroupNameP = SHM_ADDRESS2HANDLE(executionGroupNameP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
		globalPostitInstanceP->messageFlowNameP = SHM_ADDRESS2HANDLE(messageFlowNameP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
		globalPostitInstanceP->savedDataP = SHM_ADDRESS2HANDLE(dataP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
		globalPostitInstanceP->postitName = SHM_ADDRESS2HANDLE(postitNameP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

		globalPostitInstanceP->global = TRUE;
		globalPostitInstanceP->executionGroupPID = currentPID;
		globalPostitInstanceP->instanceLength = allocSize;
		globalPostitInstanceP->numApplications = postitInstanceP->deployedPostitCreateP->numApplications;

		free(serialisedBuffer);
		free(postitInstanceP);  /* free postit in local memory */

		/* use same pointer to address either local or global for simplicity */
		postitInstanceP = (postitCreateInstance_t *)globalPostitInstanceP;
	}
	TRACE_MID(TRACE_INFO, ("Postit updated with serialised data at %.8x",
		                    scope==SCOPE_EXECUTION_GROUP ? postitInstanceP->savedDataP : 
	                        SHM_HANDLE2ADDRESS(globalPostitInstanceP->savedDataP, void *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
	TRACE_MIDHEX(TRACE_INFO, scope==SCOPE_EXECUTION_GROUP ? postitInstanceP->savedDataP : 
	                         SHM_HANDLE2ADDRESS(globalPostitInstanceP->savedDataP, void *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform),
								 scope==SCOPE_EXECUTION_GROUP ? postitInstanceP->savedDataLength :
							     globalPostitInstanceP->savedDataLength);

	TRACE_MID(TRACE_INFO, ("About to lock postitChainMutex"));
	threadMutexLock(postitChainMutex);

	/* add the postit to the deployed create node */
	instanceAnchorP = &deployedPostitCreateP->instanceAnchor;
	INSERT_AFTER_IN_LINKED_LIST(instanceAnchorP, postitInstanceP);

	/* set the postit timeout and add to alarm chain */
	if (postitInstanceP->deployedPostitCreateP->timeOutInterval != INFINITE)
	{
	    setPostitTimeoutAlarm(postitInstanceP);
	
		/* signal the alarm thread to process new postit instance */
		threadCondSignal(alarmThreadWorkCondVar);
	}
	else
	{
		TRACE_MID(TRACE_INFO, ("Postit has no timeout value"));
	}

	/* add the postit to required indices and chains */
	updateAllIndexes(postitInstanceP, scope);

	TRACE_MID(TRACE_INFO, ("About to unlock postitChainMutex"));
	threadMutexUnlock(postitChainMutex);

    /* propagate the message to the output terminal */
	TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'out' "
			               "terminal", message));
	retCode = propagateMessage(contextP,
							   C.TERM_OUT,
							   message,
							   destinationList,
							   exceptionList,
							   FALSE); /* this is not a new message */

	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'out' terminal"));
		RETHROW_IF_EXCEPTION(retCode);
		TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
	}

	TRACE_RETURN(RSN_013, TT_SUCCESS);
}

/******************************************************************************/
/* updateAllIndexes                                                           */
/******************************************************************************/
void updateAllIndexes(IN postitCreateInstance_t *newInstanceP,
					  IN scope_t scope)
{
	postitCreateInstance_t *currentInstanceP;
	globalPostitCreateInstance_t *currentGlobalInstanceP,
		                         *newGlobalInstanceP;

    TRACE_ENTRY("updateAllIndexes", ("newInstanceP=%.8x, scope=%s", 
		                             newInstanceP,
									 scope==SCOPE_EXECUTION_GROUP ? "SCOPE_EXECUTION_GROUP" :
									 scope==SCOPE_BROKER ? "SCOPE_BROKER" : "SCOPE_PLATFORM"));

	/* 
	 * Assume caller has postitChainMutex locked on entry... 
	 *
	 * NB: Even though some of the postits on the local chain may in fact reside in shared
	 * storage, we can read these without taking any locks.  This is because they are only
	 * ever deleted by the owning process (and that's us).  Furthermore, we can also 
	 * update the local chain pointers of these postits without a lock since these fields
	 * are only ever referenced by the owning process (again, that's us).
	 */

	/* 
	 * We always update the local chains since the Execution Group will always be in scope.
	 */ 


	/* 
	 * Update the MsgId index
	 */
	currentInstanceP = msgIdIndexAnchorP;

	/* insert postit instance into chain in message id order */
	while(currentInstanceP)
	{
		TRACE_MID(TRACE_INFO, ("Processing MsgId index element at %.8x", currentInstanceP));
		if (!currentInstanceP->nextMsgIdP ||
			memcmp(newInstanceP->msgId, 
				   currentInstanceP->nextMsgIdP->msgId, MQ_MSG_ID_LENGTH) < 0)
		{
			TRACE_MID(TRACE_INFO, ("Inserting new element in MsgId index"));
			break;
		}
		currentInstanceP = currentInstanceP->nextMsgIdP;
	}
	INSERT_AFTER_IN_MSG_ID_CHAIN(currentInstanceP, newInstanceP);

	/* 
	 * Update the CorrelId index
	 */
	currentInstanceP = correlIdIndexAnchorP;

	/* insert postit instance into chain in correl id order */
	while(currentInstanceP)
	{
		TRACE_MID(TRACE_INFO, ("Processing CorrelId index element at %.8x", currentInstanceP));
		if (!currentInstanceP->nextCorrelIdP ||
			memcmp(newInstanceP->correlId, 
				   currentInstanceP->nextCorrelIdP->msgId, MQ_CORREL_ID_LENGTH) < 0)
		{
			TRACE_MID(TRACE_INFO, ("Inserting new element in CorrelId index"));
			break;
		}
		currentInstanceP = currentInstanceP->nextCorrelIdP;
	}
	INSERT_AFTER_IN_CORREL_ID_CHAIN(currentInstanceP, newInstanceP);

	/* 
	 * Update the Both index
	 */
	currentInstanceP = bothIndexAnchorP;

	/* insert postit instance into chain in Msgid+CorrelId order */
	/* Since MsgId and CorrelId are contiguous in storage, we can simply compare 48 bytes */
	while(currentInstanceP)
	{
		TRACE_MID(TRACE_INFO, ("Processing Both index element at %.8x", currentInstanceP));
		if (!currentInstanceP->nextBothP ||
			memcmp(newInstanceP->msgId, 
				   currentInstanceP->nextBothP->msgId, MQ_MSG_ID_LENGTH+MQ_CORREL_ID_LENGTH) < 0)
		{
			TRACE_MID(TRACE_INFO, ("Inserting new element in MsgId+CorrelId index"));
			break;
		}
		currentInstanceP = currentInstanceP->nextBothP;
	}
	INSERT_AFTER_IN_BOTH_CHAIN(currentInstanceP, newInstanceP);

	/**********************************************************************************/
	/**********************************************************************************/
	/* Update global chains                                                           */
	/**********************************************************************************/
	/**********************************************************************************/
	if (scope != SCOPE_EXECUTION_GROUP)
	{
		TRACE_MID(TRACE_INFO, ("Updating global indexes"));

        newGlobalInstanceP = (globalPostitCreateInstance_t *)newInstanceP;

		/* lock shared memory */
		shmLock(newGlobalInstanceP->scope==SCOPE_BROKER);

		/* Since we are chaining global storage, we must use handles as opposed to pointers */
		newGlobalInstanceP->nextGlobalP = SHM_NULL_HANDLE;
		newGlobalInstanceP->prevGlobalP = SHM_NULL_HANDLE;
		newGlobalInstanceP->nextGlobalMsgIdP = SHM_NULL_HANDLE;
		newGlobalInstanceP->prevGlobalMsgIdP = SHM_NULL_HANDLE;
		newGlobalInstanceP->nextGlobalCorrelIdP = SHM_NULL_HANDLE;
		newGlobalInstanceP->prevGlobalCorrelIdP = SHM_NULL_HANDLE;
		newGlobalInstanceP->nextGlobalBothP = SHM_NULL_HANDLE;
		newGlobalInstanceP->prevGlobalBothP = SHM_NULL_HANDLE;

		/* 
		 * Update the global MsgId index
		 */
		currentGlobalInstanceP = &(scope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor;

		/* insert postit instance into chain in message id order */
		while(currentGlobalInstanceP)
		{
			TRACE_MID(TRACE_INFO, ("Processing global MsgId index element at %.8x", currentGlobalInstanceP));
			if (!SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalMsgIdP, void *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform) ||
				memcmp(newGlobalInstanceP->msgId, 
					   SHM_HANDLE2ADDRESS(
					   currentGlobalInstanceP->nextGlobalMsgIdP, globalPostitCreateInstance_t *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)->msgId, 
						  MQ_MSG_ID_LENGTH) < 0)
			{
				TRACE_MID(TRACE_INFO, ("Inserting new element in global MsgId index"));
				break;
			}
			currentGlobalInstanceP = 
				SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalMsgIdP, globalPostitCreateInstance_t *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
		}
		INSERT_AFTER_IN_GLOBAL_POSTIT_MSG_ID_CHAIN(currentGlobalInstanceP, newGlobalInstanceP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

		/* 
		 * Update the CorrelId index
		 */
		currentGlobalInstanceP = &(scope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor;

		/* insert postit instance into chain in correl id order */
		while(currentGlobalInstanceP)
		{
			TRACE_MID(TRACE_INFO, ("Processing global CorrelId index element at %.8x", currentGlobalInstanceP));
			if (!SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalCorrelIdP, void *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform) ||
				memcmp(newGlobalInstanceP->correlId, 
					   SHM_HANDLE2ADDRESS(
					      currentGlobalInstanceP->nextGlobalCorrelIdP, globalPostitCreateInstance_t *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)->correlId, 
						  MQ_MSG_ID_LENGTH) < 0)
			{
				TRACE_MID(TRACE_INFO, ("Inserting new element in global CorrelId index"));
				break;
			}
			currentGlobalInstanceP = 
				SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalCorrelIdP, globalPostitCreateInstance_t *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
		}
		INSERT_AFTER_IN_GLOBAL_POSTIT_CORREL_ID_CHAIN(currentGlobalInstanceP, newGlobalInstanceP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

		/* 
		 * Update the Both index
		 */
		currentGlobalInstanceP = &(scope==SCOPE_BROKER ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor;

		/* insert postit instance into chain in Msgid+CorrelId order */
		/* Since MsgId and CorrelId are contiguous in storage, we can simply compare 48 bytes */
		while(currentGlobalInstanceP)
		{
			TRACE_MID(TRACE_INFO, ("Processing global Both index element at %.8x", currentGlobalInstanceP));

			
			if (!SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalBothP, void *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform) ||
				memcmp(newGlobalInstanceP->msgId, 
					   SHM_HANDLE2ADDRESS(
					      currentGlobalInstanceP->nextGlobalBothP, globalPostitCreateInstance_t *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)->msgId, 
						  MQ_MSG_ID_LENGTH+MQ_CORREL_ID_LENGTH) < 0)
			{
				TRACE_MID(TRACE_INFO, ("Inserting new element in global MsgId+CorrelId index"));
				break;
			}
			currentGlobalInstanceP = 
				SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalBothP, globalPostitCreateInstance_t *, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
		}
		INSERT_AFTER_IN_GLOBAL_POSTIT_BOTH_CHAIN(currentGlobalInstanceP, newGlobalInstanceP, scope==SCOPE_BROKER ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

		/* unlock shared memory */
		shmUnlock(newGlobalInstanceP->scope==SCOPE_BROKER);
	}

	TRACE_RETURN_VOID(RSN_001);
}

