#define GLOBAL_POSTIT_DATA 1
/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : postit.c                                                 */
/* Output name     : postit.dll                                               */
/*                   (NOTE: must be renamed to postit.lil and moved to        */
/*                   <install dir>\bin for use with MQSI V2.0 broker          */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\include\plugin        */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : This pair of plug-in nodes create allow data to be       */
/*                   copied from one message and restored in another.         */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Jun 2001                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include <postit.h> /* Specific header for this dll */                       
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <plugin.h>   /* common plugin node functions */
#include <BipCni.h>	  /* Function prototypes for Plug-in Node functions     */

/* local function prototypes */
int attachSharedMemoryIfRequired(IN context_t *contextP,
								 IN scope_t scope);
executionGroupEntry_t *findExecutionGroupEntry(IN bool_t scopeIsBroker);
void removeEGPostitsFromGlobalStorage(IN executionGroupEntry_t *currentEGP,
									  IN bool_t scopeIsBroker);
executionGroupEntry_t *createNewExecutionGroupEntry(IN bool_t scopeIsBroker);
void removeExecutionGroupFromGlobalStorage();

/******************************************************************************/
/* Node Implementation Functions                                              */
/*                                                                            */
/* These are mandatory and MUST be implemented by the programmer.             */
/*                                                                            */
/* createNodeContext                                                          */
/* deleteNodeContext                                                          */
/* evaluate                                                                   */
/* getAttribute                                                               */
/* getAttributeName                                                           */
/* setAttribute                                                               */
/* bipGetMessageflowNodeFactory                                               */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/* createNodeContext                                                          */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  factoryObject (input) - address of factory object that owns the node      */
/*                          being created.                                    */
/*  nodeName      (input) - name of node being created.                       */
/*  nodeObject    (input) - address of the node object that has just been     */
/*                          created.                                          */
/* Returns:                                                                   */
/*                                                                            */
/*  Pointer to the local context.                                             */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Creates any plug-in context for an instance of the node object. This is    */
/* invoked by the broker whenever a node instance is constructed. Nodes are   */
/* constructed whenever a messageflow containing the node is deployed by the  */
/* broker.                                                                    */
/*                                                                            */
/* This function has 4 responsibilities:                                      */
/*                                                                            */
/* 1. Optionally verify the name of the node is supported by the factory.     */
/* 2. Allocate node instance specific data areas. e.g. For attribute data or  */
/*    terminals.                                                              */
/* 3. Perform any additional resource acquisition or initialisation that is   */
/*    required for the processing of the node.                                */
/* 4. Return the address of the context to the calling function.              */
/*                                                                            */
/******************************************************************************/
CciContext *createNodeContext(CciFactory *factoryObject, 
                              CciChar *nodeName, 
                              CciNode *nodeObject)
{
    int rc;
	static char *functionName="createNodeContext()";
    context_t *contextP = NULL;
#if defined(POST_201_VERSION)
	CNI_BROKER_INFO_ST brokerInfo={CNI_BROKER_INFO_ST_DEFAULT};
#else
    CNI_BROKER_INFO_ST brokerInfo;
#endif
	CciChar messageFlowNameUnicode[MAX_STRING_SIZE + 1];
	char messageFlowNameMB[MAX_STRING_SIZE + 1];
	messageFlow_t *mfP,
	              defaultMsgFlow = {MESSAGE_FLOW_DEFAULT},
		          *messageFlowAnchorP;
	deployedPostitCreate_t *deployedPostitCreateP;
	deployedPostitApply_t *deployedPostitApplyP;
	deployedPostitDelete_t *deployedPostitDeleteP;
	char msgBuffer[200],
		 nodeNameMB[100];

    TRACE_ENTRY("createNodeContext", ("factoryObject=%.8x, nodeName=%S, "
		                              "nodeObject=%.8x",
									  factoryObject,
									  nodeName,
									  nodeObject));

	/* initialise the broker info structure */
	ALLOCATE_STRING_STORAGE(brokerInfo.brokerName, 300);
	ALLOCATE_STRING_STORAGE(brokerInfo.executionGroupName, 300);
	ALLOCATE_STRING_STORAGE(brokerInfo.messageFlowName, 300);
	ALLOCATE_STRING_STORAGE(brokerInfo.queueManagerName, 300);

	/* get the name of our queue manager */
    TRACE_MID(TRACE_INFO, ("About to call cniGetBrokerInfo"));

	cniGetBrokerInfo(0,           /* retcode does not appear to be set */
			         nodeObject,
					 &brokerInfo);

	UCsprintf(messageFlowNameMB, "%S", brokerInfo.messageFlowName.buffer);
	UCcpy(messageFlowNameUnicode, brokerInfo.messageFlowName.buffer);
    TRACE_MID(TRACE_INFO, ("Extracted messageFlowName is '%s'",
			               messageFlowNameMB));

	/* Is this the first postit for this Message Flow */
	threadMutexLock(postitChainMutex);
	mfP = findMessageFlow(messageFlowNameUnicode, &messageFlowAnchor);

	if (!mfP)
	{
		/* create new message flow entry */
		mfP = (messageFlow_t *)malloc(sizeof(messageFlow_t));
		memcpy(mfP, &defaultMsgFlow, sizeof(messageFlow_t));
		mfP->name = UCdup(messageFlowNameUnicode);

		/* add to message flow chain */
		messageFlowAnchorP = &messageFlowAnchor;
		INSERT_AFTER_IN_LINKED_LIST(messageFlowAnchorP, mfP);
	}
	threadMutexUnlock(postitChainMutex);

    TRACE_MID(TRACE_INFO, ("Current messageFlow entry at %.8x", mfP));

	if (!brokerInfoObtained)
	{
		/* copy the process specific information, both as single and Unicode strings */
		UCsprintf(brokerName, "%S", brokerInfo.brokerName.buffer);
		UCcpy(brokerNameUC, brokerInfo.brokerName.buffer);
        TRACE_MID(TRACE_INFO, ("Extracted brokerName is '%s'",
			                   brokerName));

		UCsprintf(executionGroupName, "%S", brokerInfo.executionGroupName.buffer);
		UCcpy(executionGroupNameUC, brokerInfo.executionGroupName.buffer);
        TRACE_MID(TRACE_INFO, ("Extracted executionGroupName is '%s'",
			                   executionGroupName));

		UCsprintf(queueManagerName, "%S", brokerInfo.queueManagerName.buffer);
		UCcpy(queueManagerNameUC, brokerInfo.queueManagerName.buffer);
        TRACE_MID(TRACE_INFO, ("Extracted queueManagerName is '%s'",
			                   queueManagerName));

		brokerInfoObtained = TRUE;
	}

	FREE_STRING_STORAGE(brokerInfo.brokerName);
	FREE_STRING_STORAGE(brokerInfo.executionGroupName);
	FREE_STRING_STORAGE(brokerInfo.messageFlowName);
	FREE_STRING_STORAGE(brokerInfo.queueManagerName);

    /* Allocate a pointer to the local context */
    contextP = malloc(sizeof(context_t));

    /* If malloc was successful */
    if (!contextP)
	{
        /* Error handling if malloc failed */
        TRACE_RETURN_TYPE(RSN_022, void *, contextP);
	}

    TRACE_MID(TRACE_INFO, ("Context structure allocated at %.8x", contextP));

    /* initialise returned storage for context anchor */
    memset (contextP, 0, sizeof(context_t)); /* Clear the context area */
    contextP->nodeObject = nodeObject;
	contextP->messageFlowName = mfP->name;

    TRACE_MID(TRACE_INFO, ("<===========================>"));
    TRACE_MID(TRACE_INFO, ("%S node current", nodeName));
    TRACE_MID(TRACE_INFO, ("<===========================>"));

    /* Verify the name of the node is supported by the factory */
    if (!UCcmp(nodeName, C.POSTIT_CREATE_NODE))
	{
		/* postit save node */
		contextP->nodeType = POSTIT_CREATE;
        contextP->nodeName = UCdup(nodeName);

		/* Build attribute list for the node */
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_POSTIT_NAME, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_COPY_DATA_FROM, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_TIMEOUT_SECONDS, CCI_VALUE_TYPE_INTEGER);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_NUM_APPLICATIONS, CCI_VALUE_TYPE_INTEGER);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_POSTIT_SCOPE, CCI_VALUE_TYPE_BYTEARRAY);

		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_LABEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_USER_TRACE_LEVEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_TRACE_LEVEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_USER_TRACE_FILTER, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_TRACE_FILTER, CCI_VALUE_TYPE_BYTEARRAY);

		/* this node has a single input and 3 output terminals */
		insertInputTerminalListEntry(contextP, C.TERM_IN);

		insertOutputTerminalListEntry(contextP, C.TERM_OUT);
		insertOutputTerminalListEntry(contextP, C.TERM_FAILURE);
		insertOutputTerminalListEntry(contextP, C.TERM_NO_DATA_COPIED);

		/* create a new deployed postit create control block */
		deployedPostitCreateP = (deployedPostitCreate_t *)
			                         malloc(sizeof(deployedPostitCreate_t));
		memset(deployedPostitCreateP, 0, sizeof(deployedPostitCreate_t));
	    TRACE_MID(TRACE_INFO, ("Deployed PostitCreate control block allocated at %.8x",
			                   deployedPostitCreateP));

		/* lock the postit mutex and add to global chain */
		threadMutexLock(postitChainMutex);
		INSERT_AFTER_IN_LINKED_LIST(postitCreateAnchorP,
					                deployedPostitCreateP);
		contextP->nodeDataP = deployedPostitCreateP;

		/* increment the global node counter */
		numPostitCreateNodes++;
	    TRACE_MID(TRACE_INFO, ("Global count of PostitCreate nodes incremented to %d",
			                   numPostitCreateNodes));
		threadMutexUnlock(postitChainMutex);

		/* As of MQSI V2.0.2, an attribute for which the user has accepted the default, is
		 * no longer included in the deployment message.  We therefore need to initialise
		 * all attributes to their default values within the plug-in.
		 * The exception to this is a value for which no 'value=xxx' label is specified and 
		 * for which 'valueMandatory=true' has been specified.  In this case, the control 
		 * centre will force the user to enter a value into the field, and this will be 
		 * included in the deployment message.  Of course, we must still check for invalid
		 * values such as blanks.
		 */
	    TRACE_MID(TRACE_INFO, ("<***********> Setting default values"));
		setAttribute(contextP,
		             C.ATTR_POSTIT_NAME,
                     C.ATTR_POSTIT_NAME_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_COPY_DATA_FROM,
                     C.ATTR_COPY_DATA_FROM_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_TIMEOUT_SECONDS,
                     C.ATTR_TIMEOUT_SECONDS_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_NUM_APPLICATIONS,
                     C.ATTR_NUM_APPLICATIONS_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_POSTIT_SCOPE,
                     C.ATTR_POSTIT_SCOPE_DEFAULT_VALUE);
	    TRACE_MID(TRACE_INFO, ("<***********> Default values set"));
	}
	else
    if (!UCcmp(nodeName, C.POSTIT_APPLY_NODE))
	{
		/* postit restore node */
		contextP->nodeType = POSTIT_APPLY;
        contextP->nodeName = UCdup(nodeName);

		/* Build attribute list for the node */
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_POSTIT_NAME, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_COPY_DATA_TO, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_MATCH_ON, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_POSTIT_SCOPE, CCI_VALUE_TYPE_BYTEARRAY);

		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_LABEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_USER_TRACE_LEVEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_TRACE_LEVEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_USER_TRACE_FILTER, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_TRACE_FILTER, CCI_VALUE_TYPE_BYTEARRAY);

		/* this node has a single input terminal and 4 output terminals */
		insertInputTerminalListEntry(contextP, C.TERM_IN);

		insertOutputTerminalListEntry(contextP, C.TERM_OUT);
		insertOutputTerminalListEntry(contextP, C.TERM_FAILURE);
		insertOutputTerminalListEntry(contextP, C.TERM_NO_MATCH);

		/* create a new deployed postit apply control block */
		deployedPostitApplyP = (deployedPostitApply_t *)
			                         malloc(sizeof(deployedPostitApply_t));
		memset(deployedPostitApplyP, 0, sizeof(deployedPostitApply_t));
	    TRACE_MID(TRACE_INFO, ("Deployed PostitApply control block allocated at %.8x",
			                   deployedPostitApplyP));

		/* lock the postit mutex and add to global chain */
		threadMutexLock(postitChainMutex);
		INSERT_AFTER_IN_LINKED_LIST(postitApplyAnchorP,
					                deployedPostitApplyP);

		contextP->nodeDataP = deployedPostitApplyP;

		/* increment the global node counter */
		numPostitApplyNodes++;
	    TRACE_MID(TRACE_INFO, ("Global count of PostitApply nodes incremented to %d",
			                   numPostitApplyNodes));
		threadMutexUnlock(postitChainMutex);

		/* As of MQSI V2.0.2, an attribute for which the user has accepted the default, is
		 * no longer included in the deployment message.  We therefore need to initialise
		 * all attributes to their default values within the plug-in.
		 * The exception to this is a value for which no 'value=xxx' label is specified and 
		 * for which 'valueMandatory=true' has been specified.  In this case, the control 
		 * centre will force the user to enter a value into the field, and this will be 
		 * included in the deployment message.  Of course, we must still check for invalid
		 * values such as blanks.
		 */
	    TRACE_MID(TRACE_INFO, ("Setting default values"));
		setAttribute(contextP,
		             C.ATTR_POSTIT_NAME,
                     C.ATTR_POSTIT_NAME_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_COPY_DATA_TO,
                     C.ATTR_COPY_DATA_TO_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_MATCH_ON,
                     C.ATTR_MATCH_ON_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_POSTIT_SCOPE,
                     C.ATTR_POSTIT_SCOPE_DEFAULT_VALUE);
	    TRACE_MID(TRACE_INFO, ("<***********> Default values set"));
	}
	else
    if (!UCcmp(nodeName, C.POSTIT_DELETE_NODE))
	{
		/* postit restore node */
		contextP->nodeType = POSTIT_DELETE;
        contextP->nodeName = UCdup(nodeName);

		/* Build attribute list for the node */
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_POSTIT_NAME, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_COPY_DATA_FROM, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_POSTIT_SCOPE, CCI_VALUE_TYPE_BYTEARRAY);

		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_LABEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_USER_TRACE_LEVEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_TRACE_LEVEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_USER_TRACE_FILTER, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_TRACE_FILTER, CCI_VALUE_TYPE_BYTEARRAY);

		/* this node has a single input terminal and 4 output terminals */
		insertInputTerminalListEntry(contextP, C.TERM_IN);

		insertOutputTerminalListEntry(contextP, C.TERM_OUT);
		insertOutputTerminalListEntry(contextP, C.TERM_FAILURE);
		insertOutputTerminalListEntry(contextP, C.TERM_NO_MATCH);

		/* create a new deployed postit delete control block */
		deployedPostitDeleteP = (deployedPostitDelete_t *)
			                         malloc(sizeof(deployedPostitDelete_t));
		memset(deployedPostitDeleteP, 0, sizeof(deployedPostitDelete_t));
	    TRACE_MID(TRACE_INFO, ("Deployed PostitDelete control block allocated at %.8x",
			                   deployedPostitDeleteP));

		/* lock the postit mutex and add to global chain */
		threadMutexLock(postitChainMutex);
		INSERT_AFTER_IN_LINKED_LIST(postitDeleteAnchorP,
					                deployedPostitDeleteP);

		contextP->nodeDataP = deployedPostitDeleteP;

		/* increment the global node counter */
		numPostitDeleteNodes++;
	    TRACE_MID(TRACE_INFO, ("Global count of PostitDelete nodes incremented to %d",
			                   numPostitDeleteNodes));
		threadMutexUnlock(postitChainMutex);

		/* As of MQSI V2.0.2, an attribute for which the user has accepted the default, is
		 * no longer included in the deployment message.  We therefore need to initialise
		 * all attributes to their default values within the plug-in.
		 * The exception to this is a value for which no 'value=xxx' label is specified and 
		 * for which 'valueMandatory=true' has been specified.  In this case, the control 
		 * centre will force the user to enter a value into the field, and this will be 
		 * included in the deployment message.  Of course, we must still check for invalid
		 * values such as blanks.
		 */
	    TRACE_MID(TRACE_INFO, ("Setting default values"));
		setAttribute(contextP,
		             C.ATTR_POSTIT_NAME,
                     C.ATTR_POSTIT_NAME_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_COPY_DATA_FROM,
                     C.ATTR_COPY_DATA_FROM_DEFAULT_VALUE);
		setAttribute(contextP,
		             C.ATTR_POSTIT_SCOPE,
                     C.ATTR_POSTIT_SCOPE_DEFAULT_VALUE);
	    TRACE_MID(TRACE_INFO, ("<***********> Default values set"));
	}
	else
	{
	    /* Error handling if name of node is not supported by factory */
        TRACE_MID(TRACE_ERROR, ("Unrecognised node"));
		/* throw an exception here...*/
		UCsprintf(msgBuffer, "Unrecognised node ('%S') passed to createContext()",
			               contextP->nodeName);
		UCsprintf(nodeNameMB, "'%S'", contextP->nodeName);
		cciThrowException(&rc, 
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      C.PLUGIN_MESSAGE_CATALOGUE,
			              MJB_UNRECOGNISED_NODE_CREATE_CONTEXT,  /* messageNumber */ 
			              msgBuffer,              /* traceText */
#if !defined(WIN32)
						  brokerName,
#endif
						  nodeNameMB,
			              (char *)0);
	    free(contextP);
	}

	/* 
	 * If this is the first postit node encountered, create the
	 * alarm thread
	 */
	if (!alarmThreadCreated)
	{
		rc = createAlarmThread(contextP);

		if (rc != TT_SUCCESS)
		{
			TRACE_RETURN_TYPE(RSN_005, void *, contextP);
		}
	}

    /* Return the address of the context to the calling function */
    TRACE_RETURN_TYPE(RSN_008, void *, contextP);
}

/******************************************************************************/
/* createAlarmThread                                                          */
/******************************************************************************/
int createAlarmThread(IN context_t *contextP)
{
	char errorMessage[1000];
	static char *functionName="createAlarmThread()";
	int rc;

    TRACE_ENTRY("createAlarmThread", ("contextP=%.8x",
		                              contextP));

    /* 
	 * During message flow initialisation, it's unlikely that this code
	 * will be run concurrently.  However, just to be sure, we'll 
	 * serialise the code through this section.
	 */
	TRACE_MID(TRACE_INFO, ("About to lock alarmThreadSyncMutex"));
	threadMutexLock(alarmThreadSyncMutex);

	/* check again in case we lost the race... */
	if (alarmThreadCreated)
	{
	 	TRACE_MID(TRACE_INFO, ("Lost the race to create alarm thread"));
		TRACE_MID(TRACE_INFO, ("About to unlock alarmThreadSyncMutex"));
		threadMutexUnlock(alarmThreadSyncMutex);
		TRACE_RETURN(RSN_005, TT_SUCCESS);
	}
		
	/* create the alarm thread */
	TRACE_MID(TRACE_INFO, ("About to create alarm thread"));
	if (threadCreateThread((threadFunction_t)processAlarms,
						  (threadParameter_t)contextP,
						   &alarmThreadHandle) != THREAD_OK)
	{
		/* need an exception here */
		TRACE_MID(TRACE_INFO, ("About to unlock alarmThreadSyncMutex"));
		threadMutexUnlock(alarmThreadSyncMutex);
		TRACE_RETURN(RSN_006, TT_FAILED);
	}
	TRACE_MID(TRACE_INFO, ("Alarm thread (%d) successfully created",
						   threadGetThreadId(alarmThreadHandle)));

	/* wait up to 10 seconds for child thread to signal us that he's up and running */
	TRACE_MID(TRACE_INFO, ("About to wait for alarm thread to initialise"));
	
	rc = threadCondTimedWait(alarmThreadSyncCondVar, 10000);

	if (rc != THREAD_OK && rc != THREAD_TIMED_OUT)
	{
		/* need an exception here */
		TRACE_MID(TRACE_INFO, ("About to unlock alarmThreadSyncMutex"));
		threadMutexUnlock(alarmThreadSyncMutex);
		TRACE_RETURN(RSN_007, TT_FAILED);
	}

	/* check response from alarm thread */
	if (!alarmThreadInitialised)
	{
		sprintf(errorMessage, "Alarm thread failed to initialise.  Reason='%s'",
							  alarmThreadErrorMessage);
		TRACE_MID(TRACE_ERROR, ("%s", errorMessage));

		/* throw exception */
		cciThrowException(&rc,                       /* returnCode */
			   CCI_FATAL_EXCEPTION,       /* type */
			   __FILE__,                  /* file */
			   __LINE__,                  /* line */
			   functionName,              /* function */
			   C.PLUGIN_MESSAGE_CATALOGUE,    /* messageSource */
			   MJB_ALARM_THREAD_INIT_FAILED, /* messageNumber */ 
			   errorMessage,              /* traceText */
#if !defined(WIN32)
			   brokerName,
#endif
			   errorMessage,
			   (char *)0);
	}
	else
	{
		alarmThreadCreated = TRUE;

		TRACE_MID(TRACE_INFO, ("Alarm thread initialised ok"));
	}

	TRACE_MID(TRACE_INFO, ("About to unlock alarmThreadSyncMutex"));
	threadMutexUnlock(alarmThreadSyncMutex);
	TRACE_RETURN(RSN_008, TT_SUCCESS);
}

/******************************************************************************/
/* createSharedMemoryCleanupThread                                            */
/******************************************************************************/
int createSharedMemoryGarbageCollectionThread(IN context_t *contextP)
{
	char errorMessage[1000];
	static char *functionName="createSharedMemoryGarbageCollectionThread()";
	int rc;

    TRACE_ENTRY("createSharedMemoryGarbageCollectionThread", ("contextP=%.8x",
		                                                       contextP));

    /* 
	 * During message flow initialisation, it's unlikely that this code
	 * will be run concurrently.  However, just to be sure, we'll 
	 * serialise the code through this section.
	 */
	TRACE_MID(TRACE_INFO, ("About to lock alarmThreadSyncMutex"));
	threadMutexLock(shmemThreadSyncMutex);

	/* check again in case we lost the race... */
	if (shmemThreadCreated)
	{
	 	TRACE_MID(TRACE_INFO, ("Lost the race to create shared memory garbage collection thread"));
		TRACE_MID(TRACE_INFO, ("About to unlock shmemThreadSyncMutex"));
		threadMutexUnlock(shmemThreadSyncMutex);
		TRACE_RETURN(RSN_005, TT_SUCCESS);
	}
		
	/* create the garbage collection thread */
	TRACE_MID(TRACE_INFO, ("About to create shared memory garbage collection thread"));
	if (threadCreateThread((threadFunction_t)sharedMemoryGarbageCollection,
						  (threadParameter_t)contextP,
						   &shmemThreadHandle) != THREAD_OK)
	{
		/* need an exception here */
		TRACE_MID(TRACE_INFO, ("About to unlock shmemThreadSyncMutex"));
		threadMutexUnlock(shmemThreadSyncMutex);
		TRACE_RETURN(RSN_006, TT_FAILED);
	}
	TRACE_MID(TRACE_INFO, ("Garbage collection thread (%d) successfully created",
						   threadGetThreadId(shmemThreadHandle)));

	/* wait up to 10 seconds for child thread to signal us that he's up and running */
	TRACE_MID(TRACE_INFO, ("About to wait for garbage collection thread to initialise"));
	
	rc = threadCondTimedWait(shmemThreadSyncCondVar, 10000);

	if (rc != THREAD_OK && rc != THREAD_TIMED_OUT)
	{
		/* need an exception here */
		TRACE_MID(TRACE_INFO, ("About to unlock shmemThreadSyncMutex"));
		threadMutexUnlock(shmemThreadSyncMutex);
		TRACE_RETURN(RSN_007, TT_FAILED);
	}

	/* check response from alarm thread */
	if (!shmemThreadInitialised)
	{
		sprintf(errorMessage, "Garbage collection thread failed to initialise.  Reason='%s'",
							  shmemThreadErrorMessage);
		TRACE_MID(TRACE_ERROR, ("%s", errorMessage));

		/* throw an exception */
		cciThrowException(&rc,                       /* returnCode */
			   CCI_FATAL_EXCEPTION,       /* type */
			   __FILE__,                  /* file */
			   __LINE__,                  /* line */
			   functionName,              /* function */
			   C.PLUGIN_MESSAGE_CATALOGUE,    /* messageSource */
			   MJB_SHARED_MEMORY_GARBAGE_COLLECTION_THREAD_INIT_FAILED, /* messageNumber */ 
			   errorMessage,              /* traceText */
#if !defined(WIN32)
			   brokerName,
#endif
			   errorMessage,
			   (char *)0);
	}
	else
	{
		shmemThreadCreated = TRUE;

		TRACE_MID(TRACE_INFO, ("Shared memory garbage collection thread initialised ok"));
	}

	TRACE_MID(TRACE_INFO, ("About to unlock shmemThreadSyncMutex"));
	threadMutexUnlock(shmemThreadSyncMutex);
	TRACE_RETURN(RSN_008, TT_SUCCESS);
}

/******************************************************************************/
/* deleteNodeContext                                                          */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context       (input)  - pointer to the current context                   */
/*                          created.                                          */
/* Returns:                                                                   */
/*                                                                            */
/*  Nothing.                                                                  */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Called whenever an instance of a node object is destroyed.  Any resources  */
/* aquired by the node e.g. context or other resources, should be released    */
/* at this time.                                                              */
/*                                                                            */
/******************************************************************************/
void deleteNodeContext(CciContext *dataP)
{
	context_t *contextP = (context_t *)dataP;
	int rc;
	bool_t scopeIsBroker;
    static char *functionName="deleteNodeContext()";
	deployedPostitCreate_t *deployedPostitCreateP;
	deployedPostitApply_t *deployedPostitApplyP;
	deployedPostitDelete_t *deployedPostitDeleteP;
	postitCreateInstance_t *postitInstanceP,
		                   *nextInstanceP;
	char infoMessage[1000],
		 *nodeNameP,
	     mbLabelString[MAX_STRING_SIZE + 1];

    TRACE_ENTRY("deleteNodeContext", ("dataP=%.8x", dataP));

	/* lock mutex just to be sure there's no race */
	TRACE_MID(TRACE_INFO, ("About to lock postitChainMutex"));
	threadMutexLock(postitChainMutex);

	/* delete deployed node structure pointed to by this context */
	if (contextP->nodeType == POSTIT_CREATE)
	{
		deployedPostitCreateP = (deployedPostitCreate_t *)contextP->nodeDataP;
		TRACE_MID(TRACE_INFO, ("Processing deployed postit create at %.8x", deployedPostitCreateP));

		/* delete all postit instances belonging to this deployed postit */
        postitInstanceP = deployedPostitCreateP->instanceAnchor.nextP;

		while (postitInstanceP)
		{
			TRACE_MID(TRACE_INFO, ("Processing postit instance at %.8x", postitInstanceP));
			nextInstanceP = postitInstanceP->nextP;

			if (postitInstanceP->global)
			{
				/* shared memory must be locked around the free operation */
				shmLock(deployedPostitCreateP->scope==SCOPE_BROKER);
				freePostit(postitInstanceP);
				shmUnlock(deployedPostitCreateP->scope==SCOPE_BROKER);
			}
			else
			{
				freePostit(postitInstanceP);
			}

			postitInstanceP = nextInstanceP;
		}

		/* remove deployed postit from global chain */
		REMOVE_FROM_LINKED_LIST(deployedPostitCreateP);

		/* free any linked resources */
		free(deployedPostitCreateP->copyFromLocation);
		if (deployedPostitCreateP->postitName)
		{
			TRACE_MID(TRACE_INFO, ("Freeing deployed postit name"));
			free(deployedPostitCreateP->postitName);
		}

		/* finally, free the deployed node control block itself */
		free(deployedPostitCreateP);
		TRACE_MID(TRACE_INFO, ("Freeing deployed postit at %.8x", deployedPostitCreateP));

		/* decrement number of create nodes */
		numPostitCreateNodes--;
		TRACE_MID(TRACE_INFO, ("Number of postit create nodes decremented to %d",
			                   numPostitCreateNodes));
	}
	else
	if (contextP->nodeType == POSTIT_APPLY)
	{
		deployedPostitApplyP = (deployedPostitApply_t *)contextP->nodeDataP;
		TRACE_MID(TRACE_INFO, ("Processing deployed postit apply at %.8x", deployedPostitApplyP));

		/* remove deployed postit from global chain */
		REMOVE_FROM_LINKED_LIST(deployedPostitApplyP);

		/* free any linked resources */
		free(deployedPostitApplyP->copyToLocation);
		if (deployedPostitApplyP->postitName)
		{
			TRACE_MID(TRACE_INFO, ("Freeing deployed postit name"));
			free(deployedPostitApplyP->postitName);
		}

		/* finally, free the deployed node control block itself */
		free(deployedPostitApplyP);
		TRACE_MID(TRACE_INFO, ("Freeing deployed postit at %.8x", deployedPostitApplyP));

		numPostitApplyNodes--;
		TRACE_MID(TRACE_INFO, ("Number of postit apply nodes decremented to %d",
			                   numPostitApplyNodes));
	}
	else
	{
		deployedPostitDeleteP = (deployedPostitDelete_t *)contextP->nodeDataP;
		TRACE_MID(TRACE_INFO, ("Processing deployed postit delete at %.8x", deployedPostitDeleteP));

		/* remove deployed postit delete from global chain */
		REMOVE_FROM_LINKED_LIST(deployedPostitDeleteP);

		/* free any linked resources */
		free(deployedPostitDeleteP->copyFromLocation);
		if (deployedPostitDeleteP->postitName)
		{
			TRACE_MID(TRACE_INFO, ("Freeing deployed postit name"));
			free(deployedPostitDeleteP->postitName);
		}

		/* finally, free the deployed node control block itself */
		free(deployedPostitDeleteP);
		TRACE_MID(TRACE_INFO, ("Freeing deployed postit at %.8x", deployedPostitDeleteP));

		/* decrement number of create nodes */
		numPostitDeleteNodes--;
		TRACE_MID(TRACE_INFO, ("Number of postit delete nodes decremented to %d",
			                   numPostitDeleteNodes));
	}

	TRACE_MID(TRACE_INFO, ("alarmThreadInitialised is %s",
		alarmThreadInitialised ? "TRUE" :"FALSE"));


	/* shutdown the alarm thread if there's no more active postit nodes */
	if (!numPostitCreateNodes && !numPostitApplyNodes &&
	    !alarmThreadShuttingDown && alarmThreadInitialised)
	{
		alarmThreadShuttingDown = TRUE;

		/* shutdown the alarm thread */
		shutdownAlarmRequested = TRUE;
		TRACE_MID(TRACE_INFO, ("Signalling alarm thread to shutdown"));
		threadCondSignal(alarmThreadWorkCondVar);
		TRACE_MID(TRACE_INFO, ("About to unlock postitChainMutex"));
		threadMutexUnlock(postitChainMutex);

		/* wait for alarm to terminate before proceeding */
		TRACE_MID(TRACE_INFO, ("About to wait for alarm thread to terminate"));
		rc = threadJoin(alarmThreadHandle);

		TRACE_MID(TRACE_INFO, ("About to lock postitChainMutex"));
		threadMutexLock(postitChainMutex);
		if (rc == THREAD_OK)
		{
			shutdownAlarmRequested = FALSE;
			alarmThreadCreated = FALSE;
			alarmThreadInitialised = FALSE;
			alarmThreadHandle = 0;
			alarmThreadShuttingDown = FALSE;
		}
	}

	/* shutdown the garbage collector thread if there's no more active postit nodes */
	if (!numPostitCreateNodes && !numPostitApplyNodes &&
	    !shmemThreadShuttingDown && shmemThreadInitialised)
	{
		shmemThreadShuttingDown = TRUE;

		/* shutdown the garbage collector thread */
		shutdownShmemRequested = TRUE;
		TRACE_MID(TRACE_INFO, ("Signalling garbage collection thread to shutdown"));

		TRACE_MID(TRACE_INFO, ("About to lock shmemThreadWorkMutex"));
		threadMutexLock(shmemThreadWorkMutex);
		threadCondSignal(shmemThreadWorkCondVar);
		TRACE_MID(TRACE_INFO, ("About to unlock shmemThreadWorkMutex"));
		threadMutexUnlock(shmemThreadWorkMutex);

		/* wait for gc to terminate before proceeding */
		TRACE_MID(TRACE_INFO, ("About to wait for garbage collector thread to terminate"));
		rc = threadJoin(shmemThreadHandle);

		if (rc == THREAD_OK)
		{
			shutdownShmemRequested = FALSE;
			shmemThreadCreated = FALSE;
			shmemThreadInitialised = FALSE;
			shmemThreadHandle = 0;
			shmemThreadShuttingDown = FALSE;
		}
	}

	/* detach shared memory if attached and no more active postit nodes */
	if (!numPostitCreateNodes && !numPostitApplyNodes && 
		(sharedMemoryAttached_Broker || sharedMemoryAttached_Platform))
	{
		TRACE_MID(TRACE_INFO, ("About to lock sharedMemoryInitTermMutex"));
		threadMutexLock(shmemInitTermMutex);

		if (sharedMemoryAttached_Broker)
		{
			scopeIsBroker = TRUE;
			/* remove our execution group entry first... */
			shmLock(scopeIsBroker);
			removeExecutionGroupFromGlobalStorage(scopeIsBroker);
			shmUnlock(scopeIsBroker);
			rc = shmDetach(currentPID, scopeIsBroker);
			if (rc != TT_SUCCESS)
			{
				/* deal with errors... */
			}
			sharedMemoryAttached_Broker = FALSE;
			sharedMemoryCBP_Broker = NULL;
		}

		if (sharedMemoryAttached_Platform)
		{
			scopeIsBroker = FALSE;
			/* remove our execution group entry first... */
			shmLock(scopeIsBroker);
			removeExecutionGroupFromGlobalStorage(scopeIsBroker);
			shmUnlock(scopeIsBroker);
			rc = shmDetach(currentPID, scopeIsBroker);
			if (rc != TT_SUCCESS)
			{
				/* deal with errors... */
			}
			sharedMemoryAttached_Platform = FALSE;
			sharedMemoryCBP_Platform = NULL;
		}

		TRACE_MID(TRACE_INFO, ("About to unlock sharedMemoryInitTermMutex"));
		threadMutexUnlock(shmemInitTermMutex);
	}

	TRACE_MID(TRACE_INFO, ("About to unlock postitChainMutex"));
	threadMutexUnlock(postitChainMutex);

	/* clear the attributes associated with this context */
    deleteAttributeTableEntries(contextP);

	/* clear the terminals associated with this context */
    deleteAllTerminalListEntries(contextP);

	/* 
	 * Finally, free up the context structure, but first create multibyte strings for
	 * inclusion in log message.
	 */
    UCToStr(contextP->nodeLabel,
		    mbLabelString,
		    sizeof(mbLabelString),
		    NULL);

	if (contextP->nodeType == POSTIT_CREATE)
	{
		nodeNameP = POSTIT_CREATE_NODE_MB;
	}
	else
	{
		nodeNameP = POSTIT_APPLY_NODE_MB;
	}

	free(contextP->nodeName);
	free(contextP->nodeLabel);
    free(contextP);

	sprintf(infoMessage, "Successfully deleted node context for %s (%s)", 
		                  nodeNameP, mbLabelString);
	TRACE_MID(TRACE_INFO, ("%s", infoMessage));

    cciLog(&rc,                       /* returnCode */
	       CCI_LOG_INFORMATION,       /* type */
		   __FILE__,                  /* file */
		   __LINE__,                  /* line */
		   functionName,              /* function */
		   C.PLUGIN_MESSAGE_CATALOGUE, /* messageSource */
           MJB_NODE_CONTEXT_DELETED,  /* messageNumber */ 
		   infoMessage, /* traceText */
#if !defined(WIN32)
		   brokerName,
#endif
           nodeNameP,
		   mbLabelString,
		   (char *)0);

	TRACE_RETURN_VOID(RSN_009);
}


/******************************************************************************/
/* setAttribute                                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context        (input)  - pointer to the current context                  */
/*  attributeName  (input)  - attribute name for which value is to be set     */
/*  attributeValue (input) -  value to be set for the attribute               */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/*  1 - failure, unknown attribute name                                       */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Sets the value of an attribute on a specific node instance. It is invoked  */
/* by the broker when a configuration request is received that attempts to    */
/* set the value of a node attribute. The broker will have verified that the  */
/* attribute name is valid for the node.                                      */
/*                                                                            */
/* This function has 3 responsibilities:                                      */
/*                                                                            */
/* 1. Verify value of attribute is correctly specified, otherwise a           */
/*    configuration exception should be thrown.                               */
/* 2. Store the attribute value in the context.                               */
/* 3. Throw a configuration exception if errors occur.                        */
/*                                                                            */
/******************************************************************************/
int setAttribute(CciContext *dataP,
				 CciChar *attributeName,
				 CciChar *attributeValue)
{
	int rc;
	context_t *contextP = (context_t *)dataP;

    TRACE_ENTRY("setAttribute", ("dataP=%.8x, attributeName=%S, "
		                         "attributeValue='%S'",
								 dataP, attributeName, attributeValue));

	/* call common plugin function */
	rc = commonSetAttribute(contextP,
		                    attributeName,
						    attributeValue);

	if (rc != CCI_SUCCESS)
	{
        TRACE_RETURN_TYPE(RSN_011, int, rc);
	}

	/* special processing for postit attributes */
    rc = processAttribute(contextP,
		                  attributeName,
		                  attributeValue);
	if (rc != TT_SUCCESS)
	{
	    TRACE_RETURN_TYPE(RSN_012, int, CCI_EXCEPTION_CONFIGURATION);
	}
    TRACE_RETURN_TYPE(RSN_012, int, TT_SUCCESS);
}


/******************************************************************************/
/* processAttribute                                                           */
/******************************************************************************/
int processAttribute(IN context_t *contextP,
				     IN CciChar *attributeName,
				     IN CciChar *attributeValue)
{
	deployedPostitCreate_t *deployedPostitCreateP;
	deployedPostitApply_t *deployedPostitApplyP;
	deployedPostitDelete_t *deployedPostitDeleteP;
    CciChar copyString[MAX_STRING_SIZE + 1],
            matchString[MAX_STRING_SIZE + 1],
            scopeString[MAX_STRING_SIZE + 1];
	int copyStringLength,
		matchStringLength,
		scopeStringLength,
		timeoutSeconds,
		numApplications,
		retCode;
	char msgBuffer[2000],
		 nodeName[1000],
		 propertyName[100];
	static char *functionName = "processAttribute";

    TRACE_ENTRY("processAttribute", ("contextP=%.8x, attributeName=%S, "
		                             "attributeValue='%S'",
								     contextP, attributeName, attributeValue));

	/* 
	 * During message flow initialisation, it's unlikely that this code
	 * will be run concurrently.  However, just to be sure, we'll 
	 * serialise the code through this section.
	 */
	TRACE_MID(TRACE_INFO, ("About to lock postitChainMutex"));
	threadMutexLock(postitChainMutex);

	if (contextP->nodeType == POSTIT_CREATE)
	{
		deployedPostitCreateP = (deployedPostitCreate_t *)(contextP->nodeDataP);

		/* initialise the deployed postit create structure */

		/* deal with postitName */
		if (!UCcmp(attributeName, C.ATTR_POSTIT_NAME))
		{
			copyStringLength = sizeof(copyString);
			if (getStringAttribute(contextP, 
					               C.ATTR_POSTIT_NAME,
								   &copyStringLength,
					               copyString) == TT_SUCCESS)
			{
				/* postit name is an optional property, so check for null string */
				if (UClen(copyString))
				{
					/* free default value first */
					if (deployedPostitCreateP->postitName)
					{
						free(deployedPostitCreateP->postitName);
					}
					/* name was supplied so copy string into deployed postit */
					deployedPostitCreateP->postitName = UCdup(copyString);
				}
			}
		}

		/* process the CopyDataFrom property */
		if (!UCcmp(attributeName, C.ATTR_COPY_DATA_FROM))
		{
			copyStringLength = sizeof(copyString);
			if (getStringAttribute(contextP, 
					               C.ATTR_COPY_DATA_FROM,
								   &copyStringLength,
					               copyString) == TT_SUCCESS)
			{
				/* 
				 * String can be 'Root' or it must begin with 'Root.' - anything else
				 * is invalid.
				 */
				if (!(UClen(copyString) == 4 && !UCcmp(copyString, C.ROOT) ||
				      UClen(copyString) > 5 && !UCncmp(copyString, C.ROOT_DOT, UClen(C.ROOT_DOT))))
				{
					/* throw an exception here...*/
					UCsprintf(msgBuffer, "The syntax element name specified in the %S "
						               "property of node %S(%S) is invalid. "
									   "Valid names must be either 'Root' or must begin "
									   "with 'Root.'.",
									   C.ATTR_COPY_DATA_FROM,
									   contextP->nodeName,
									   contextP->nodeLabel ? contextP->nodeLabel : C.NULL_LITERAL);

					TRACE_MID(TRACE_ERROR, ("%s", msgBuffer));

					UCsprintf(propertyName, "%S", C.ATTR_COPY_DATA_FROM);

					UCsprintf(nodeName, "%S(%S)", contextP->nodeName,
							contextP->nodeLabel ? contextP->nodeLabel : C.NULL_LITERAL);

					threadMutexUnlock(postitChainMutex);

					cciThrowException(0, 
						   CCI_CONFIGURATION_EXCEPTION,
						   __FILE__,
						   __LINE__,
						   functionName,
						   C.PLUGIN_MESSAGE_CATALOGUE,
						   MJB_INVALID_COPY_FROM_SYNTAX_ELEMENT_NAME,  /* messageNumber */ 
						   msgBuffer,              /* traceText */
#if !defined(WIN32)
						   brokerName,
#endif
						   propertyName,
						   nodeName,
						   (char *)0);
				    TRACE_RETURN(RSN_013, TT_INVALID_ATTRIBUTE_INITIALISATION_VALUE_SPECIFIED);
				}

				/* copy string into deployed postit, freeing any default value first */
				if (deployedPostitCreateP->copyFromLocation)
				{
					free(deployedPostitCreateP->copyFromLocation);
				}
				deployedPostitCreateP->copyFromLocation = UCdup(copyString);
			}
		}

		/* process the PostitScope property */
		if (!UCcmp(attributeName, C.ATTR_POSTIT_SCOPE))
		{
			scopeStringLength = sizeof(scopeString);
			if (getStringAttribute(contextP, 
					               C.ATTR_POSTIT_SCOPE,
								   &scopeStringLength,
					               scopeString) == TT_SUCCESS)
			{
				/* set enum value according to specified match on string */
				if (!UCcmp(scopeString, C.SCOPE_VALUE_EG))
				{
					deployedPostitCreateP->scope = SCOPE_EXECUTION_GROUP;
			        TRACE_MID(TRACE_INFO, ("scope set to SCOPE_EXECUTION_GROUP"));
				}
				else
				{
					if (!UCcmp(scopeString, C.SCOPE_VALUE_BROKER))
					{
						deployedPostitCreateP->scope = SCOPE_BROKER;
						TRACE_MID(TRACE_INFO, ("scope set to SCOPE_BROKER"));
					}
					else
					{
						deployedPostitCreateP->scope = SCOPE_PLATFORM;
						TRACE_MID(TRACE_INFO, ("scope set to SCOPE_PLATFORM"));
					}

					retCode = attachSharedMemoryIfRequired(contextP,
						                                   deployedPostitCreateP->scope);
					if (retCode != TT_SUCCESS)
					{
						/* raise an exception */
						TRACE_MID(TRACE_ERROR, ("Failed to attach shared memory segment"));
						/* generate an exception here... */
						cciThrowException(0,
										  CCI_FATAL_EXCEPTION,
										  __FILE__,
										  __LINE__,
										  functionName,
										  C.PLUGIN_MESSAGE_CATALOGUE,
										  MJB_FAILED_TO_ATTACH_SHARED_MEMORY,  /* messageNumber */ 
										  "Failed to attach shared memory segment", /* traceText */
#if !defined(WIN32)
										  brokerName,
#endif
										  (char *)0);

					}
				}
			}
		}

		/* process the TimeoutSeconds property */
		if (!UCcmp(attributeName, C.ATTR_TIMEOUT_SECONDS))
		{
			if (getIntegerAttribute(contextP, 
					                C.ATTR_TIMEOUT_SECONDS,
					                &timeoutSeconds) == TT_SUCCESS)
			{
				/* copy timeout value into deployed postit - 0=no timeout */
				if (!timeoutSeconds)
				{
					deployedPostitCreateP->timeOutInterval = INFINITE;
			        TRACE_MID(TRACE_INFO, ("deployed postit timeout set to INFINTE"));
				}
				else
				{
					deployedPostitCreateP->timeOutInterval = timeoutSeconds;
				}
			}
		}

		if (!UCcmp(attributeName, C.ATTR_NUM_APPLICATIONS))
		{
			if (getIntegerAttribute(contextP, 
					                C.ATTR_NUM_APPLICATIONS,
					                &numApplications) == TT_SUCCESS)
			{
				/* copy number of applications value into deployed postit */
				deployedPostitCreateP->numApplications = numApplications;
			}
		}

	}
	else
	if (contextP->nodeType == POSTIT_APPLY)
	{
		deployedPostitApplyP = (deployedPostitApply_t *)(contextP->nodeDataP);

		/* initialise the deployed postit apply structure */

		/* deal with postitName */
		if (!UCcmp(attributeName, C.ATTR_POSTIT_NAME))
		{
			copyStringLength = sizeof(copyString);
			if (getStringAttribute(contextP, 
					               C.ATTR_POSTIT_NAME,
								   &copyStringLength,
					               copyString) == TT_SUCCESS)
			{
				/* postit name is an optional property, so check for null string */
				if (UClen(copyString))
				{
					/* free default value first */
					if (deployedPostitApplyP->postitName)
					{
						free(deployedPostitApplyP->postitName);
					}
					/* name was supplied so copy string into deployed postit */
					deployedPostitApplyP->postitName = UCdup(copyString);
				}
			}
		}

		/* deal with CopyDataTo property */
		if (!UCcmp(attributeName, C.ATTR_COPY_DATA_TO))
		{
			copyStringLength = sizeof(copyString);
			if (getStringAttribute(contextP, 
					               C.ATTR_COPY_DATA_TO,
								   &copyStringLength,
					               copyString) == TT_SUCCESS)
			{
				/* 
				 * String must begin with 'Root.' - anything else is invalid.
				 */
				if (!(UClen(copyString) > 5 && !UCncmp(copyString, C.ROOT_DOT, UClen(C.ROOT_DOT))))
				{
					/* throw an exception here...*/
					UCsprintf(msgBuffer, "The syntax element name specified in the %S "
						               "property of node %S(%S) is invalid. "
									   "Valid names must begin with 'Root.'.",
									   C.ATTR_COPY_DATA_TO,
									   contextP->nodeName,
									   contextP->nodeLabel ? contextP->nodeLabel : C.NULL_LITERAL);

					TRACE_MID(TRACE_ERROR, ("%s", msgBuffer));

					UCsprintf(propertyName, "%S", C.ATTR_COPY_DATA_TO);

					UCsprintf(nodeName, "%S(%S)", contextP->nodeName,
							contextP->nodeLabel ? contextP->nodeLabel : C.NULL_LITERAL);

					threadMutexUnlock(postitChainMutex);

					cciThrowException(0, 
						   CCI_CONFIGURATION_EXCEPTION,
						   __FILE__,
						   __LINE__,
						   functionName,
						   C.PLUGIN_MESSAGE_CATALOGUE,
						   MJB_INVALID_COPY_TO_SYNTAX_ELEMENT_NAME,  /* messageNumber */ 
						   msgBuffer,              /* traceText */
#if !defined(WIN32)
						   brokerName,
#endif
						   propertyName,
						   nodeName,
						   (char *)0);
				    TRACE_RETURN(RSN_013, TT_INVALID_ATTRIBUTE_INITIALISATION_VALUE_SPECIFIED);
				}

				/* copy string into deployed postit, freeing any default value first */
				if (deployedPostitApplyP->copyToLocation)
				{
					free(deployedPostitApplyP->copyToLocation);
				}
				deployedPostitApplyP->copyToLocation = UCdup(copyString);
			}
		}

		/* process the PostitScope property */
		if (!UCcmp(attributeName, C.ATTR_POSTIT_SCOPE))
		{
			scopeStringLength = sizeof(scopeString);
			if (getStringAttribute(contextP, 
					               C.ATTR_POSTIT_SCOPE,
								   &scopeStringLength,
					               scopeString) == TT_SUCCESS)
			{
				/* set enum value according to specified match on string */
				if (!UCcmp(scopeString, C.SCOPE_VALUE_EG))
				{
					deployedPostitApplyP->scope = SCOPE_EXECUTION_GROUP;
			        TRACE_MID(TRACE_INFO, ("scope set to SCOPE_EXECUTION_GROUP"));
				}
				else
				{
					if (!UCcmp(scopeString, C.SCOPE_VALUE_BROKER))
					{
						deployedPostitApplyP->scope = SCOPE_BROKER;
						TRACE_MID(TRACE_INFO, ("scope set to SCOPE_BROKER"));
					}
					else
					{
						deployedPostitApplyP->scope = SCOPE_PLATFORM;
						TRACE_MID(TRACE_INFO, ("scope set to SCOPE_PLATFORM"));
					}

					retCode = attachSharedMemoryIfRequired(contextP,
						                                   deployedPostitApplyP->scope);
					if (retCode != TT_SUCCESS)
					{
						/* raise an exception */
						TRACE_MID(TRACE_ERROR, ("Failed to attach shared memory segment"));
						/* generate an exception here... */
						cciThrowException(0,
										  CCI_FATAL_EXCEPTION,
										  __FILE__,
										  __LINE__,
										  functionName,
										  C.PLUGIN_MESSAGE_CATALOGUE,
										  MJB_FAILED_TO_ATTACH_SHARED_MEMORY,  /* messageNumber */ 
										  "Failed to attach shared memory segment", /* traceText */
#if !defined(WIN32)
										  brokerName,
#endif
										  (char *)0);

					}
				}
			}
		}

		/* process the MatchOn property */
		if (!UCcmp(attributeName, C.ATTR_MATCH_ON))
		{
			matchStringLength = sizeof(matchString);
			if (getStringAttribute(contextP, 
					               C.ATTR_MATCH_ON,
								   &matchStringLength,
					               matchString) == TT_SUCCESS)
			{

				/* set enum value according to specified match on string */
				if (!UCcmp(matchString, C.MATCH_CORRELID_TO_MSGID))
				{
					deployedPostitApplyP->matchValue = MATCH_TYPE_CORRELID_MSGID;
			        TRACE_MID(TRACE_INFO, ("matchValue set to MATCH_TYPE_CORRELID_MSGID"));
				}
				else
				if (!UCcmp(matchString, C.MATCH_MSGID_TO_MSGID))
				{
					deployedPostitApplyP->matchValue = MATCH_TYPE_MSGID_MSGID;
			        TRACE_MID(TRACE_INFO, ("matchValue set to MATCH_TYPE_MSGID_MSGID"));
				}
				else
				if (!UCcmp(matchString, C.MATCH_CORRELID_TO_CORRELID))
				{
					deployedPostitApplyP->matchValue = MATCH_TYPE_CORRELID_CORRELID;
			        TRACE_MID(TRACE_INFO, ("matchValue set to MATCH_TYPE_CORRELID_CORRELID"));
				}
				else
				if (!UCcmp(matchString, C.MATCH_MSGID_TO_CORRELID))
				{
					deployedPostitApplyP->matchValue = MATCH_TYPE_MSGID_CORRELID;
			        TRACE_MID(TRACE_INFO, ("matchValue set to MATCH_TYPE_MSGID_CORRELID"));
				}
				else
				if (!UCcmp(matchString, C.MATCH_BOTHIDS))
				{
					deployedPostitApplyP->matchValue = MATCH_TYPE_BOTH;
			        TRACE_MID(TRACE_INFO, ("matchValue set to MATCH_TYPE_BOTH"));
				}
				else
				{
					deployedPostitApplyP->matchValue = MATCH_TYPE_ALL;
			        TRACE_MID(TRACE_INFO, ("matchValue set to MATCH_TYPE_ALL"));
				}
			}
		}
	}
	else
	if (contextP->nodeType == POSTIT_DELETE)
	{
		deployedPostitDeleteP = (deployedPostitDelete_t *)(contextP->nodeDataP);

		/* initialise the deployed postit delete structure */

		/* deal with postitName */
		if (!UCcmp(attributeName, C.ATTR_POSTIT_NAME))
		{
			copyStringLength = sizeof(copyString);
			if (getStringAttribute(contextP, 
					               C.ATTR_POSTIT_NAME,
								   &copyStringLength,
					               copyString) == TT_SUCCESS)
			{
				/* postit name is an optional property, so check for null string */
				if (UClen(copyString))
				{
					/* free default value first */
					if (deployedPostitDeleteP->postitName)
					{
						free(deployedPostitDeleteP->postitName);
					}
					/* name was supplied so copy string into deployed postit */
					deployedPostitDeleteP->postitName = UCdup(copyString);
				}
			}
		}

		/* process the CopyDataFrom property */
		if (!UCcmp(attributeName, C.ATTR_COPY_DATA_FROM))
		{
			copyStringLength = sizeof(copyString);
			if (getStringAttribute(contextP, 
					               C.ATTR_COPY_DATA_FROM,
								   &copyStringLength,
					               copyString) == TT_SUCCESS)
			{
				/* 
				 * String can be 'Root' or it must begin with 'Root.' - anything else
				 * is invalid.
				 */
				if (!(UClen(copyString) == 4 && !UCcmp(copyString, C.ROOT) ||
				      UClen(copyString) > 5 && !UCncmp(copyString, C.ROOT_DOT, UClen(C.ROOT_DOT))))
				{
					/* throw an exception here...*/
					UCsprintf(msgBuffer, "The syntax element name specified in the %S "
						               "property of node %S(%S) is invalid. "
									   "Valid names must be either 'Root' or must begin "
									   "with 'Root.'.",
									   C.ATTR_COPY_DATA_FROM,
									   contextP->nodeName,
									   contextP->nodeLabel ? contextP->nodeLabel : C.NULL_LITERAL);

					TRACE_MID(TRACE_ERROR, ("%s", msgBuffer));

					UCsprintf(propertyName, "%S", C.ATTR_COPY_DATA_FROM);

					UCsprintf(nodeName, "%S(%S)", contextP->nodeName,
							contextP->nodeLabel ? contextP->nodeLabel : C.NULL_LITERAL);

					threadMutexUnlock(postitChainMutex);

					cciThrowException(0, 
						   CCI_CONFIGURATION_EXCEPTION,
						   __FILE__,
						   __LINE__,
						   functionName,
						   C.PLUGIN_MESSAGE_CATALOGUE,
						   MJB_INVALID_COPY_FROM_SYNTAX_ELEMENT_NAME,  /* messageNumber */ 
						   msgBuffer,              /* traceText */
#if !defined(WIN32)
						   brokerName,
#endif
						   propertyName,
						   nodeName,
						   (char *)0);
				    TRACE_RETURN(RSN_013, TT_INVALID_ATTRIBUTE_INITIALISATION_VALUE_SPECIFIED);
				}

				/* copy string into deployed postit, freeing any default value first */
				if (deployedPostitDeleteP->copyFromLocation)
				{
					free(deployedPostitDeleteP->copyFromLocation);
				}
				deployedPostitDeleteP->copyFromLocation = UCdup(copyString);
			}
		}

		/* process the PostitScope property */
		if (!UCcmp(attributeName, C.ATTR_POSTIT_SCOPE))
		{
			scopeStringLength = sizeof(scopeString);
			if (getStringAttribute(contextP, 
					               C.ATTR_POSTIT_SCOPE,
								   &scopeStringLength,
					               scopeString) == TT_SUCCESS)
			{
				/* set enum value according to specified match on string */
				if (!UCcmp(scopeString, C.SCOPE_VALUE_EG))
				{
					deployedPostitDeleteP->scope = SCOPE_EXECUTION_GROUP;
			        TRACE_MID(TRACE_INFO, ("scope set to SCOPE_EXECUTION_GROUP"));
				}
				else
				{
					if (!UCcmp(scopeString, C.SCOPE_VALUE_BROKER))
					{
						deployedPostitDeleteP->scope = SCOPE_BROKER;
						TRACE_MID(TRACE_INFO, ("scope set to SCOPE_BROKER"));
					}
					else
					{
						deployedPostitDeleteP->scope = SCOPE_PLATFORM;
						TRACE_MID(TRACE_INFO, ("scope set to SCOPE_PLATFORM"));
					}

					retCode = attachSharedMemoryIfRequired(contextP,
						                                   deployedPostitDeleteP->scope);
					if (retCode != TT_SUCCESS)
					{
						/* raise an exception */
						TRACE_MID(TRACE_ERROR, ("Failed to attach shared memory segment"));
						/* generate an exception here... */
						cciThrowException(0,
										  CCI_FATAL_EXCEPTION,
										  __FILE__,
										  __LINE__,
										  functionName,
										  C.PLUGIN_MESSAGE_CATALOGUE,
										  MJB_FAILED_TO_ATTACH_SHARED_MEMORY,  /* messageNumber */ 
										  "Failed to attach shared memory segment", /* traceText */
#if !defined(WIN32)
										  brokerName,
#endif
										  (char *)0);

					}
				}
			}
		}
	}

	TRACE_MID(TRACE_INFO, ("About to unlock postitChainMutex"));
	threadMutexUnlock(postitChainMutex);
								  
    TRACE_RETURN(RSN_013, TT_SUCCESS);
}

/******************************************************************************/
/* evaluate                                                                   */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  destinationList (input) - address of the input destination list for the   */
/*                            message                                         */
/*  exceptionList   (input) - address of the exception list for the message   */
/*  message         (input) - address of the input message object             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Performs node processing. This function contains the main processing logic */
/* of the message flow node. It is invoked by the broker when a message is    */
/* received on one of the input terminals of an instance of a node object.    */
/*                                                                            */
/* This function has 4 responsibilities:                                      */
/*                                                                            */
/* 1. Process the message in accordance with the values of any attribute on   */
/*    the node instance.                                                      */
/* 2. Process the message based on content if required.                       */
/* 3. Propagate the message to the appropriate output terminal                */
/* 4. Throw an exception if errors occur.                                     */
/*                                                                            */
/******************************************************************************/
void evaluate(CciContext *dataP,
		      CciMessage *destinationList,
		      CciMessage *exceptionList,
		      CciMessage *message)
{
	context_t *contextP;
	int rc;
	static char *functionName = "evaluate";
	char msgBuffer[200],
		 nodeName[100];

    TRACE_ENTRY("evaluate", ("dataP=%.8x, destinationList=%.8x, "
		                     "exceptionList=%.8x, message=%.8x",
							 dataP, destinationList,
							 exceptionList, message));

	contextP = (context_t *)dataP;

	/* determine which function to call based on node name */
    TRACE_MID(TRACE_INFO, ("nodeName='%S'", contextP->nodeName));
	if (!UCcmp(contextP->nodeName, C.POSTIT_CREATE_NODE))
	{
        TRACE_MID(TRACE_INFO, (">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"));
        TRACE_MID(TRACE_INFO, (">>>>>>>>>>> PostitCreate >>>>>>>>>>>>"));
        TRACE_MID(TRACE_INFO, (">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"));
		rc = evaluatePostitCreate(contextP,
			                      destinationList,
							      exceptionList,
							      message);
	}
	else
	if (!UCcmp(contextP->nodeName, C.POSTIT_APPLY_NODE))
	{
        TRACE_MID(TRACE_INFO, ("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<"));
        TRACE_MID(TRACE_INFO, ("<<<<<<<<<<< PostitApply <<<<<<<<<<<<<"));
        TRACE_MID(TRACE_INFO, ("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<"));
		rc = evaluatePostitApply(contextP,
			                     destinationList,
							     exceptionList,
							     message);
	}
	else
	if (!UCcmp(contextP->nodeName, C.POSTIT_DELETE_NODE))
	{
        TRACE_MID(TRACE_INFO, ("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<"));
        TRACE_MID(TRACE_INFO, ("<<<<<<<<<< PostitDelete <<<<<<<<<<<<<"));
        TRACE_MID(TRACE_INFO, ("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<"));
		rc = evaluatePostitDelete(contextP,
			                     destinationList,
							     exceptionList,
							     message);
	}
	else
	{
        TRACE_MID(TRACE_ERROR, ("Unrecognised node"));
		/* throw an exception here...*/
		UCsprintf(msgBuffer, "Unrecognised node ('%S') passed to evaluate()",
			               contextP->nodeName);
		UCsprintf(nodeName, "'%S'", contextP->nodeName);
		cciThrowException(&rc, 
		                  CCI_CONFIGURATION_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      C.PLUGIN_MESSAGE_CATALOGUE,
			              MJB_UNRECOGNISED_NODE,  /* messageNumber */ 
			              msgBuffer,              /* traceText */
#if !defined(WIN32)
						  brokerName,
#endif
						  nodeName,
			              (char *)0);
	}

	// check response from evaluate subroutines and throw exception if needed.

    TRACE_RETURN_VOID(RSN_016);
}

/******************************************************************************/
/* bipGetMessageflowNodeFactory                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  Address of factory object.                                                */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* This function is called when the message broker loads the "lil" during the */
/* initialization of the broker engine                                        */
/*                                                                            */
/* This function has 5 responsibilities:                                      */
/*                                                                            */
/* 1. create the node factory (or factories) which the "lil" supports, using  */
/*    the createNodeFactory() function, saving the returned pointer to the    */
/*    node factory for use at runtime.                                        */
/* 2. define each node supported by the (or each) factory) and pass a pointer */
/*    to a function table which contains pointers to the node implementation  */
/*    functions.                                                              */
/* 3. define the name of each attribute supported by each node, including a   */
/*    pointer to a validation function.                                       */
/* 4. define input and output terminals supported by each node.               */
/* 5. return the pointer to the factory, unless an error occurred, in which   */
/*    case zero should be returned.                                           */    
/*                                                                            */
/******************************************************************************/
#if defined(WIN32)
    #include <process.h>
#elif defined(_AIX) || defined(_SOLARIS) || defined(_HPUX)
    #include <unistd.h>
#endif
#if defined(WIN32)
void __declspec(dllexport) *bipGetMessageflowNodeFactory()
#else
CciFactory LilFactoryExportPrefix * LilFactoryExportSuffix bipGetMessageflowNodeFactory()
#endif
{
    static char *functionName="bipGetMessageflowNodeFactory()";
    void* factoryObject = NULL;
    CciChar *factoryName = C.POSTIT_PLUGIN_FACTORY;
#if defined(POST_201_VERSION)
	CCI_EXCEPTION_ST exception_st={CCI_EXCEPTION_ST_DEFAULT};
#else
    CCI_EXCEPTION_ST exception_st;
#endif
    int rc;
#if defined(POST_201_VERSION)
	static CNI_VFT vftable={CNI_VFT_DEFAULT};
#else
    static CNI_VFT vftable;
#endif
	bool_t resourceCreated = TRUE;
	char *resourceName,
		 msgBuffer[200];

    TRACE_ENTRY("bipGetMessageflowNodeFactory", ("void"));
    currentPID = getpid();
    TRACE_MID(TRACE_INFO, ("<------------------------------------------------------>"));
    TRACE_MID(TRACE_INFO, ("<**** Start of postit deployment (pid=%d) ****>",
		                   currentPID));
    TRACE_MID(TRACE_INFO, ("<**** Patch level %s ****>", patchLevel));
    TRACE_MID(TRACE_INFO, ("<------------------------------------------------------>"));
	/* create the global syncronisation resources we need here */

	/* create required mutexes */
	if (resourceCreated)
	{
        TRACE_MID(TRACE_INFO, ("About to create postitChainMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &postitChainMutex) != THREAD_OK)
		{
			resourceName = "postitChainMutex";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create alarmThreadSyncMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &alarmThreadSyncMutex) != THREAD_OK)
		{
			resourceName = "alarmThreadSyncMutex";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create shmemInitTermMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &shmemInitTermMutex) != THREAD_OK)
		{
			resourceName = "shmemInitTermMutex";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create shmemThreadWorkMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &shmemThreadWorkMutex) != THREAD_OK)
		{
			resourceName = "shmemInitTermMutex";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create shmemThreadSyncMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &shmemThreadSyncMutex) != THREAD_OK)
		{
			resourceName = "shmemThreadSyncMutex";
			resourceCreated = FALSE;
		}
	}

	/* create required condition variables */
	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create alarmThreadWorkCondVar"));
		if (threadCreateCondVar(postitChainMutex,
								&alarmThreadWorkCondVar) != THREAD_OK)
		{
			resourceName = "alarmThreadWorkCondVar";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create alarmThreadSyncCondVar"));
		if (threadCreateCondVar(alarmThreadSyncMutex,
								&alarmThreadSyncCondVar) != THREAD_OK)
		{
			resourceName = "alarmThreadSyncCondVar";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create shmemThreadWorkCondVar"));
		if (threadCreateCondVar(shmemThreadWorkMutex,
								&shmemThreadWorkCondVar) != THREAD_OK)
		{
			resourceName = "shmemThreadWorkCondVar";
			resourceCreated = FALSE;
		}
	}

	if (resourceCreated)
	{
		TRACE_MID(TRACE_INFO, ("About to create shmemThreadSyncCondVar"));
		if (threadCreateCondVar(shmemThreadSyncMutex,
								&shmemThreadSyncCondVar) != THREAD_OK)
		{
			resourceName = "shmemThreadSyncCondVar";
			resourceCreated = FALSE;
		}
	}

	/* process any errors that have occurred */
	if (!resourceCreated)
	{
		sprintf(msgBuffer, "Postit Plugin Factory failed to create resource %s",
						   resourceName);
		/* generate an exception here... */
		cciThrowException(&rc, 
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      C.PLUGIN_MESSAGE_CATALOGUE,
			              MJB_RESOURCE_CREATION_FAILED,  /* messageNumber */ 
			              msgBuffer, /* traceText */
#if !defined(WIN32)
						  brokerName,
#endif
			              "PostitPlugInFactory",
						  resourceName,
			              (char *)0);
	}


    /* Setup virtual function table with addresses of implementation functions */
#if !defined(POST_201_VERSION)
    memset(&vftable, 0, sizeof(vftable));
#endif
    vftable.iFpCreateNodeContext     = createNodeContext;
    vftable.iFpDeleteNodeContext     = deleteNodeContext;
    vftable.iFpGetAttributeName      = getAttributeName;
    vftable.iFpSetAttribute          = setAttribute;
    vftable.iFpGetAttribute          = getAttribute;
    vftable.iFpEvaluate              = evaluate;


    /* Create the Node Factory for this plug-in */
    factoryObject = cniCreateNodeFactory(&rc, factoryName);
    if (factoryObject == CCI_NULL_ADDR)
	{
        if (rc == CCI_EXCEPTION)
		{
            cciGetLastExceptionData(&rc, 
				                    &exception_st);
            /* Any local error handling may go here */
            cciRethrowLastException(&rc);
		}
        /* Any further local error handling can go here */
        return(factoryObject);
	}
    /* Define the PostitCreate and PostitApply nodes supported by our factory */

	/*
	 * Define the PostitCreate node
	 */
    cniDefineNodeClass(&rc, 
			           factoryObject,
					   C.POSTIT_CREATE_NODE,
					   &vftable); 
    if (rc != CCI_SUCCESS)
	{
        TRACE_MID(TRACE_ERROR, ("Failed to create PostitCreate node "
			                    "(cniDefineNodeClass rc=%d)", rc));
		/* generate an exception here... */
		cciThrowException(&rc,
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      C.PLUGIN_MESSAGE_CATALOGUE,
			              MJB_NODE_CLASS_CREATION_FAILED,  /* messageNumber */ 
			              "Postit Plugin Factory failed to create PostitCreate node", /* traceText */
#if !defined(WIN32)
						  brokerName,
#endif
			              "PostitPlugInFactory",
			              POSTIT_CREATE_NODE_MB,
			              (char *)0);
	}

	/*
	 * Define the PostitApply node
	 */
    cniDefineNodeClass(&rc, 
			           factoryObject, 
					   C.POSTIT_APPLY_NODE, 
					   &vftable); 
    if (rc != CCI_SUCCESS)
	{
        TRACE_MID(TRACE_ERROR, ("Failed to create PostitApply node "
			                    "(cniDefineNodeClass rc=%d)", rc));
		/* generate an exception here... */
		cciThrowException(&rc,
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      C.PLUGIN_MESSAGE_CATALOGUE,
			              MJB_NODE_CLASS_CREATION_FAILED,  /* messageNumber */ 
			              "Postit Plugin Factory failed to create PostitApply node", /* traceText */
#if !defined(WIN32)
						  brokerName,
#endif
			              "PostitPlugInFactory",
			              POSTIT_APPLY_NODE_MB,
			              (char *)0);
	}

	/*
	 * Define the PostitDelete node
	 */
    cniDefineNodeClass(&rc, 
			           factoryObject, 
					   C.POSTIT_DELETE_NODE, 
					   &vftable); 
    if (rc != CCI_SUCCESS)
	{
        TRACE_MID(TRACE_ERROR, ("Failed to create PostitDelete node "
			                    "(cniDefineNodeClass rc=%d)", rc));
		/* generate an exception here... */
		cciThrowException(&rc,
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      C.PLUGIN_MESSAGE_CATALOGUE,
			              MJB_NODE_CLASS_CREATION_FAILED,  /* messageNumber */ 
			              "Postit Plugin Factory failed to create PostitDelete node", /* traceText */
#if !defined(WIN32)
						  brokerName,
#endif
			              "PostitPlugInFactory",
			              POSTIT_DELETE_NODE_MB,
			              (char *)0);
	}

    /* Log a success message */
    TRACE_MID(TRACE_INFO, ("About to issue cciLog success message"));
    cciLog(&rc,                       /* returnCode */
	       CCI_LOG_INFORMATION,       /* type */
		   __FILE__,                  /* file */
		   __LINE__,                  /* line */
		   functionName,              /* function */
		   C.PLUGIN_MESSAGE_CATALOGUE,  /* messageSource */
#if defined(WIN32)
           MJB_NODE_FACTORY_CREATED,  /* messageNumber */ 
#else
           MJB_NODE_FACTORY_CREATED_UNIX,  /* messageNumber */ 
#endif
		   "Postit Plugin Factory loading", /* traceText */
		   "PostitPlugInFactory",
		   (char *)0);

    /* Return address of this factory object to the broker */
    TRACE_RETURN_TYPE(RSN_021, void *, factoryObject);
}


/* local functions */


/******************************************************************************/
/* attachSharedMemoryIfRequired                                               */
/******************************************************************************/
int attachSharedMemoryIfRequired(IN context_t *contextP,
								 IN scope_t scope)
{
	int shmSize,
		retCode;
	globalCB_t dummyCB = {GLOBAL_CB_DEFAULT};
	executionGroupEntry_t *currentEGP;
    static char *functionName="attachSharedMemoryIfRequired";
	char ownerKey[MAX_STRING_SIZE + 1],
	     segNameKey[MAX_STRING_SIZE + 1];
	bool_t scopeIsBroker;
#if !defined(WIN32)
	unsigned long *numericKeyP;
	int	i;
#endif

    TRACE_ENTRY("attachSharedMemoryIfRequired", ("contextP=%.8x, scope=%s", 
		                                          contextP,
												  scope==SCOPE_BROKER ? "Broker" :
	                                              "Platform"));

    scopeIsBroker = (scope == SCOPE_BROKER);

	TRACE_MID(TRACE_INFO, ("About to lock shared memory mutex"));
	threadMutexLock(shmemInitTermMutex);
	if (scopeIsBroker && !sharedMemoryAttached_Broker ||
	    !scopeIsBroker && !sharedMemoryAttached_Platform)
	{
		shmSize = getSharedSegmentSize(sharedSegmentEnvironmentVariableName,
							           SHARED_SEGMENT_DEFAULT_KSIZE); 
#if defined(WIN32)
		strcpy(segNameKey, SHARED_MEMORY_SEGMENT_NAME_PREFIX);
		if (scopeIsBroker)
		{
			strcat(segNameKey, brokerName);
		}
		else
		{
			strcat(segNameKey, "Platform");
		}
		TRACE_MID(TRACE_INFO, ("Segment name key is: %s", segNameKey));
#else
		strcpy(segNameKey, SHARED_SEGMENT_KEY);
		numericKeyP = (unsigned long *)segNameKey;
		TRACE_MID(TRACE_INFO, ("*numericKeyP value is %.8x", *numericKeyP));
		if (scopeIsBroker)
		{
			// hash the broker name and incorporate into the segment key
			// need to find a good hashing algorithm to replace this temporary one
			for (i=0; i<(int)strlen(brokerName); i++)
			{
				*numericKeyP ^= (int)brokerName[i] << 8;
				*numericKeyP += (int)brokerName[i];
				*numericKeyP <<= 1;
				TRACE_MID(TRACE_INFO, ("*numericKeyP value is %.8x", *numericKeyP));
			}
		}
#endif

		/* build owner key for broker scope */
		if (scopeIsBroker)
		{
			strcpy(ownerKey, SHARED_SEGMENT_BROKER_SCOPE_KEY_PREFIX);
			strcat(ownerKey, brokerName);
			TRACE_MID(TRACE_INFO, ("Owner key is: %s", ownerKey));
		}
		TRACE_MID(TRACE_INFO, ("About to attempt to request/attach shared "
			                   "memory segment '%s' of %d bytes",
							    segNameKey,
								shmSize));

		retCode = shmAttach(segNameKey,
							shmSize,
							(void *)&dummyCB,
							sizeof(dummyCB),
							currentPID,
							scopeIsBroker ? ownerKey : SHARED_SEGMENT_PLATFORM_SCOPE_KEY,
							scopeIsBroker);

		if (retCode != TT_SUCCESS)
		{
			/* deal with error, including unlocking mutex */
			TRACE_MID(TRACE_ERROR, ("Failed to attach shared memory"));

			TRACE_MID(TRACE_INFO, ("About to unlock shared memory mutex"));
			threadMutexUnlock(shmemInitTermMutex);
			TRACE_RETURN(RSN_001, TT_FAILED_TO_ATTACH_GLOBAL_STORAGE);
		}

		if (scopeIsBroker)
		{
			sharedMemoryCBP_Broker = (globalCB_t *)shmGetCBAddress(TRUE);
			sharedMemoryAttached_Broker = TRUE;
		}
		else
		{
			sharedMemoryCBP_Platform = (globalCB_t *)shmGetCBAddress(FALSE);
			sharedMemoryAttached_Platform = TRUE;
		}

		/* 
		 * Tidy up shared memory if any previous instance of this execution group failed without
		 * properly removing its resources.
		 */

		/* lock shared memory */
		TRACE_MID(TRACE_INFO, ("Taking shared memory lock for %s", 
			      scopeIsBroker ? "Broker" : "Platform"));
		shmLock(scopeIsBroker);

		// test locking...sleep for a while and wait for other process to try lock
		// TRACE_MID(TRACE_INFO, ("Sleeping to test global lock"));
		// SLEEP_THREAD_MILLISECONDS(20);  // sleep for 20 seconds

		currentEGP = findExecutionGroupEntry(scopeIsBroker);

		if (currentEGP)
		{
			// make sure Work In Progress (WIP) flag is reset in shared memory 
		    shmResetWIP(currentEGP->pid, TRUE, scopeIsBroker);

			removeEGPostitsFromGlobalStorage(currentEGP,
											 scopeIsBroker);

			/* remove stale pid from pid list */
			retCode = shmRemovePidFromList(currentEGP->pid,
										   TRUE,  // we've got memory locked
										   scopeIsBroker,  // Broker scope
				   						   NULL);
		}

		/* 
		 * If an execution group entry already exists, then store our pid in it, otherwise create 
		 * a new entry.
		 */
		if (!currentEGP)
		{
			currentEGP = createNewExecutionGroupEntry(scopeIsBroker);
		}

		/* store our pid in the entry */
		currentEGP->pid = currentPID;

		/* unlock shared memory */
		shmUnlock(scopeIsBroker);

		/* 
		 * We also need to start a background thread to do garbage collection for the areas
		 * we allocate in the shared memory space.
		 */
		if (!shmemThreadCreated)
		{
			retCode = createSharedMemoryGarbageCollectionThread(contextP);

			if (retCode != TT_SUCCESS)
			{
				TRACE_MID(TRACE_INFO, ("About to unlock shared memory mutex"));
				threadMutexUnlock(shmemInitTermMutex);
				TRACE_RETURN(RSN_005, TT_FAILED_TO_CREATE_GLOBAL_STORAGE_GARBAGE_COLLECTOR);
			}
		}
	}

	TRACE_MID(TRACE_INFO, ("About to unlock shared memory mutex"));
	threadMutexUnlock(shmemInitTermMutex);

	TRACE_RETURN(RSN_001, TT_SUCCESS);
}


/******************************************************************************/
/* findExecutionGroupEntry                                                    */
/******************************************************************************/
executionGroupEntry_t *findExecutionGroupEntry(IN bool_t scopeIsBroker)
{
    executionGroupEntry_t *currentEGP = NULL;
	char fullyQualifiedName[MAX_STRING_SIZE * 2 + 1 + 1];
    TRACE_ENTRY("findExecutionGroupEntry", ("scopeIsBroker=%s", 
		                                     scopeIsBroker ? "TRUE" : "FALSE"));

	/* on entry, assume caller has shared memory locked */

	/* see if this execution group already exists */
	currentEGP = SHM_HANDLE2ADDRESS((scopeIsBroker ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->executionGroupAnchor.nextEntry, 
		                            executionGroupEntry_t *, 
									scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

	/* build the fully qualified EG name by prepending with broker name and underscore */
	strcpy(fullyQualifiedName, brokerName);
	strcat(fullyQualifiedName, "_");
	strcat(fullyQualifiedName, executionGroupName);

	TRACE_MID(TRACE_INFO, ("Looking for execution group entry with fully qualified name '%s'",
			                fullyQualifiedName));
	while(currentEGP)
	{
		TRACE_MID(TRACE_INFO, ("Processing entry at %.8x (handle=%.8x)",
			                   currentEGP, SHM_ADDRESS2HANDLE(currentEGP,
							   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
		if (!strcmp(fullyQualifiedName,
			        SHM_HANDLE2ADDRESS(currentEGP->name, char *,
					                   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)))
		{
			TRACE_MID(TRACE_INFO, ("Found a match containing pid %d", currentEGP->pid));
			break;
		}

		/* address the next in the list */
		currentEGP = SHM_HANDLE2ADDRESS(currentEGP->nextEntry, executionGroupEntry_t *,
			                            scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	}

	TRACE_RETURN_TYPE(RSN_001, executionGroupEntry_t *, currentEGP);
}


/******************************************************************************/
/* removeEGPostitsFromGlobalStorage                               */
/******************************************************************************/
void removeEGPostitsFromGlobalStorage(IN executionGroupEntry_t *currentEGP,
									  IN bool_t scopeIsBroker)
{
	globalPostitCreateInstance_t *currentGlobalInstanceP,
								 *nextGlobalInstanceP;
    static char *functionName="removeEGPostitsFromGlobalStorage";
	int rc;
    TRACE_ENTRY("removeEGPostitsFromGlobalStorage", ("currentEGP=%.8x, scopeIsBroker=%s",
		                                             currentEGP,
													 scopeIsBroker ? "TRUE" : "FALSE"));

	/* run down global chains and remove any postits owned by dead pid */

	// process Broker scope chain
	currentGlobalInstanceP = 
		SHM_HANDLE2ADDRESS((scopeIsBroker ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor.nextGlobalMsgIdP,
			               globalPostitCreateInstance_t *,
						   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("Checking global MsgId chain"));
	while(currentGlobalInstanceP)
	{
		TRACE_MID(TRACE_INFO, ("Checking for dead global entry at %.8x (handle=%.8x",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP,
									                   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
		nextGlobalInstanceP = SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalMsgIdP, 
				                                 globalPostitCreateInstance_t *,
												 scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

		if (currentGlobalInstanceP->executionGroupPID = currentEGP->pid)
		{
			TRACE_MID(TRACE_INFO, ("Freeing dead postit at %.8x (handle=%.8x)",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP,
									                   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
			/* remove from global chains and free global storage */
			unchainPostit((postitCreateInstance_t *)currentGlobalInstanceP);
			rc = shmFree(currentGlobalInstanceP, TRUE, scopeIsBroker);
	        CHECK_SHMFREE_ERROR(rc, 
		                        functionName, 
					            platformIsUnix);  
		}
		currentGlobalInstanceP = nextGlobalInstanceP;
	}

	// process Platform scope chain
	currentGlobalInstanceP = 
		SHM_HANDLE2ADDRESS((scopeIsBroker ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor.nextGlobalCorrelIdP,
			               globalPostitCreateInstance_t *,
						   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("Checking global CorrelId chain"));
	while(currentGlobalInstanceP)
	{
		TRACE_MID(TRACE_INFO, ("Checking for dead global entry at %.8x (handle=%.8x",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP,
									                   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
		nextGlobalInstanceP = SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalCorrelIdP, 
				                                 globalPostitCreateInstance_t *,
												 scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

		if (currentGlobalInstanceP->executionGroupPID = currentEGP->pid)
		{
			TRACE_MID(TRACE_INFO, ("Freeing dead postit at %.8x (handle=%.8x)",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP,
									                   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
			/* remove from global chains and free global storage */
			unchainPostit((postitCreateInstance_t *)currentGlobalInstanceP);
			rc = shmFree(currentGlobalInstanceP, TRUE, scopeIsBroker);
	        CHECK_SHMFREE_ERROR(rc, 
		                        functionName, 
					            platformIsUnix);  

		}
		currentGlobalInstanceP = nextGlobalInstanceP;
	}

	currentGlobalInstanceP = 
		SHM_HANDLE2ADDRESS((scopeIsBroker ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalInstanceAnchor.nextGlobalBothP,
			               globalPostitCreateInstance_t *,
						   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("Checking global Both chain"));
	while(currentGlobalInstanceP)
	{
		TRACE_MID(TRACE_INFO, ("Checking for dead global entry at %.8x (handle=%.8x",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP,
									                   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
		nextGlobalInstanceP = SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalBothP, 
				                                 globalPostitCreateInstance_t *,
												 scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

		if (currentGlobalInstanceP->executionGroupPID = currentEGP->pid)
		{
			TRACE_MID(TRACE_INFO, ("Freeing dead postit at %.8x (handle=%.8x)",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP,
									                   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
			/* remove from global chains and free global storage */
			unchainPostit((postitCreateInstance_t *)currentGlobalInstanceP);
			rc = shmFree(currentGlobalInstanceP, TRUE, scopeIsBroker);
	        CHECK_SHMFREE_ERROR(rc, 
		                        functionName, 
					            platformIsUnix);  
		}
		currentGlobalInstanceP = nextGlobalInstanceP;
	}

	/* check the dead postit chain also */
	currentGlobalInstanceP = 
		SHM_HANDLE2ADDRESS((scopeIsBroker ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->globalDeadInstanceAnchor.nextGlobalP,
			               globalPostitCreateInstance_t *,
						   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("Checking global dead postit chain"));
	while(currentGlobalInstanceP)
	{
		TRACE_MID(TRACE_INFO, ("Checking for dead global entry at %.8x (handle=%.8x",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP,
									                   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
		nextGlobalInstanceP = SHM_HANDLE2ADDRESS(currentGlobalInstanceP->nextGlobalP, 
				                                 globalPostitCreateInstance_t *,
												 scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);

		if (currentGlobalInstanceP->executionGroupPID = currentEGP->pid)
		{
			TRACE_MID(TRACE_INFO, ("Freeing dead postit at %.8x (handle=%.8x)",
					                currentGlobalInstanceP,
					                SHM_ADDRESS2HANDLE(currentGlobalInstanceP,
									                   scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));
			/* remove from dead chain and free global storage */
			REMOVE_FROM_GLOBAL_POSTIT_CHAIN(currentGlobalInstanceP,
				                            scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
			rc = shmFree(currentGlobalInstanceP, TRUE, scopeIsBroker);
	        CHECK_SHMFREE_ERROR(rc, 
		                        functionName, 
					            platformIsUnix);  
		}
		currentGlobalInstanceP = nextGlobalInstanceP;
	}

	TRACE_RETURN_VOID(RSN_001);
}


/******************************************************************************/
/* createNewExecutionGroupEntry                                               */
/******************************************************************************/
executionGroupEntry_t *createNewExecutionGroupEntry(IN bool_t scopeIsBroker)
{
	int retCode;
	char *egName;
	char fullyQualifiedName[MAX_STRING_SIZE * 2 + 1 + 1];
    static char *functionName="createNewExecutionGroupEntry";
	executionGroupEntry_t defaultEGEntry = {EXECUTION_GROUP_ENTRY_DEFAULT},
		                  *anchorEGP,
		                  *currentEGP;
    TRACE_ENTRY("createNewExecutionGroupEntry", ("scopeIsBroker=%s",
												  scopeIsBroker ? "TRUE" : "FALSE"));

	/* build the fully qualified EG name by prepending with broker name and underscore */
	strcpy(fullyQualifiedName, brokerName);
	strcat(fullyQualifiedName, "_");
	strcat(fullyQualifiedName, executionGroupName);

	TRACE_MID(TRACE_INFO, ("Fully qualified name is '%s'",
			                fullyQualifiedName));

	currentEGP = (executionGroupEntry_t *)shmMalloc(sizeof(executionGroupEntry_t),
			                                        TRUE,
						                            scopeIsBroker,
						                            &retCode);
    // check if we're SOS - no point in trying to poke gc thread since it doesn't exist
	CHECK_SHMEM_ERROR(retCode, 
		              functionName, 
					  scopeIsBroker, 
			          currentEGP, 
					  executionGroupEntry_t *, 
					  sizeof(executionGroupEntry_t),
				      TRUE,   /* lock is already held */
				      platformIsUnix,
					  FALSE);  // gc thread doesn't exist yet, so retry=FALSE

	TRACE_MID(TRACE_INFO, ("New execution group entry block allocated at %.8x (handle=%.8x)",
			               currentEGP, SHM_ADDRESS2HANDLE(currentEGP, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));

	/* initialise control block to default values */
    memcpy(currentEGP, &defaultEGEntry, sizeof(executionGroupEntry_t));

	egName = (char *)shmMalloc(strlen(fullyQualifiedName) + 1,  // allow for null
			                   TRUE,
					           scopeIsBroker,
					           &retCode);

	CHECK_SHMEM_ERROR(retCode, 
		              functionName, 
					  scopeIsBroker, 
			          egName, 
					  char *, 
					  strlen(fullyQualifiedName) + 1,
				      TRUE,   /* lock is already held */
				      platformIsUnix,
					  FALSE);  // gc thread doesn't exist yet, so retry=FALSE

	TRACE_MID(TRACE_INFO, ("New execution group name buffer allocated at %.8x (handle=%.8x)",
			               egName, SHM_ADDRESS2HANDLE(egName, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform)));

	/* copy the EG name into the buffer */
	strcpy(egName, fullyQualifiedName);

	/* store the name in the control block as a handle value */
	currentEGP->name = SHM_ADDRESS2HANDLE(egName, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("Name handle %.8x stored in execution group entry", 
		                    currentEGP->name));

	/* add the entry to the global chain */
	anchorEGP = &(scopeIsBroker ? sharedMemoryCBP_Broker : sharedMemoryCBP_Platform)->executionGroupAnchor;
	TRACE_MID(TRACE_INFO, ("About to chain entry to anchor at %.8x", anchorEGP));
	INSERT_AFTER_IN_GLOBAL_EG_NAME_CHAIN(anchorEGP, currentEGP, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("New execution group entry added to global list"));

	TRACE_RETURN_TYPE(RSN_001, executionGroupEntry_t *, currentEGP);
}

/******************************************************************************/
/* removeExecutionGroupFromGlobalStorage                                      */
/******************************************************************************/
void removeExecutionGroupFromGlobalStorage(IN bool_t scopeIsBroker)
{
	executionGroupEntry_t *currentEGP = NULL;
	int rc;
	bool_t lastPID;
    static char *functionName="removeExecutionGroupFromGlobalStorage";
    TRACE_ENTRY("removeExecutionGroupFromGlobalStorage", ("scopeIsBroker=%s",
												           scopeIsBroker ? "TRUE" : "FALSE"));

	/* on entry, assume caller has shared memory locked */
	currentEGP = findExecutionGroupEntry(scopeIsBroker);
	if (!currentEGP)
	{
		TRACE_MID(TRACE_INFO, ("No execution group found"));
		TRACE_RETURN_VOID(RSN_001);
	}
	TRACE_MID(TRACE_INFO, ("Execution group entry located at %.8x", currentEGP));
    
	/* unchain execution group entry */
	REMOVE_FROM_GLOBAL_EG_NAME_CHAIN(currentEGP, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform);
	TRACE_MID(TRACE_INFO, ("Execution group entry removed from global chains"));

	/* free the global storage */
	if (currentEGP->name != SHM_NULL_HANDLE)
	{
		rc = shmFree(SHM_HANDLE2ADDRESS(currentEGP->name, void *, scopeIsBroker ? sharedMemorySegment_Broker : sharedMemorySegment_Platform),
 			    TRUE,  // global lock already taken
				scopeIsBroker);
        CHECK_SHMFREE_ERROR(rc, 
	                        functionName, 
				            platformIsUnix);  
		TRACE_MID(TRACE_INFO, ("Execution group name buffer freed"));
	}

    /* remove our pid from pid list */
	rc = shmRemovePidFromList(currentEGP->pid,
							  TRUE,  // we've got memory locked
							  scopeIsBroker,  // Broker scope
	   						  &lastPID);

	if (lastPID) {
		// remove the pid list
		TRACE_MID(TRACE_INFO, ("We are the last process, removing pid list"));
        rc = shmFreePidList(TRUE,  // we've got memory locked already
			                scopeIsBroker);
	}

	// finally, remove the execution group control block 
	rc = shmFree(currentEGP,
			    TRUE,  // global lock already taken
				scopeIsBroker);
    CHECK_SHMFREE_ERROR(rc, 
                        functionName, 
			            platformIsUnix);  

	TRACE_MID(TRACE_INFO, ("Execution group entry control block freed"));

	TRACE_RETURN_VOID(RSN_001);
}


