/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : reply.c                                                  */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : Contains functions for dealing with reply messages       */
/*                   belonging to an aggregate.                               */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Feb 2000                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include <aggregator.h>      /* Specific header for this dll                       */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <BipCni.h>		/* Function prototypes for Plug-in Node functions     */
#include <log.h>


/******************************************************************************/
/* Node Utility Functions                                                     */
/*                                                                            */
/* evaluateAggregateReply                                                     */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/* evaluateAggregateReply                                                     */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  destinationList (input) - address of the input destination list for the   */
/*                            message                                         */
/*  exceptionList   (input) - address of the exception list for the message   */
/*  message         (input) - address of the input message object             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Performs alarm processing.  This function receives the modified alarm input*/
/* message generated by the alarm thread and replaces the correlid with its   */
/* original value.                                                            */
/*                                                                            */
/******************************************************************************/
int evaluateAggregateReply(IN context_t* contextP,
 			   		       IN CciMessage *destinationList,
		                   IN CciMessage *exceptionList,
		                   IN CciMessage *message)
{
	CciMessage *newMsg,
		       *newDestList;
	void *correlIdElementP,
		 *putApplNameElementP;
    struct CciByteArray byteArrayDetails;
	char correlIdArray[MQ_CORREL_ID_LENGTH],
		 *bufferStart;
    CciSize numBytes;
	CciChar putApplName[MQ_PUT_APPL_NAME_LENGTH + 1];

	int rc,
        retCode;
    msgIdIndexElement_t *msgIndexP;
	aggregateMember_t *memberP;
	aggregateInstance_t *aggregateP;
	CciSize bufferSize;
	static char *functionName = "evaluateAggregateReply";
	char msgBuffer[1000],
		 mbString[MAX_STRING_SIZE + 1],
		 hexBuffer[2*MQ_CORREL_ID_LENGTH + 1];
	deployedAggregate_t *depAggregateP;
#if 0
	logTruncationRequest_t *logTruncReqP;
#endif

    TRACE_ENTRY("evaluateAggregateReply", ("contextP=%.8x, destinationList=%.8x, "
		                     "exceptionList=%.8x, message=%.8x",
							 contextP, destinationList,
							 exceptionList, message));

	/* 
	 * Check that the aggregate consists of at least one AggregateCreate node
	 * and at most, one AggregateReply node
	 */

	if (!contextP->nodeDataP || 
		!((deployedAggregate_t *)(contextP->nodeDataP))->aggregateCreateAndReplyExist)
	{
		if (!contextP->nodeDataP)
		{
	        TRACE_MID(TRACE_ERROR, ("Aggregate has not been correctly configured"));

			/* throw an exception */
			sprintf(msgBuffer, "Aggregate has not been correctly configured");

			cciThrowException(&rc,
							  CCI_FATAL_EXCEPTION,
							  __FILE__,
							  __LINE__,
							  functionName,
							  PLUGIN_MESSAGE_CATALOGUE,
							  MJB_INVALID_AGGREGATE_CONFIGURATION, /* messageNumber */ 
							  msgBuffer,           /* traceText */
							  (char *)0);
		}
		else
		{
			depAggregateP = (deployedAggregate_t *)(contextP->nodeDataP);


	        sprintf(msgBuffer, "Aggregate %S is missing either an "
				               "SpAggregateCreate node or an SpAggregateReply "
							   "node", depAggregateP->name);
	        TRACE_MID(TRACE_ERROR, ("%s", msgBuffer));

			/* throw an exception */
			UCToStr(depAggregateP->name,
				    sizeof(mbString),
				    mbString,
				    NULL),
			cciThrowException(&rc,
							  CCI_FATAL_EXCEPTION,
							  __FILE__,
							  __LINE__,
							  functionName,
							  PLUGIN_MESSAGE_CATALOGUE,
							  MJB_MISSING_AGGREGATE_NODE, /* messageNumber */ 
							  msgBuffer,           /* traceText */
#if defined(_AIX)
						      brokerName,
#endif
							  mbString,
							  (char *)0);
		}
	}

	/* extract the correlid */
	correlIdElementP = getMessageElement(message, CORREL_ID);

	if (!correlIdElementP)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Correlation Id not found in message"));
		TRACE_RETURN(RSN_001, TT_CORRELID_NOT_FOUND);
	}

	/* copy correl id, assume it is of type CCI_VALUE_TYPE_BYTEARRAY */
	byteArrayDetails.pointer = (void *)correlIdArray;
	byteArrayDetails.size = sizeof(correlIdArray);
	numBytes = cniElementByteArrayValue(&rc,
										correlIdElementP,
										&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract Correl Id from message"));
		TRACE_RETURN(RSN_002, TT_CORRELID_NOT_FOUND);
	}

	if (traceEnabled)
	{
		dispHex((char *)(correlIdArray),
				MQ_CORREL_ID_LENGTH,
				hexBuffer);
		TRACE_MID(TRACE_INFO, ("CorrelId value is %s", hexBuffer));
	}

	/* extract the PutApplName value */
	putApplNameElementP = getMessageElement(message, PUT_APPL_NAME);

	if (!putApplNameElementP)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("PutApplName not found in message"));
		TRACE_RETURN(RSN_001, TT_PUT_APPL_NAME_NOT_FOUND);
	}

	/* copy PutApplName, assume it is of type CCI_VALUE_TYPE_BYTEARRAY */
	/* PutApplName value is stored in unicode format */
	TRACE_MID(TRACE_INFO, ("Calling cniElementCharacterValue"));
	numBytes = cniElementCharacterValue(&rc,
			                            putApplNameElementP,
								        putApplName,
										sizeof(putApplName));
	if (rc != CCI_SUCCESS)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract PutApplName from message, rc=%d, "
			      "numBytes=%d", rc, numBytes));
		TRACE_RETURN(RSN_002, TT_CORRELID_NOT_FOUND);
	}
	TRACE_MID(TRACE_INFO, ("PutApplName value is '%S'", putApplName));

	/* 
	 * Check if this is a TIMED OUT message.  
	 *
	 * A timed out message has the following (hopefully) unique features...
	 *  - a CorrelId value starting with 'Agg-'
	 *  - a PutApplName value set to 'AggregatorTimerThread'
	 *  - a datalength of 0
	 */
	if (!memcmp(correlIdArray, AGG_TIMER_ID_MB, strlen(AGG_TIMER_ID_MB)) &&
	    !wcscmp(putApplName, TIMER_THREAD_ID))
	{
        retCode = processTimerMessage(contextP,
			                           destinationList,
									   exceptionList,
			                           message);
		TRACE_RETURN(RSN_003, retCode);
	}

	/* search the msg id index for a match on correl id */
	TRACE_MID(TRACE_INFO, ("About to lock aggregateChainMutex"));
	threadMutexLock(aggregateChainMutex);

	msgIndexP = findMsgIndexElement(correlIdArray);

	/* no aggregate found for this message, assume it is has already timed out */
	if (!msgIndexP)
	{
		TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
		threadMutexUnlock(aggregateChainMutex);

		/* assume message has timed out, propagate to the 'late reply' terminal */
		retCode = propagateMessage(contextP,
								   L"lateReplies",
								   message,
								   destinationList,
								   exceptionList,
   								   FALSE); /* not a new message */

		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'lateReplies' terminal"));
			TRACE_RETURN(RSN_004, TT_PROPAGATION_ERROR);
		}

		TRACE_RETURN(RSN_005, TT_SUCCESS);
	}

	/* message reply belongs to an active aggregate */
	memberP = msgIndexP->memberP;
	TRACE_MID(TRACE_INFO, ("Processing aggregate member at %.8x", memberP));

	/* remove the message id index element */
    removeMsgIdIndex(msgIndexP);
	memberP->indexP = NULL;

	/* set reply time for member */
	time(&(memberP->replyArrivalTime));

	/* store message and decrement member count */
	memberP->aggregateInstanceP->outstandingReplies--;
	TRACE_MID(TRACE_INFO, ("Outstanding reply count decremented to %d",
			               memberP->aggregateInstanceP->outstandingReplies));

	/* get the message length */
	bufferSize = cniBufferSize(&rc, message);

	/* serialise syntax tree to buffer if buffer is empty */
	if (!bufferSize)
	{
		cniWriteBuffer(&rc, message);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to serialize message to buffer "
									"(cniWriteBuffer rc=%d", rc));
			TRACE_RETURN(RSN_006, TT_FAILED_TO_SERIALIZE_MESSAGE);
		}
		TRACE_MID(TRACE_INFO, ("Message serialised to buffer"));
		bufferSize = cniBufferSize(&rc, message);
	}

	bufferStart = (char *)cniBufferPointer(&rc, message);
	TRACE_MID(TRACE_INFO, ("Message buffer at %.8x, length %d",
			               bufferStart, bufferSize));
    
	memberP->mqmdP = (char *)malloc(sizeof(MQMD));
	memcpy(memberP->mqmdP, bufferStart, sizeof(MQMD));
	TRACE_MID(TRACE_INFO, ("MQMD stored at %.8x", memberP->mqmdP));
    TRACE_MID(TRACE_INFO, ("STORAGE(3): alloc %.8x", memberP->mqmdP));

	memberP->messageLength = bufferSize - sizeof(MQMD);
	if (memberP->messageLength)
	{
		memberP->messageP = (char *)malloc(memberP->messageLength);
		memcpy(memberP->messageP, bufferStart+sizeof(MQMD),
			   memberP->messageLength);
		TRACE_MID(TRACE_INFO, ("Message payload stored at %.8x, length %d",
							   memberP->messageP, memberP->messageLength));
        TRACE_MID(TRACE_INFO, ("STORAGE(4): alloc %.8x", memberP->messageP));


	}
	else
	{
		TRACE_MID(TRACE_INFO, ("Empty reply message - nothing stored"));
		memberP->messageP = NULL;
	}

	/* if this is not the last reply of the aggregate, then just return */
	if (memberP->aggregateInstanceP->outstandingReplies)
	{

		/* log change to aggregate if recovery is on */
		if (memberP->aggregateInstanceP->deployedAggregateP->persistent)
		{
			retCode = logReply(memberP);

			if (retCode != TT_SUCCESS)
			{
				TRACE_MID(TRACE_ERROR, ("Failed to log aggregate reply details"));
				TRACE_RETURN(RSN_012, TT_LOG_ERROR);
			}
		}

		TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
		threadMutexUnlock(aggregateChainMutex);

		TRACE_RETURN(RSN_007, TT_SUCCESS);
	}

	/* remove all messages for this aggregate from the log if recovery is on */
	if (memberP->aggregateInstanceP->deployedAggregateP->persistent)
	{
#if 1
		/* read all records which match this aggregate id and commit */
		retCode = logRemoveEntries(memberP->aggregateInstanceP->aggregateId, FALSE);
#else
		/* write an aggregate FINISH record to log */
		retCode = logFinish(memberP->aggregateInstanceP->aggregateId,
			                memberP->aggregateInstanceP->deployedAggregateP->name);

#endif

		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to log aggregate FINISH record"));
			TRACE_RETURN(RSN_012, TT_LOG_ERROR);
		}

#if 0
		/* signal log thread to remove aggregate log entries */
		TRACE_MID(TRACE_INFO, ("About to lock logThreadWorkMutex"));
		threadMutexLock(logThreadWorkMutex);

		/* add an aggregate complete work entry for the log thread */
        logTruncReqP = (logTruncationRequest_t *)malloc(sizeof(logTruncationRequest_t));
		memset(logTruncReqP, 0, sizeof(logTruncationRequest_t));
		memcpy(logTruncReqP->aggregateId, memberP->aggregateInstanceP->aggregateId, MQ_CORREL_ID_LENGTH);

		INSERT_AFTER_IN_LINKED_LIST(logTruncationRequestAnchorP, logTruncReqP);

		numLogTruncationRequests++;
		threadCondSignal(logThreadWorkCondVar);
		TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
		threadMutexUnlock(logThreadWorkMutex);
#endif
	}

	/* this is the last reply so build the aggregate message (if required) and propagate */

	/* first, remove this aggregate from the aggregate chain */
	aggregateP = memberP->aggregateInstanceP;
	TRACE_MID(TRACE_INFO, ("Unchaining aggregate %.8x", aggregateP));

	REMOVE_FROM_LINKED_LIST(aggregateP);
	REMOVE_FROM_ALARM_CHAIN(aggregateP);

	/* unlock the aggregate chain to allow other threads to proceed */
	TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
	threadMutexUnlock(aggregateChainMutex);

	/******************************************************************************/
	/* This is the last reply - the aggregate is now complete.                    */ 
	/* Build the aggregate message (if required) and propagate.                   */
	/******************************************************************************/
	retCode = buildCompletedAggregateReplyMessage(message,
									              aggregateP,
										          &newMsg);
	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to build aggregated reply message(s)"));
		TRACE_RETURN(RSN_008, retCode);
	}

	newDestList = NULL;
	if (aggregateP->commonMsgId)
	{
		if (aggregateP->deployedAggregateP->outputToMQReply)
		{
			/* 
			 * Set the ReplyToQ and ReplyToQueueManager in output message 
			 */
			retCode = swapMessageReplyDetails(newMsg, 
											  aggregateP->memberAnchor.nextP->replyToQueue,
											  aggregateP->memberAnchor.nextP->replyToQueueManager,
											  NULL,  /* don't save current value */
											  0,
											  NULL,  /* don't save current value */
											  0);

			if (retCode != TT_SUCCESS)
			{
				/* failed to modify message reply details */
				TRACE_MID(TRACE_ERROR, ("Failed to replace reply details in new message"));
				TRACE_RETURN(RSN_009, TT_FAILED_TO_SWAP_MESSAGE_REPLY_DETAILS);
			}

			/*
			 * When the message is processed by either
			 *   - an MQReply node or 
			 *   - an MQOutput node with DestinationMode set to replyToQueue
			 * the MsgId and CorrelId are swapped by MQSI.  
			 * In order that the final CorrelId value will be the same as the
			 * original request MsgId value, we need to nullify this action by 
			 * 'pre-swapping' these values before the message is propagated.
			 */
			retCode = swapMessageIdAndCorrelId(newMsg);

			if (retCode != TT_SUCCESS)
			{
				/* failed to swap id's */
				TRACE_MID(TRACE_ERROR, ("Failed to swap MsgId and CorrelId values in new message"));
				TRACE_RETURN(RSN_010, TT_FAILED_TO_SWAP_CORREL_AND_MSG_IDS);
			}
		}
		else
		{
			TRACE_MID(TRACE_INFO, ("setOutputReplyToQueue is FALSE"));

		}

		if (aggregateP->deployedAggregateP->outputToDestinationList)
		{
			/* 
			 * Add the ReplyToQ and ReplyToQMgr details to the destination listAdd reply to queue and queue manager from first member - assume all members
			 * share the same details
			 */
			retCode = addAddressToDestinationList(destinationList,
									  aggregateP->memberAnchor.nextP->replyToQueue,
									  aggregateP->memberAnchor.nextP->replyToQueueManager,
									  &newDestList);
			if (retCode != TT_SUCCESS)
			{
				TRACE_MID(TRACE_ERROR, ("Failed to add reply details to destination list"));
				TRACE_RETURN(RSN_011, TT_FAILED_TO_SET_DESTINATION_LIST);
			}
		}
	}

	/* propagate message */
	TRACE_MID(TRACE_INFO, ("Propagating message to 'out' terminal"));
	retCode = propagateMessage(contextP,
							   L"out",
							   newMsg,
							   newDestList ? newDestList : destinationList,
							   exceptionList,
						       TRUE); /* this is a new message */

	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'out' terminal"));
		TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
	}

	/* delete the new message */
    cniDeleteMessage(&rc,
		             newMsg);

	if (newDestList)
	{
	    TRACE_MID(TRACE_INFO, ("Freeing destination list structure"));
		/* delete the new destination list */
        cniDeleteMessage(&rc,
		                 newDestList);
	}

	/* free each member structure */
	freeAllAggregateMembers(aggregateP);
	free(aggregateP);
    TRACE_MID(TRACE_INFO, ("STORAGE(50): free %.8x", aggregateP));

	TRACE_RETURN(RSN_013, TT_SUCCESS);
}

/******************************************************************************/
/* buildCompletedAggregateReplyMessage                                        */
/******************************************************************************/
int buildCompletedAggregateReplyMessage(IN CciMessage *message,
						                IN aggregateInstance_t *aggregateP,
							            OUT CciMessage **newMessageP)
{
	char *aggregatedBodyP;
	CciMessageContext *msgContextP;
	int rc,
		numberOfReplies,
		aggregatedBodyLength;
	CciMessage *newMsg;
	CciElement *correlIdElementP;
    struct CciByteArray byteArrayDetails;
    MQMD md = {MQMD_DEFAULT};           /* Message Descriptor */

    TRACE_ENTRY("buildCompletedAggregateReplyMessage", ("message=%.8x, "
		                                      "aggregateP=%.8x", message,
		                                       aggregateP));

	/* 
	 * The aggregate descriptor we've been passed is no longer part of either
	 * the aggregate chain or the alarm chain.  We can process it and then
	 * free it.
	 */

	/* deal with aggregated reply */
	rc = buildAggregateMessagePayload(aggregateP,
							          &aggregatedBodyP,
							          &aggregatedBodyLength,
								      &numberOfReplies);

	/* create a new message, using current message as a template */
	msgContextP = cniGetMessageContext(&rc, message);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to get message context "
			                    "(cniGetMessageContext rc=%d)", rc));
		TRACE_RETURN(RSN_014, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}
	TRACE_MID(TRACE_INFO, ("Message context got at %.8x", msgContextP));

    newMsg = cniCreateMessage(&rc, msgContextP);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to create new message "
			                    "(cniCreateMessage rc=%d)", rc));
		TRACE_RETURN(RSN_015, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}
	TRACE_MID(TRACE_INFO, ("New message created at %.8x", newMsg));

	/* Copy old message into new */
	rc = copyMessage(newMsg, message);
	if (rc != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to copy old message to new message"));
		TRACE_RETURN(RSN_016, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}

	/* Set the CorrelId of new message to MsgId of original message(s) */
	if (aggregateP->commonMsgId)
	{
		/* address CorrelId element in new message */
		correlIdElementP = getMessageElement(newMsg, CORREL_ID);

		if (!correlIdElementP)
		{
			/* generate an error, exit */
			TRACE_MID(TRACE_ERROR, ("CorrelId element not found in new message"));
			TRACE_RETURN(RSN_017, TT_FAILED_TO_CREATE_NEW_MESSAGE);
		}

		byteArrayDetails.pointer = (void *)(aggregateP->memberAnchor.nextP->originalMessageId);
		byteArrayDetails.size = MQ_CORREL_ID_LENGTH;
		TRACE_MID(TRACE_INFO, ("Setting CorrelId to..."));
		TRACE_MIDHEX(TRACE_INFO, byteArrayDetails.pointer, MQ_CORREL_ID_LENGTH);

		cniSetElementByteArrayValue(&rc,
									correlIdElementP,
									&byteArrayDetails);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to set CorrelId (cniSetElementByteArrayValue rc=%d)", rc));
			TRACE_RETURN(RSN_018, TT_FAILED_TO_CREATE_NEW_MESSAGE);
		}
		TRACE_MID(TRACE_INFO, ("CorrelId successfully set to original message id"));
	}

	/********************************************************/
	/* replace the message body with the aggregated message */
	/********************************************************/
    rc = replaceMessageBodyWithBLOB(newMsg,
		                            aggregatedBodyP,
								    aggregatedBodyLength);
	if (rc != CCI_SUCCESS)
	{
	    TRACE_MID(TRACE_ERROR, ("Failed to replace message body"));
		TRACE_RETURN(RSN_019, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}
    TRACE_MID(TRACE_INFO, ("New message payload successfully set"));

	if (traceEnabled)
	{
		traceMessageBuffer(newMsg);
	}

	/* free the aggregated message */
	if (aggregatedBodyP)
	{
        TRACE_MID(TRACE_INFO, ("Freeing aggregated message"));
		free(aggregatedBodyP);
        TRACE_MID(TRACE_INFO, ("STORAGE(8): free  %.8x", aggregatedBodyP));
	}

	/* return new message to caller */
    TRACE_MID(TRACE_INFO, ("Returning new message at %.8x", newMsg));
	*newMessageP = newMsg;

	TRACE_RETURN(RSN_020, TT_SUCCESS);
}

/******************************************************************************/
/* deleteMessageBody                                                          */
/******************************************************************************/
int deleteMessageBody(INOUT CciMessage *message)
{
	int rc;
	CciElement *mqmdElementP,
			   *siblingP;
    CciChar elementName[256];
    CciSize elementNameSize;

    TRACE_ENTRY("deleteMessageBody", ("message=%.8x", message));

	/* remove all sibling elements after MQMD */
	mqmdElementP = getMessageElement(message, MQMD_ELEMENT);

	if (!mqmdElementP)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to address MQMD element"));
		TRACE_RETURN(RSN_021, TT_FAILED_TO_GET_MESSAGE_MQMD);
	}
	TRACE_MID(TRACE_INFO, ("Addressed MQMD element at %.8x", mqmdElementP));

	/* delete all elements following the MQMD */
	siblingP = cniNextSibling(&rc, mqmdElementP);
	while(siblingP)
	{
		/* Get the element's name */
		elementNameSize = cniElementName(&rc, 
										 siblingP, 
										 (CciChar *)&elementName, 
										 sizeof(elementName));

		/* empty wc string seems to have a length of 1 */
		if (elementNameSize <= 1)
		{
			TRACE_MID(TRACE_INFO, ("Deleting element with no name"));
		}
		else
		{
		    TRACE_MID(TRACE_INFO, ("Deleting element with name %S", elementName));
		}

		cniDetach(&rc, siblingP);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to detach sibling element "
									"(cniDetach rc=%d)", rc));
			TRACE_RETURN(RSN_022, TT_FAILED_TO_DELETE_MESSAGE_BODY);
		}
		TRACE_MID(TRACE_INFO, ("Detached sibling element"));

		/* address the next sibling */
	    siblingP = cniNextSibling(&rc, mqmdElementP);
	}

	TRACE_RETURN(RSN_023, TT_SUCCESS);
}

/******************************************************************************/
/* replaceMessageBodyWithBLOB                                                 */
/******************************************************************************/
int replaceMessageBodyWithBLOB(INOUT CciMessage *message,
						       IN char *payload,
							   IN int payloadLength)
{
	int rc;
	CciElement *bodyElementP,
		       *childElementP,
			   *rootElementP;
	CciChar nullString[] = L"";
    struct CciByteArray byteArrayDetails;

    TRACE_ENTRY("replaceMessageBodyWithBLOB", ("message=%.8x, payload=%.8x, "
		                                       "payloadLength=%d",
		                                        message, payload, payloadLength));

	/* address the root element of new message */
    rootElementP = cniRootElement(&rc, message);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to address message root "
			                    "(cniRootElement rc=%d)", rc));
		TRACE_RETURN(RSN_024, TT_FAILED_TO_GET_MESSAGE_ROOT);
	}
	TRACE_MID(TRACE_INFO, ("Addressed message root element at %.8x", 
		                    rootElementP));

	/* remove existing body from message */
    rc = deleteMessageBody(message);
	if (rc != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to delete existing message body element"));
		TRACE_RETURN(RSN_025, TT_FAILED_TO_DELETE_MESSAGE_BODY);
	}

    /* create new body using the blob parser */
	bodyElementP = cniCreateElementAsLastChildUsingParser(&rc,
		                                                  rootElementP,
														  L"none");
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to create new message body "
			                    "(cniCreateElementAsLastChildUsingParser rc=%d)", rc));
		TRACE_RETURN(RSN_026, TT_FAILED_TO_CREATE_MESSAGE_BODY);
	}
	TRACE_MID(TRACE_INFO, ("New message body created at %.8x", bodyElementP));

	/* now add the payload child element */
	childElementP = cniCreateElementAsLastChild(&rc,
	                                            bodyElementP);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to create second child element "
			                    "(cniCreateElementAsLastChild rc=%d)", rc));
		TRACE_RETURN(RSN_027, TT_FAILED_TO_CREATE_MESSAGE_BODY);
	}
	TRACE_MID(TRACE_INFO, ("Second child element created at %.8x", childElementP));
	cniSetElementName(&rc,
		              childElementP,
					  L"BLOB");
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set second child element name "
			                    "(cniSetElementName rc=%d)", rc));
		TRACE_RETURN(RSN_028, TT_FAILED_TO_CREATE_MESSAGE_BODY);
	}
	TRACE_MID(TRACE_INFO, ("First child element name set to %S", L"BLOB"));
	cniSetElementType(&rc,
		              childElementP,
					  CCI_ELEMENT_TYPE_NAME_VALUE);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set second child element type "
			                    "(cniSetElementType rc=%d)", rc));
		TRACE_RETURN(RSN_029, TT_FAILED_TO_CREATE_MESSAGE_BODY);
	}
	TRACE_MID(TRACE_INFO, ("Second child element type set"));

	/* set the element value to new payload */
	byteArrayDetails.pointer = (void *)payload;
	byteArrayDetails.size = payloadLength;
#if 0
	TRACE_MID(TRACE_INFO, ("Setting body payload to..."));
	TRACE_MIDHEX(TRACE_INFO, payload, payloadLength);
#endif
	cniSetElementByteArrayValue(&rc,
								childElementP,
								&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
	    TRACE_MID(TRACE_ERROR, ("Failed to set payload (cniSetElementByteArrayValue rc=%d)", rc));
		TRACE_RETURN(RSN_030, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}
    TRACE_MID(TRACE_INFO, ("New message payload successfully set"));

	TRACE_RETURN(RSN_031, TT_SUCCESS);
}


/******************************************************************************/
/* processExpiredAggregate                                                    */
/******************************************************************************/
int processExpiredAggregate(IN context_t *contextP,
						    IN aggregateInstance_t *aggregateP,
                            IN CciMessage *message,
		   		            IN CciMessage *destinationList,
	                        IN CciMessage *exceptionList)
{
	char *aggregatedBodyP;
	CciMessageContext *msgContextP;
	int rc,
		retCode,
		numberOfReplies,
		aggregatedBodyLength;
	CciMessage *newMsg;
	CciElement *correlIdElementP;
    struct CciByteArray byteArrayDetails;
    MQMD md = {MQMD_DEFAULT};           /* Message Descriptor */
	static char *functionName = "processExpiredAggregate";
#if 0
	logTruncationRequest_t *logTruncReqP;
#endif

    TRACE_ENTRY("processExpiredAggregate", ("contextP=%.8x, aggregateP=%.8x, "
		                                    "message=%.8x, destinationList=%.8x, "
											"exceptionList=%.8x",
		                                     contextP, aggregateP, message,
											 destinationList,
											 exceptionList));

	/* 
	 * The aggregate descriptor we've been passed is no longer part of either
	 * the aggregate chain or the alarm chain.  We can process it and then
	 * free it.
	 */

	rc = buildAggregateMessagePayload(aggregateP,
							          &aggregatedBodyP,
							          &aggregatedBodyLength,
								      &numberOfReplies);

	/* create a new message using the input message as a template */
	msgContextP = cniGetMessageContext(&rc, message);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to get message context "
			                    "(cniGetMessageContext rc=%d)", rc));
		TRACE_RETURN(RSN_032, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}
	TRACE_MID(TRACE_INFO, ("Message context got at %.8x", msgContextP));

    newMsg = cniCreateMessage(&rc, msgContextP);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to create new message "
			                    "(cniCreateMessage rc=%d)", rc));
		TRACE_RETURN(RSN_033, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}
	TRACE_MID(TRACE_INFO, ("New message created at %.8x", newMsg));

	/* Copy old message into new */
	rc = copyMessage(newMsg, message);
	if (rc != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to copy old message to new message"));
		TRACE_RETURN(RSN_034, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}

	if (aggregateP->commonMsgId)
	{

		/* address CorrelId element in new message */
		correlIdElementP = getMessageElement(newMsg, CORREL_ID);

		if (!correlIdElementP)
		{
			/* generate an error, exit */
			TRACE_MID(TRACE_ERROR, ("CorrelId element not found in new message"));
			TRACE_RETURN(RSN_017, TT_FAILED_TO_CREATE_NEW_MESSAGE);
		}

		byteArrayDetails.pointer = (void *)(aggregateP->memberAnchor.nextP->originalMessageId);
		/*
		byteArrayDetails.pointer = (void *)aggregateP->aggregateId;
		*/
		byteArrayDetails.size = MQ_CORREL_ID_LENGTH;
		TRACE_MID(TRACE_INFO, ("Setting CorrelId to..."));
		TRACE_MIDHEX(TRACE_INFO, byteArrayDetails.pointer, MQ_CORREL_ID_LENGTH);

		cniSetElementByteArrayValue(&rc,
									correlIdElementP,
									&byteArrayDetails);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to set CorrelId (cniSetElementByteArrayValue rc=%d)", rc));
			TRACE_RETURN(RSN_018, TT_FAILED_TO_CREATE_NEW_MESSAGE);
		}
		TRACE_MID(TRACE_INFO, ("CorrelId successfully set to original message id"));
	}

	/* build a new BLOB body if there's aggregated data */
	if (aggregatedBodyLength)
	{
		rc = replaceMessageBodyWithBLOB(newMsg,
										aggregatedBodyP,
										aggregatedBodyLength);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to replace message body"));
			TRACE_RETURN(RSN_037, TT_FAILED_TO_CREATE_NEW_MESSAGE);
		}
		TRACE_MID(TRACE_INFO, ("New message payload successfully set"));
	}
	else
	{
		TRACE_MID(TRACE_INFO, ("Empty message being built"));
	}

	if (traceEnabled)
	{
		traceMessageBuffer(newMsg);
	}

	if (aggregateP->deployedAggregateP->persistent)
	{
#if 1
		/* read all records which match this aggregate id and commit */
		retCode = logRemoveEntries(aggregateP->aggregateId, FALSE);
#else
		/* write an aggregate FINISH record to log */
		retCode = logFinish(aggregateP->aggregateId,
			                aggregateP->deployedAggregateP->name);

#endif
		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to log aggregate FINISH record"));
			TRACE_RETURN(RSN_012, TT_LOG_ERROR);
		}

#if 0
		/* signal log thread to remove aggregate log entries */
		TRACE_MID(TRACE_INFO, ("About to lock logThreadWorkMutex"));

		threadMutexLock(logThreadWorkMutex);

		/* add an aggregate complete work entry for the log thread */
        logTruncReqP = (logTruncationRequest_t *)malloc(sizeof(logTruncationRequest_t));
		memset(logTruncReqP, 0, sizeof(logTruncationRequest_t));
		memcpy(logTruncReqP->aggregateId, aggregateP->aggregateId, MQ_CORREL_ID_LENGTH);

		INSERT_AFTER_IN_LINKED_LIST(logTruncationRequestAnchorP, logTruncReqP);

		numLogTruncationRequests++;
		threadCondSignal(logThreadWorkCondVar);
		TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
		threadMutexUnlock(logThreadWorkMutex);
#endif
	}

	/* propagate message to timed out terminal */
	retCode = propagateMessage(contextP,
							   L"timedOut",
							   newMsg,
							   destinationList,
							   exceptionList,
   							   TRUE); /* this is a new message */

	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'timedOut' terminal"));
		TRACE_RETURN(RSN_039, TT_PROPAGATION_ERROR);
	}

	/* free the aggregated message */
	if (aggregatedBodyP)
	{
        TRACE_MID(TRACE_INFO, ("Freeing aggregated message"));
		free(aggregatedBodyP);
        TRACE_MID(TRACE_INFO, ("STORAGE(28): free  %.8x", aggregatedBodyP));
	}

	/* free each member structure and the aggregate instance itself */
	freeAllAggregateMembers(aggregateP);
	free(aggregateP);
    TRACE_MID(TRACE_INFO, ("STORAGE(29): free  %.8x", aggregateP));

	/* free the message structure */
    cniDeleteMessage(&rc,
		             newMsg);

    TRACE_RETURN_TYPE(RSN_040, int, 0);
}

/******************************************************************************/
/* removeMsgIdIndex                                                           */
/******************************************************************************/
void removeMsgIdIndex(IN msgIdIndexElement_t *msgIndexP)
{

    TRACE_ENTRY("removeMsgIdIndex", ("msgIndexP=%.8x", msgIndexP));

	/* remove entry from message id index */
	REMOVE_FROM_LINKED_LIST(msgIndexP);

	free(msgIndexP);
    TRACE_MID(TRACE_INFO, ("STORAGE(27): free  %.8x", msgIndexP));

	TRACE_RETURN_VOID(RSN_041);
}

/******************************************************************************/
/* freeAllAggregateMembers                                                    */
/******************************************************************************/
void freeAllAggregateMembers(aggregateInstance_t *aggregateP)
{
	aggregateMember_t *memberP,
		              *nextMemberP;
    TRACE_ENTRY("freeAllAggregateMembers", ("aggregateP=%.8x", aggregateP));

	/* address the first member structure */
	memberP = aggregateP->memberAnchor.nextP;

	while (memberP)
	{
	    TRACE_MID(TRACE_INFO, ("Processing member at %.8x", memberP));

		/* address next member structure */
		nextMemberP = memberP->nextP;

		/* free any message buffers */
		if (memberP->messageP)
		{
	        TRACE_MID(TRACE_INFO, ("memberP->messageP = %.8x", &memberP->messageP));
	        TRACE_MID(TRACE_INFO, ("Freeing message at %.8x", memberP->messageP));
			free(memberP->messageP);
            TRACE_MID(TRACE_INFO, ("STORAGE(5): free  %.8x", memberP->messageP));
		}

		/* free any message headers */
		if (memberP->mqmdP)
		{
	        TRACE_MID(TRACE_INFO, ("Freeing message header at %.8x", memberP->mqmdP));
			free(memberP->mqmdP);
            TRACE_MID(TRACE_INFO, ("STORAGE(51): free  %.8x", memberP->mqmdP));
		}

		/* free up member structure itself */
	    TRACE_MID(TRACE_INFO, ("Freeing member structure at %.8x", memberP));
		free(memberP);
        TRACE_MID(TRACE_INFO, ("STORAGE(6): free  %.8x", memberP));

        memberP = nextMemberP;
	}

    TRACE_RETURN_VOID(RSN_042);
}

/******************************************************************************/
/* buildAggregateMessagePayload                                               */
/******************************************************************************/
int buildAggregateMessagePayload(IN aggregateInstance_t *aggregateP,
								 OUT char **bufferP,
								 OUT int *lengthP,
								 OUT int *numberOfRepliesP)
{
	aggregateMember_t *memberP;
	char *localBuffer = NULL,
		 *charP;
	int messageLength = 0,
		numReplies = 0,
		numMembers = 0,
		sepStringLength = 0;
	int	localSeparatorStringLength;
	char *localSeparatorString;

    TRACE_ENTRY("buildAggregateMessagePayload", ("aggregateP=%.8x, bufferP=%.8x",
							                     aggregateP, bufferP));

	/* calculate length of any separator string */
	localSeparatorStringLength = 0;
    localSeparatorString = "";
	if (aggregateP->deployedAggregateP->separatorStringSet)
	{
		localSeparatorString = aggregateP->deployedAggregateP->separatorString;
		localSeparatorStringLength = aggregateP->deployedAggregateP->separatorStringLength;
	    TRACE_MID(TRACE_INFO, ("separatorStringSet is TRUE, length=%d", 
			                    localSeparatorStringLength));
	}

	/* calculate the size of buffer required for aggregated message */
	memberP = aggregateP->memberAnchor.nextP;

	while (memberP)
	{
        TRACE_MID(TRACE_INFO, ("Processing member at %.8x", memberP));

		/* add separator string, if any */
        TRACE_MID(TRACE_INFO, ("Adding separator string length of %d bytes", 
			                    localSeparatorStringLength));
		messageLength += localSeparatorStringLength;

		/* ignore members with outstanding replies */
		if (!memberP->replyArrivalTime)
		{
	        TRACE_MID(TRACE_INFO, ("Member has no reply, ignoring member"));
			memberP = memberP->nextP;
			continue;
		}

		/* add reply message length */
        messageLength += memberP->messageLength;
	    TRACE_MID(TRACE_INFO, ("Aggregate member (%.8x), message length %d, "
			                   "total=%d",
			                   memberP, memberP->messageLength,
							   messageLength));
		numReplies++;
		memberP = memberP->nextP;
	}

	/* there will always be one less separator string than reply messages */
	if (messageLength)
	{
		messageLength -= localSeparatorStringLength;
	}

	/* just return if no messages and no separator string specified */
	if (!messageLength)
	{
		TRACE_MID(TRACE_INFO, ("No replies found - returning empty aggregated message"));
		*lengthP = 0;
		*bufferP = NULL;
		*numberOfRepliesP = numReplies;
		TRACE_RETURN(RSN_043, TT_SUCCESS);
	}

	/* allocate aggregated reply buffer */
	localBuffer = (char *)malloc(messageLength);
	TRACE_MID(TRACE_INFO, ("Message buffer allocated at %.8x, %d bytes",
			               localBuffer,
						   messageLength));
    TRACE_MID(TRACE_INFO, ("STORAGE(7): alloc %.8x", localBuffer));

    /* move message data into buffer - assume members are in seq order */
	charP = localBuffer;
	memberP = aggregateP->memberAnchor.nextP;

	while (memberP)
	{
		if (memberP->messageP)
		{
			TRACE_MID(TRACE_INFO, ("Adding %d bytes from member (%.8x), "
					               "at %.8x",
								   memberP->messageLength,
								   memberP,
								   charP));
			memcpy(charP, memberP->messageP, memberP->messageLength);
			charP += memberP->messageLength;
		}
		memberP = memberP->nextP;

		if (memberP && localSeparatorStringLength)
		{
			memcpy(charP, localSeparatorString, localSeparatorStringLength);
			charP += localSeparatorStringLength;
		}
	}

	TRACE_MID(TRACE_INFO, ("Returning %d repl%s",
		                   numReplies,
						   numReplies>1 ? "ies" : "y"));
	TRACE_MID(TRACE_INFO, ("Returning aggregated message at %.8x, length=%d",
		                   localBuffer,
						   messageLength));

	*lengthP = messageLength;
	*bufferP = localBuffer;
	*numberOfRepliesP = numReplies;

#if 1
	TRACE_MID(TRACE_INFO, ("Aggregated message data is..."));
	TRACE_MIDHEX(TRACE_INFO, *bufferP, *lengthP);
#endif

    TRACE_RETURN(RSN_044, TT_SUCCESS);
}

/******************************************************************************/
/* processTimerMessage                                                        */
/******************************************************************************/
int processTimerMessage(IN context_t *contextP,
 			   	        IN CciMessage *destinationList,
		                IN CciMessage *exceptionList,
		                IN CciMessage *message)
{
	CciSize bufferSize;
	int rc;
	void *messageIdElementP;
    struct CciByteArray byteArrayDetails;
	char messageIdArray[MQ_MSG_ID_LENGTH];
    CciSize numBytes;
    msgIdIndexElement_t *msgIndexP;
	aggregateInstance_t *aggregateP;

    TRACE_ENTRY("processTimerMessage", ("contextP=%.8x, destinationList=%.8x, "
		                     "exceptionList=%.8x, message=%.8x",
							 contextP, destinationList,
							 exceptionList, message));

	/* get the message length */
	rc = getMessageLength(message, &bufferSize);
	if (rc != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to get message length"));
		TRACE_RETURN(RSN_045, TT_FAILED_TO_GET_MESSAGE_LENGTH);
	}

	/* check for non-zero data length - this should not happen */
	if (bufferSize != sizeof(MQMD)) 
	{
		/* unexpected message so send to failure output */
		rc = propagateMessage(contextP,
						      L"failure",
						      message,
						      destinationList,
						      exceptionList,
						      FALSE); /* not a new message */

		if (rc != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message "
					                "to 'failure' terminal"));
			TRACE_RETURN(RSN_046, TT_PROPAGATION_ERROR);
		}

		TRACE_RETURN(RSN_047, TT_SUCCESS);
	}

	/* Process the TIMED OUT message */
	TRACE_MID(TRACE_INFO, ("Message is a TIMED OUT message (has zero length data)"));

	/* extract the messageId */
	messageIdElementP = getMessageElement(message, MESSAGE_ID);

	if (!messageIdElementP)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Message Id not found in message"));
		TRACE_RETURN(RSN_048, TT_MESSAGE_ID_NOT_FOUND);
	}

	/* copy message id, assume it is of type CCI_VALUE_TYPE_BYTEARRAY */
	byteArrayDetails.pointer = (void *)messageIdArray;
	byteArrayDetails.size = sizeof(messageIdArray);
	numBytes = cniElementByteArrayValue(&rc,
										messageIdElementP,
										&byteArrayDetails);
	if (rc != CCI_SUCCESS)
	{
		/* generate an error, exit */
		TRACE_MID(TRACE_ERROR, ("Failed to extract Message Id from message"));
		TRACE_RETURN(RSN_049, TT_MESSAGE_ID_NOT_FOUND);
	}

	TRACE_MID(TRACE_INFO, ("Message Id value is..."));
	TRACE_MIDHEX(TRACE_INFO, messageIdArray, MQ_MSG_ID_LENGTH);

	/* search the msg id index for a match on message id */
	TRACE_MID(TRACE_INFO, ("About to lock aggregateChainMutex"));
	threadMutexLock(aggregateChainMutex);

	msgIndexP = findMsgIndexElement(messageIdArray);

	if (!msgIndexP)
	{
		/* unexpected message so send to failure output */
		TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
		threadMutexUnlock(aggregateChainMutex);

		/* assume message has timed out, propagate to failure terminal */
		rc = propagateMessage(contextP,
						      L"failure",
						      message,
						      destinationList,
						      exceptionList,
							  FALSE); /* not a new message */

		if (rc != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message "
						            "to 'failure' terminal"));
			TRACE_RETURN(RSN_050, TT_PROPAGATION_ERROR);
		}

		TRACE_RETURN(RSN_051, TT_SUCCESS);
	}

	/* found a hit, so process the timed out aggregate */
	aggregateP = (aggregateInstance_t *)(msgIndexP->memberP);

	/* remove from timed out chain */
	REMOVE_FROM_LINKED_LIST(aggregateP);

	if (traceEnabled)
	{
		char hexBuffer[2*sizeof(aggregateP->aggregateId) + 1];
		dispHex((char *)(aggregateP->aggregateId),
				MQ_CORREL_ID_LENGTH,
				hexBuffer);
		TRACE_MID(TRACE_INFO, ("Expired aggregate is %s", hexBuffer));
	}

	/* remove the message id index element */
	removeMsgIdIndex(msgIndexP);

	TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
	threadMutexUnlock(aggregateChainMutex);

	/* now build the aggregate message (if required) and propagate */
	rc = processExpiredAggregate(contextP,
							     aggregateP,
								 message,
								 destinationList,
								 exceptionList);
	if (rc != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to process expired reply messages"));
		TRACE_RETURN(RSN_052, rc);
	}

	TRACE_RETURN(RSN_001, TT_SUCCESS);
}


